library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;

-- Package Declaration Section
package comp_encoder is

	-----------------------------------------------------------------------------------------------------------------
	-------------------------------------------------- ENCODER TOP --------------------------------------------------
	-----------------------------------------------------------------------------------------------------------------
	
	-- Encoder top entity
	component top_encoder is
		generic (
			-- 0: "signed" type samples, 1: "unsigned" type samples
			SMPL_TYPE_G				: std_logic;
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G			: std_logic_vector(1 downto 0);
			-- 0: Encoder Header disabled, 1: Encoder Header enabled
			HEADER_EN_G				: std_logic;
			-- 1: Full prediction mode, 0: Reduced prediction mode
			PREDICT_MODE_G			: std_logic;
			-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
			LSUM_TYPE_G				: std_logic_vector(1 downto 0);
			-- 1: Custom weight init, 0: Default weight init
			W_INIT_TYPE_G			: std_logic;
			-- 1: enabled, 0: disabled
			PER_ERR_LIM_UPD_G		: std_logic;
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G		: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G		: std_logic;
			REL_ERR_BAND_TYPE_G		: std_logic;
			-- 00: Sample-Adaptive Entropy, 01: Hybrid Entropy, 10: Block-Adaptive Entropy
			ENCODER_TYPE_G			: std_logic_vector(1 downto 0);
			-- Flag to add the "Weight Initialization Table"
			W_INIT_TABL_FLAG_G		: std_logic;
			-- Flag to add the "Weight Exponent Offset Table"
			W_EXP_OFF_TABL_FLAG_G	: std_logic;
			-- Flag to add the "Damping Table"
			DAMP_TABLE_FLAG_G		: std_logic;
			-- Flag to add the "Offset Table"
			OFFSET_TABLE_FLAG_G		: std_logic;
			-- Flag to add the "Accumulator Initialization Table"
			ACCU_INIT_TABLE_FLAG_G	: std_logic;
			-- 1: Restricted set of code options used, 0: Restricted set of code options NOT used
			RESTRICT_CODE_G			: std_logic;
			-- User Defined Data
			UDEF_DATA_G				: std_logic_vector(7 downto 0);
			-- Array with values -> 00: unsigned integer, 01: signed integer, 10: float
			SUPL_TABLE_TYPE_G		: supl_table_type_t;
			-- Array with values -> From 0 to 15 (check Table 3-1)
			SUPL_TABLE_PURPOSE_G	: supl_table_purpose_t;
			-- Array with values -> 00: zero-dimensional, 01: one-dimensional, 10: two-dimensional-zx, 11: two-dimensional-yx
			SUPL_TABLE_STRUCT_G		: supl_table_struct_t;
			-- Array with values -> Suplementary User-Defined Data
			SUPL_TABLE_UDATA_G		: supl_table_udata_t;
			-- 0: Little Endian order, 1: Big Endian order
			ENDIANNESS_G			: std_logic
		);
		port (
			clock_i					: in  std_logic;
			reset_i					: in  std_logic;

			enable_i				: in  std_logic;
			enable_o				: out std_logic;
			img_coord_i				: in  img_coord_t;
			img_coord_o				: out img_coord_t;
			
			init_fifo_o				: out std_logic;
			image_start_o			: out std_logic;							-- Flag to inform a new image is coming out
			image_end_o				: out std_logic;							-- Flag to inform current image has been finished
			
			err_lim_upd_en_i	    : in  std_logic;
			err_lim_val_i			: in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i			: in  std_logic_vector(D_C-1 downto 0);
			valid_data_o			: out std_logic;							-- Flag to inform valid data is in the output line
			output_data_o			: out std_logic_vector(B_C*8-1 downto 0)
		);
	end component top_encoder;

	-----------------------------------------------------------------------------------------------------------------
	------------------------------------------------- ENCODER HEADER ------------------------------------------------
	-----------------------------------------------------------------------------------------------------------------
	
	-- Encoder header top
	component top_enc_header is
		generic (
			-- 0: "signed" type samples, 1: "unsigned" type samples
			SMPL_TYPE_G				: std_logic;
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G			: std_logic_vector(1 downto 0);
			-- 1: Full prediction mode, 0: Reduced prediction mode
			PREDICT_MODE_G			: std_logic;
			-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
			LSUM_TYPE_G				: std_logic_vector(1 downto 0);
			-- 1: Custom weight init, 0: Default weight init
			W_INIT_TYPE_G			: std_logic;
			-- 1: enabled, 0: disabled
			PER_ERR_LIM_UPD_G		: std_logic;
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G		: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G		: std_logic;
			REL_ERR_BAND_TYPE_G		: std_logic;
			-- 00: Sample-Adaptive Entropy, 01: Hybrid Entropy, 10: Block-Adaptive Entropy
			ENCODER_TYPE_G			: std_logic_vector(1 downto 0);
			-- Flag to add the "Weight Initialization Table"
			W_INIT_TABL_FLAG_G		: std_logic;
			-- Flag to add the "Weight Exponent Offset Table"
			W_EXP_OFF_TABL_FLAG_G	: std_logic;
			-- Flag to add the "Damping Table"
			DAMP_TABLE_FLAG_G		: std_logic;
			-- Flag to add the "Offset Table"
			OFFSET_TABLE_FLAG_G		: std_logic;
			-- Flag to add the "Accumulator Initialization Table"
			ACCU_INIT_TABLE_FLAG_G	: std_logic;
			-- 1: Restricted set of code options used, 0: Restricted set of code options NOT used
			RESTRICT_CODE_G			: std_logic;
			-- User Defined Data
			UDEF_DATA_G				: std_logic_vector(7 downto 0);
			-- Array with values -> 00: unsigned integer, 01: signed integer, 10: float
			SUPL_TABLE_TYPE_G		: supl_table_type_t;
			-- Array with values -> From 0 to 15 (check Table 3-1)
			SUPL_TABLE_PURPOSE_G	: supl_table_purpose_t;
			-- Array with values -> 00: zero-dimensional, 01: one-dimensional, 10: two-dimensional-zx, 11: two-dimensional-yx
			SUPL_TABLE_STRUCT_G		: supl_table_struct_t;
			-- Array with values -> Suplementary User-Defined Data
			SUPL_TABLE_UDATA_G		: supl_table_udata_t
		);
		port (
			enc_header_width_o		: out integer;
			enc_header_data_o		: out std_logic_vector(4095 downto 0)
		);
	end component top_enc_header;

	-- Image/Predictor/Encoder Metadata	
	component metadata_img is
		generic (
			-- 0: "signed" type samples, 1: "unsigned" type samples
			SMPL_TYPE_G			: std_logic;
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 00: Sample-Adaptive Entropy, 01: Hybrid Entropy, 10: Block-Adaptive Entropy
			ENCODER_TYPE_G		: std_logic_vector(1 downto 0);
			-- User Defined Data
			UDEF_DATA_G			: std_logic_vector(7 downto 0);
			-- Array with values -> 00: unsigned integer, 01: signed integer, 10: float
			SUPL_TABLE_TYPE_G	: supl_table_type_t;
			-- Array with values -> From 0 to 15 (check Table 3-1)
			SUPL_TABLE_PURPOSE_G: supl_table_purpose_t;
			-- Array with values -> 00: zero-dimensional, 01: one-dimensional, 10: two-dimensional-zx, 11: two-dimensional-yx
			SUPL_TABLE_STRUCT_G	: supl_table_struct_t;
			-- Array with values -> Suplementary User-Defined Data
			SUPL_TABLE_UDATA_G	: supl_table_udata_t
		);
		port (			
			md_img_width_o		: out integer;
			md_img_data_o		: out std_logic_vector(2047 downto 0)
		);
	end component metadata_img;

	component metadata_pred is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G			: std_logic_vector(1 downto 0);
			-- 1: Full prediction mode, 0: Reduced prediction mode
			PREDICT_MODE_G			: std_logic;
			-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
			LSUM_TYPE_G				: std_logic_vector(1 downto 0);
			-- 1: Custom weight init, 0: Default weight init
			W_INIT_TYPE_G			: std_logic;
			-- 1: enabled, 0: disabled
			PER_ERR_LIM_UPD_G		: std_logic;
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G		: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G		: std_logic;
			REL_ERR_BAND_TYPE_G		: std_logic;
			-- Flag to add the "Weight Initialization Table"
			W_INIT_TABL_FLAG_G		: std_logic;
			-- Flag to add the "Weight Exponent Offset Table"
			W_EXP_OFF_TABL_FLAG_G	: std_logic;
			-- Flag to add the "Damping Table"
			DAMP_TABLE_FLAG_G		: std_logic;
			-- Flag to add the "Offset Table"
			OFFSET_TABLE_FLAG_G		: std_logic
		);
		port (			
			md_pred_width_o			: out integer;
			md_pred_data_o			: out std_logic_vector(2047 downto 0)
		);
	end component metadata_pred;

	component metadata_encod is
		generic (
			-- "00": Sample-Adaptive Entropy, "01": Hybrid Entropy, "10": Block-Adaptive Entropy
			ENCODER_TYPE_G			: std_logic_vector(1 downto 0);
			-- Flag to add the "Accumulator Initialization Table"
			ACCU_INIT_TABLE_FLAG_G	: std_logic;
			-- 1: Restricted set of code options used, 0: Restricted set of code options NOT used
			RESTRICT_CODE_G			: std_logic
		);
		port (
			md_enc_width_o			: out integer;
			md_enc_data_o			: out std_logic_vector(511 downto 0)
		);
	end component metadata_encod;

	-----------------------------------------------------------------------------------------------------------------
	-------------------------------------------------- ENCODER BODY -------------------------------------------------
	-----------------------------------------------------------------------------------------------------------------
	
	-- Encoder body top
	component top_enc_body is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- 00: Sample-Adaptive Entropy, 01: Hybrid Entropy, 10: Block-Adaptive Entropy
			ENCODER_TYPE_G		: std_logic_vector(1 downto 0);
			-- Maximum number of bits (size) of every encoded data sample
			MAX_NUM_BITS_G		: integer;			
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic
		);
		port (
			clock_i			 : in  std_logic;
			reset_i			 : in  std_logic;
			
			enable_i		 : in  std_logic;
			enable_o		 : out std_logic;
			img_coord_i		 : in  img_coord_t;
			img_coord_o		 : out img_coord_t;
			
			err_lim_upd_en_i : in  std_logic;
			err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
			
			enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
			enc_body_data_o	 : out std_logic_vector(Umax_C+D_C-1 downto 0)	-- "Rk(j)" (codeword)
		);
	end component top_enc_body;

	-- Sample Adaptive Entropy Coder (top) module
	component smpl_adap_wrapper is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- Maximum number of bits (size) of every encoded data sample
			MAX_NUM_BITS_G		: integer;
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic
		);
		port (
			clock_i			 : in  std_logic;
			reset_i			 : in  std_logic;

			enable_i		 : in  std_logic;
			enable_o		 : out std_logic;
			img_coord_i		 : in  img_coord_t;
			img_coord_o		 : out img_coord_t;

			err_lim_upd_en_i : in  std_logic;
			err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
			
			enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
			codeword_o		 : out std_logic_vector(Umax_C+D_C-1 downto 0)	-- "Rk(j)" (codeword)
		);
	end component smpl_adap_wrapper;
	
	component smpl_adap_gpo2_code is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- Maximum number of bits (size) of every encoded data sample
			MAX_NUM_BITS_G		: integer;			
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic
		);
		port (
			clock_i			 : in  std_logic;
			reset_i			 : in  std_logic;
			
			enable_i		 : in  std_logic;
			enable_o		 : out std_logic;
			img_coord_i		 : in  img_coord_t;
			img_coord_o		 : out img_coord_t;

			err_lim_upd_en_i : in  std_logic;
			err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
			
			accu_i			 : in  std_logic_vector(D_C*2-1 downto 0);	-- "Σz(t)" (accumulator)
			counter_i		 : in  std_logic_vector(D_C-1 downto 0);	-- "Γ(t)"  (counter)
			
			enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
			codeword_o		 : out std_logic_vector(Umax_C+D_C-1 downto 0) -- "Rk(j)" (codeword)
		);
	end component smpl_adap_gpo2_code;

	component smpl_adap_statistic is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	 : std_logic_vector(1 downto 0)
		);
		port (
			clock_i			 : in  std_logic;
			reset_i			 : in  std_logic;

			enable_i		 : in  std_logic;
			enable_o		 : out std_logic;
			img_coord_i		 : in  img_coord_t;
			img_coord_o		 : out img_coord_t;

			err_lim_upd_en_i : in  std_logic;
			err_lim_upd_en_o : out std_logic;
			err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
			err_lim_val_o	 : out std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_o	 : out std_logic_vector(D_C-1 downto 0);
			
			accu_o			 : out std_logic_vector(D_C*2-1 downto 0);	-- "Σz(t)" (accumulator)
			counter_o		 : out std_logic_vector(D_C-1 downto 0)		-- "Γ(t)"  (counter)
		);
	end component smpl_adap_statistic;

	-- Sample Adaptive Entropy Coder (top) module
	component hybrid_wrapper is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- Maximum number of bits (size) of every encoded data sample
			MAX_NUM_BITS_G		: integer;			
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic
		);
		port (
			clock_i			 : in  std_logic;
			reset_i			 : in  std_logic;
			
			enable_i		 : in  std_logic;
			enable_o		 : out std_logic;
			img_coord_i		 : in  img_coord_t;
			img_coord_o		 : out img_coord_t;
			
			err_lim_upd_en_i : in  std_logic;
			err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
			
			enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
			rvd_codeword_o	 : out std_logic_vector(Umax_C+D_C-1 downto 0)	-- "R'k(j)" (reversed codeword)
		);
	end component hybrid_wrapper;
	
	component hybrid_compr_img_tail is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	  : std_logic_vector(1 downto 0);
			-- Maximum number of bits (size) of every encoded data sample
			MAX_NUM_BITS_G	  : integer
		);
		port (
			clock_i			  : in  std_logic;
			reset_i			  : in  std_logic;
			
			enable_i		  : in  std_logic;
			enable_o		  : out std_logic;
			img_coord_i		  : in  img_coord_t;
			img_coord_o		  : out img_coord_t;
			
			high_res_accu_i	  : in  std_logic_vector(D_C*2-1 downto 0);				-- "Σ~z(t)" (high resolution accumulator)
			curr_prefix_arr_i : in  array_string_t(0 to NUM_HYBRID_TABLES_G-1)(1 to LONGEST_STRING_SIZE_G);	-- Current prefix for every Hybrid table			
			enc_num_bits_i	  : in  std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
			rvd_codeword_i	  : in  std_logic_vector(Umax_C+D_C-1 downto 0);		-- "R'k(j)" (reversed codeword)

			out_num_bits_o	  : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
			out_codeword_o	  : out std_logic_vector(Umax_C+D_C-1 downto 0)
		);
	end component hybrid_compr_img_tail;

	component hybrid_low_entr_code is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- Maximum number of bits (size) of every encoded data sample
			MAX_NUM_BITS_G		: integer;
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic
		);
		port (
			clock_i			  : in  std_logic;
			reset_i			  : in  std_logic;

			enable_i		  : in  std_logic;
			enable_o		  : out std_logic;
			img_coord_i		  : in  img_coord_t;
			img_coord_o		  : out img_coord_t;

			err_lim_upd_en_i  : in  std_logic;
			err_lim_val_i	  : in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i	  : in  std_logic_vector(D_C-1 downto 0);
			
			high_res_accu_i	  : in  std_logic_vector(D_C*2-1 downto 0);			-- "Σ~z(t)" (high resolution accumulator)
			counter_i		  : in  std_logic_vector(D_C-1 downto 0);			-- "Γ(t)"   (counter)
			reconst_bit_i	  : in  std_logic;									-- Lets decoder reconstruct "Σ~z(t)" values
			
			curr_prefix_arr_o : out array_string_t(0 to NUM_HYBRID_TABLES_G-1)(1 to LONGEST_STRING_SIZE_G);	-- Current prefix for every Hybrid table
			enc_num_bits_o	  : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
			rvd_codeword_o	  : out std_logic_vector(Umax_C+D_C-1 downto 0)		-- "R'k(j)" (reversed codeword)
		);
	end component hybrid_low_entr_code;
	
	component hybrid_high_entr_code is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- Maximum number of bits (size) of every encoded data sample
			MAX_NUM_BITS_G		: integer;
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic
		);
		port (
			clock_i			 : in  std_logic;
			reset_i			 : in  std_logic;

			enable_i		 : in  std_logic;
			enable_o		 : out std_logic;
			img_coord_i		 : in  img_coord_t;
			img_coord_o		 : out img_coord_t;
			
			err_lim_upd_en_i : in  std_logic;
			err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
			
			high_res_accu_i	 : in  std_logic_vector(D_C*2-1 downto 0);	-- "Σ~z(t)" (high resolution accumulator)
			high_res_accu_o	 : out std_logic_vector(D_C*2-1 downto 0);	-- "Σ~z(t)" (high resolution accumulator)
			counter_i		 : in  std_logic_vector(D_C-1 downto 0);	-- "Γ(t)"   (counter)
			reconst_bit_i	 : in  std_logic;							-- Lets decoder reconstruct "Σ~z(t)" values
			
			entropy_select_i : in  std_logic;
			entropy_select_o : out std_logic;
			
			enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
			rvd_codeword_o	 : out std_logic_vector(Umax_C+D_C-1 downto 0)	-- "R'k(j)" (reversed codeword)
		);
	end component hybrid_high_entr_code;
	
	component hybrid_statistic is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	 : std_logic_vector(1 downto 0)
		);
		port (
			clock_i			 : in  std_logic;
			reset_i			 : in  std_logic;
			
			enable_i		 : in  std_logic;
			enable_o		 : out std_logic;
			img_coord_i		 : in  img_coord_t;
			img_coord_o		 : out img_coord_t;

			err_lim_upd_en_i : in  std_logic;
			err_lim_upd_en_o : out std_logic;
			err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
			err_lim_val_o	 : out std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_o	 : out std_logic_vector(D_C-1 downto 0);
			
			high_res_accu_o	 : out std_logic_vector(D_C*2-1 downto 0);	-- "Σ~z(t)" (high resolution accumulator)
			counter_o		 : out std_logic_vector(D_C-1 downto 0);	-- "Γ(t)"   (counter)
			reconst_bit_o	 : out std_logic;							-- Lets decoder reconstruct "Σ~z(t)" values
			
			entropy_select_o : out std_logic							-- Selected Entropy method (1=high-entropy, 0=low-entropy)
		);
	end component hybrid_statistic;

	-----------------------------------------------------------------------------------------------------------------
	--------------------------------------- PREDICTOR/ENCODER INTERCONNECTION ---------------------------------------
	-----------------------------------------------------------------------------------------------------------------
	
	-- Predictor/Encoder interconnection entity
	component pred_enc_interconn is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic
		);
		port (
			clock_i				: in  std_logic;
			reset_i				: in  std_logic;
			
			enable_i			: in  std_logic;
			enable_o			: out std_logic;
			img_coord_i			: in  img_coord_t;
			img_coord_o			: out img_coord_t;
			
			init_fifo_o			: out std_logic;
		
			err_lim_upd_en_o	: out std_logic;
			err_lim_val_o		: out std_logic_vector(D_C-1 downto 0);
			data_mp_quan_i		: in  std_logic_vector(D_C-1 downto 0);
			data_mp_quan_o		: out std_logic_vector(D_C-1 downto 0)
		);
	end component pred_enc_interconn;

	-----------------------------------------------------------------------------------------------------------------
	------------------------------------------------- ENCODER PACKER ------------------------------------------------
	-----------------------------------------------------------------------------------------------------------------
	
	component top_packer is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- 0: Encoder Header disabled, 1: Encoder Header enabled
			HEADER_EN_G			: std_logic;
			-- Maximum number of bits (size) of every encoded data sample
			MAX_NUM_BITS_G		: integer
		);
		port (
			clock_i				: in  std_logic;
			reset_i				: in  std_logic;

			enable_header_i		: in  std_logic;
			enable_body_i		: in  std_logic;
			enable_o			: out std_logic;
			img_coord_header_i	: in  img_coord_t;
			img_coord_body_i	: in  img_coord_t;
			img_coord_o			: out img_coord_t;
			
			enc_header_width_i	: in  integer;
			enc_header_data_i	: in  std_logic_vector(4095 downto 0);
			enc_body_num_bits_i	: in  std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
			enc_body_data_i		: in  std_logic_vector(Umax_C+D_C-1 downto 0);
			init_fifo_o			: out std_logic;
			
			image_start_o		: out std_logic;			-- Flag to inform a new image is coming out
			image_end_o			: out std_logic;			-- Flag to inform current image has been finished
			valid_data_o		: out std_logic;			-- Flag to inform valid data is in the output line
			packed_data_o		: out std_logic_vector(B_C*8-1 downto 0)
		);
	end component top_packer;

	-- Parallel XPM synchronous FIFOs 
	component parallel_sync_fifos is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	: std_logic_vector(1 downto 0);
			-- FIFO data width (for writing [input] & reading [output] channels)
			DATA_WIDTH_G	: integer;
			-- FIFO data depth (for writing [input] & reading [output] channels). MUST be a power of 2, and at least 16!!
			DATA_DEPTH_G	: integer;
			-- Number of parallel sync FIFOs to store data (apart from the 'image coordinates')
			NUM_PAR_FIFOS_G	: integer
		);
		port (
			clock_i		 	: in  std_logic;
			reset_i		 	: in  std_logic;
			
			enable_i	 	: in  std_logic;
			enable_o	 	: out std_logic;
			img_coord_i	 	: in  img_coord_t;
			img_coord_o	 	: out img_coord_t;
			
			read_fifo_i	 	: in  std_logic;
			init_fifo_o		: out std_logic;
			
			data_arr_i	 	: in  array_slv_t(0 to NUM_PAR_FIFOS_G-1)(DATA_WIDTH_G-1 downto 0);
			data_arr_o	 	: out array_slv_t(0 to NUM_PAR_FIFOS_G-1)(DATA_WIDTH_G-1 downto 0)
		);
	end component parallel_sync_fifos;
	
end package comp_encoder;