library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;

use work.param_encoder.all;
use work.types_encoder.all;

-- Package Declaration Section
package utils_encoder is

	-- ENTROPY CODER METADATA	
	pure function serial_mdata_enc_block_adapt(mdata_enc_in : in mdata_enc_block_adapt_t) return std_logic_vector;
	pure function serial_mdata_enc_hybrid(mdata_enc_in : in mdata_enc_hybrid_t) return std_logic_vector;
	pure function serial_mdata_enc_smpl_adapt(mdata_enc_in : in mdata_enc_smpl_adapt_t; accu_init_en_in : in std_logic) return std_logic_vector;
	pure function serial_mdata_enc(mdata_enc_in : in mdata_enc_t) return std_logic_vector;

	-- PREDICTOR METADATA
	pure function serial_mdata_pred_smpl_repr(mdata_pred_in : in mdata_pred_smpl_repr_t; damp_en_in : in std_logic; off_en_in : in std_logic) return std_logic_vector;
	pure function serial_mdata_pred_rel_err_limit(mdata_pred_in : in mdata_pred_rel_err_limit_t; per_err_upd_in : in std_logic) return std_logic_vector;
	pure function serial_mdata_pred_abs_err_limit(mdata_pred_in : in mdata_pred_abs_err_limit_t; per_err_upd_in : in std_logic) return std_logic_vector;
	pure function serial_mdata_pred_err_limit_upd_period(mdata_pred_in : in mdata_pred_err_limit_upd_period_t) return std_logic_vector;
	pure function serial_mdata_pred_quant(mdata_pred_in : in mdata_pred_quant_t; fidel_ctrl_in : in std_logic_vector; sample_order_in : in std_logic_vector; per_err_upd_in : in std_logic) return std_logic_vector;
	pure function serial_mdata_pred_weight_tables(mdata_pred_in : in mdata_pred_weight_tables_t; w_init_en_in : in std_logic; w_exp_off_en_in : in std_logic) return std_logic_vector;
	pure function serial_mdata_pred_primary(mdata_pred_in : in mdata_pred_primary_t) return std_logic_vector;
	pure function serial_mdata_pred(mdata_pred_in : in mdata_pred_t; w_init_en_in : in std_logic; w_exp_off_en_in : in std_logic; fidel_ctrl_in : in std_logic_vector; sample_order_in : in std_logic_vector; per_err_upd_in : in std_logic; theta_in : in integer; damp_en_in : in std_logic; off_en_in : in std_logic) return std_logic_vector;
	
	-- IMAGE METADATA
	pure function serial_mdata_img_supl_info(mdata_img_in : in mdata_img_supl_info_t) return std_logic_vector;
	pure function serial_mdata_img_supl_info_arr(mdata_img_arr_in : in mdata_img_supl_info_arr_t) return std_logic_vector;
	pure function serial_mdata_img_essential(mdata_img_in : in mdata_img_essential_t) return std_logic_vector;
	pure function serial_mdata_img(mdata_img_in : in mdata_img_t) return std_logic_vector;
	
	-- TOP ENCODER HEADER
	-- pure function serial_enc_header(enc_header_in : in enc_header_t) return std_logic_vector; --> TEMPORARY DISABLED!
	
	-- STRING FUNCTIONS
	pure function int_to_char(integer_in : in integer) return character;
	pure function char_to_int(char_in : in character) return integer;
	pure function string_to_int(string_in : in string) return integer;
	pure function clean_string(string_in : in string) return integer;
	pure function format_string(string_in : in string; max_length_in : in integer) return string;
	pure function shift_string(string_in : in string; num_steps_in : in integer; direct_in : in std_logic) return string;
	
	-- OTHER FUNCTIONS
	pure function find_pos_low_entr_table(threshold_in : in integer) return integer;
	pure function padding_bits(slv_in : in std_logic_vector; pad_bits_in : in integer; pad_value_in : in integer) return std_logic_vector;
	pure function match_code_prefix(string_arr_in : in array_string_t; table_size_in : in integer; curr_prefix_in : in string) return integer;
	pure function match_flush_prefix(string_arr_in : in array_string_t; table_size_in : in integer; curr_prefix_in : in string) return integer;
	
end package utils_encoder;

-- Package Body Section
package body utils_encoder is

	-------------------------------------------------------------------------------------------------------
	-- ENTROPY CODER METADATA
	-------------------------------------------------------------------------------------------------------
	
	-- Serializes record "mdata_enc_block_adapt_t" to "std_logic_vector" (from Table 5-15)
	pure function serial_mdata_enc_block_adapt(mdata_enc_in : in mdata_enc_block_adapt_t) return std_logic_vector is
		variable ser_mdata_enc_in_v : std_logic_vector(mdata_enc_in.total_width-1 downto 0);
	begin
		ser_mdata_enc_in_v(0 downto 0)	:= mdata_enc_in.reserved_1;
		ser_mdata_enc_in_v(2 downto 1)	:= mdata_enc_in.block_size;
		ser_mdata_enc_in_v(3 downto 3)	:= mdata_enc_in.restr_code_opt_flag;
		ser_mdata_enc_in_v(mdata_enc_in.total_width-1 downto 4) := mdata_enc_in.ref_smpl_interval;
		
		return ser_mdata_enc_in_v;
	end function serial_mdata_enc_block_adapt;
	
	-- Serializes record "mdata_enc_hybrid_t" to "std_logic_vector" (from Table 5-14)
	pure function serial_mdata_enc_hybrid(mdata_enc_in : in mdata_enc_hybrid_t) return std_logic_vector is
		variable ser_mdata_enc_in_v : std_logic_vector(mdata_enc_in.total_width-1 downto 0);
	begin
		ser_mdata_enc_in_v(4 downto 0)	:= mdata_enc_in.unary_len_limit;
		ser_mdata_enc_in_v(7 downto 5)	:= mdata_enc_in.resc_count_size;
		ser_mdata_enc_in_v(10 downto 8)	:= mdata_enc_in.init_count_exp;
		ser_mdata_enc_in_v(mdata_enc_in.total_width-1 downto 11) := (others => '0');	-- reserved_1
		
		return ser_mdata_enc_in_v;
	end function serial_mdata_enc_hybrid;

	-- Serializes record "mdata_enc_smpl_adapt_t" to "std_logic_vector" (from Table 5-13)
	pure function serial_mdata_enc_smpl_adapt(mdata_enc_in : in mdata_enc_smpl_adapt_t; accu_init_en_in : in std_logic) return std_logic_vector is
		variable ser_mdata_enc_in_v : std_logic_vector(mdata_enc_in.total_width-1 downto 0);
	begin
		ser_mdata_enc_in_v(4 downto 0)	 := mdata_enc_in.unary_len_limit;
		ser_mdata_enc_in_v(7 downto 5)	 := mdata_enc_in.resc_count_size;
		ser_mdata_enc_in_v(10 downto 8)	 := mdata_enc_in.init_count_exp;
		ser_mdata_enc_in_v(14 downto 11) := mdata_enc_in.accu_init_const;
		ser_mdata_enc_in_v(15 downto 15) := mdata_enc_in.accu_init_table_flag;
		
		if (accu_init_en_in = '1') then
			ser_mdata_enc_in_v(mdata_enc_in.total_width-1 downto 16) := mdata_enc_in.accu_init_table(mdata_enc_in.accu_init_table_size-1 downto 0);
		end if;
		
		return ser_mdata_enc_in_v;
	end function serial_mdata_enc_smpl_adapt;
	
	-- Serializes record "mdata_enc_t" to "std_logic_vector" (Additional Table)
	pure function serial_mdata_enc(mdata_enc_in : in mdata_enc_t) return std_logic_vector is
		variable ser_mdata_enc_in_v : std_logic_vector(mdata_enc_in.total_width-1 downto 0);
	begin
		ser_mdata_enc_in_v := mdata_enc_in.enc_subtype_data;

		return ser_mdata_enc_in_v;
	end function serial_mdata_enc;

	-------------------------------------------------------------------------------------------------------
	-- PREDICTOR METADATA
	-------------------------------------------------------------------------------------------------------

	-- Serializes record "mdata_pred_smpl_repr_t" to "std_logic_vector" (from Table 5-12)
	pure function serial_mdata_pred_smpl_repr(mdata_pred_in : in mdata_pred_smpl_repr_t; damp_en_in : in std_logic; off_en_in : in std_logic) return std_logic_vector is
		variable ser_mdata_pred_in_v : std_logic_vector(mdata_pred_in.total_width-1 downto 0);
	begin
		ser_mdata_pred_in_v(4 downto 0)		:= mdata_pred_in.reserved_1;
		ser_mdata_pred_in_v(7 downto 5)		:= mdata_pred_in.smpl_repr_resolution;
		ser_mdata_pred_in_v(8 downto 8)		:= mdata_pred_in.reserved_2;
		ser_mdata_pred_in_v(9 downto 9)		:= mdata_pred_in.band_var_damp_flag;
		ser_mdata_pred_in_v(10 downto 10)	:= mdata_pred_in.damp_table_flag;
		ser_mdata_pred_in_v(11 downto 11)	:= mdata_pred_in.reserved_3;
		ser_mdata_pred_in_v(15 downto 12)	:= mdata_pred_in.fixed_damp_value;
		ser_mdata_pred_in_v(16 downto 16)	:= mdata_pred_in.reserved_4;
		ser_mdata_pred_in_v(17 downto 17)	:= mdata_pred_in.band_var_offset_flag;
		ser_mdata_pred_in_v(18 downto 18)	:= mdata_pred_in.offset_table_flag;
		ser_mdata_pred_in_v(19 downto 19)	:= mdata_pred_in.reserved_5;
		ser_mdata_pred_in_v(23 downto 20)	:= mdata_pred_in.fixed_offset_value;
		
		if (damp_en_in = '1') then
			ser_mdata_pred_in_v(24+mdata_pred_in.damp_table_size-1 downto 24)							 := mdata_pred_in.damp_table_subblock(mdata_pred_in.damp_table_size-1 downto 0);
			if (off_en_in = '1') then
				ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto 24+mdata_pred_in.damp_table_size) := mdata_pred_in.offset_table_subblock(mdata_pred_in.offset_table_size-1 downto 0);
			end if;
		else
			if (off_en_in = '1') then
				ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto 24)								 := mdata_pred_in.offset_table_subblock(mdata_pred_in.offset_table_size-1 downto 0);
			end if;
		end if;
		
		return ser_mdata_pred_in_v;
	end function serial_mdata_pred_smpl_repr;
	
	-- Serializes record "mdata_pred_rel_err_limit_t" to "std_logic_vector" (from Table 5-11)
	pure function serial_mdata_pred_rel_err_limit(mdata_pred_in : in mdata_pred_rel_err_limit_t; per_err_upd_in : in std_logic) return std_logic_vector is
		variable ser_mdata_pred_in_v : std_logic_vector(mdata_pred_in.total_width-1 downto 0);
	begin
		ser_mdata_pred_in_v(0 downto 0)	:= mdata_pred_in.reserved_1;
		ser_mdata_pred_in_v(1 downto 1)	:= mdata_pred_in.rel_err_limit_assig_meth;
		ser_mdata_pred_in_v(3 downto 2)	:= mdata_pred_in.reserved_2;
		ser_mdata_pred_in_v(7 downto 4)	:= mdata_pred_in.rel_err_limit_bit_depth;
		
		if (per_err_upd_in = '0') then		-- If "Periodic Error Limit Update" enabled, relative error limit values subblock is not included!
			ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto 8) := mdata_pred_in.rel_err_limit_val_subblock(mdata_pred_in.rel_err_limit_val_size-1 downto 0);
		end if;
		
		return ser_mdata_pred_in_v;
	end function serial_mdata_pred_rel_err_limit;
	
	-- Serializes record "mdata_pred_abs_err_limit_t" to "std_logic_vector" (from Table 5-10)
	pure function serial_mdata_pred_abs_err_limit(mdata_pred_in : in mdata_pred_abs_err_limit_t; per_err_upd_in : in std_logic) return std_logic_vector is
		variable ser_mdata_pred_in_v : std_logic_vector(mdata_pred_in.total_width-1 downto 0);
	begin
		ser_mdata_pred_in_v(0 downto 0)	:= mdata_pred_in.reserved_1;
		ser_mdata_pred_in_v(1 downto 1)	:= mdata_pred_in.abs_err_limit_assig_meth;
		ser_mdata_pred_in_v(3 downto 2)	:= mdata_pred_in.reserved_2;
		ser_mdata_pred_in_v(7 downto 4)	:= mdata_pred_in.abs_err_limit_bit_depth;
		
		if (per_err_upd_in = '0') then		-- If "Periodic Error Limit Update" enabled, absolute error limit values subblock is not included!
			ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto 8) := mdata_pred_in.abs_err_limit_val_subblock(mdata_pred_in.abs_err_limit_val_size-1 downto 0);
		end if;
		
		return ser_mdata_pred_in_v;
	end function serial_mdata_pred_abs_err_limit;
	
	-- Serializes record "mdata_pred_err_limit_upd_period_t" to "std_logic_vector" (from Table 5-9)
	pure function serial_mdata_pred_err_limit_upd_period(mdata_pred_in : in mdata_pred_err_limit_upd_period_t) return std_logic_vector is
		variable ser_mdata_pred_in_v : std_logic_vector(mdata_pred_in.total_width-1 downto 0);
	begin
		ser_mdata_pred_in_v(0 downto 0)	:= mdata_pred_in.reserved_1;
		ser_mdata_pred_in_v(1 downto 1)	:= mdata_pred_in.per_upd_flag;
		ser_mdata_pred_in_v(3 downto 2)	:= mdata_pred_in.reserved_2;
		ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto 4) := mdata_pred_in.upd_period_exp;
		
		return ser_mdata_pred_in_v;
	end function serial_mdata_pred_err_limit_upd_period;

	-- Serializes record "mdata_pred_quant_t" to "std_logic_vector" (from Table 5-8)
	pure function serial_mdata_pred_quant(mdata_pred_in : in mdata_pred_quant_t; fidel_ctrl_in : in std_logic_vector; sample_order_in : in std_logic_vector; per_err_upd_in : in std_logic) return std_logic_vector is
		variable ser_mdata_pred_in_v : std_logic_vector(mdata_pred_in.total_width-1 downto 0);
	begin
		if (sample_order_in /= BSQ_C) then
			ser_mdata_pred_in_v(mdata_pred_in.err_limit_upd_period.total_width-1 downto 0)																					 	 := serial_mdata_pred_err_limit_upd_period(mdata_pred_in.err_limit_upd_period);
		end if;
		
		if (fidel_ctrl_in = ABS_ERR_C or fidel_ctrl_in = ABS_REL_C) then
			ser_mdata_pred_in_v(mdata_pred_in.err_limit_upd_period.total_width+mdata_pred_in.absol_err_limit.total_width-1 downto mdata_pred_in.err_limit_upd_period.total_width):= serial_mdata_pred_abs_err_limit(mdata_pred_in.absol_err_limit, per_err_upd_in);
		end if;
		
		if (fidel_ctrl_in = REL_ERR_C or fidel_ctrl_in = ABS_REL_C) then
			ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto mdata_pred_in.err_limit_upd_period.total_width+mdata_pred_in.absol_err_limit.total_width)					 := serial_mdata_pred_rel_err_limit(mdata_pred_in.relat_err_limit, per_err_upd_in);
		end if;

		return ser_mdata_pred_in_v;
	end function serial_mdata_pred_quant;

	-- Serializes record "mdata_pred_weight_tables_t" to "std_logic_vector" (from Table 5-7)
	pure function serial_mdata_pred_weight_tables(mdata_pred_in : in mdata_pred_weight_tables_t; w_init_en_in : in std_logic; w_exp_off_en_in : in std_logic) return std_logic_vector is
		variable ser_mdata_pred_in_v : std_logic_vector(mdata_pred_in.total_width-1 downto 0);
	begin
		if (w_init_en_in = '1') then
			ser_mdata_pred_in_v(mdata_pred_in.w_init_table_size-1 downto 0)								:= mdata_pred_in.w_init_table(mdata_pred_in.w_init_table_size-1 downto 0);
			if (w_exp_off_en_in = '1') then
				ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto mdata_pred_in.w_init_table_size) := mdata_pred_in.w_exp_off_table(mdata_pred_in.w_exp_off_table_size-1 downto 0);
			end if;
		else	-- Makes no sense to check if the 2nd table is not there, in case the 1st table is already not there
			ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto 0)									:= mdata_pred_in.w_exp_off_table(mdata_pred_in.w_exp_off_table_size-1 downto 0);
		end if;

		return ser_mdata_pred_in_v;
	end function serial_mdata_pred_weight_tables;
	
	-- Serializes record "mdata_pred_primary_t" to "std_logic_vector" (from Table 5-6)
	pure function serial_mdata_pred_primary(mdata_pred_in : in mdata_pred_primary_t) return std_logic_vector is
		variable ser_mdata_pred_in_v : std_logic_vector(mdata_pred_in.total_width-1 downto 0);
	begin
		ser_mdata_pred_in_v(0 downto 0)	  := mdata_pred_in.reserved_1;
		ser_mdata_pred_in_v(1 downto 1)	  := mdata_pred_in.smpl_repr_flag;
		ser_mdata_pred_in_v(5 downto 2)	  := mdata_pred_in.num_pred_bands;
		ser_mdata_pred_in_v(6 downto 6)	  := mdata_pred_in.pred_mode;
		ser_mdata_pred_in_v(7 downto 7)	  := mdata_pred_in.w_exp_offset_flag;
		ser_mdata_pred_in_v(9 downto 8)	  := mdata_pred_in.lsum_type;
		ser_mdata_pred_in_v(15 downto 10) := mdata_pred_in.register_size;
		ser_mdata_pred_in_v(19 downto 16) := mdata_pred_in.w_comp_res;
		ser_mdata_pred_in_v(23 downto 20) := mdata_pred_in.w_upd_scal_exp_chng_int;
		ser_mdata_pred_in_v(27 downto 24) := mdata_pred_in.w_upd_scal_exp_init_param;
		ser_mdata_pred_in_v(31 downto 28) := mdata_pred_in.w_upd_scal_exp_final_param;
		ser_mdata_pred_in_v(32 downto 32) := mdata_pred_in.w_exp_off_table_flag;
		ser_mdata_pred_in_v(33 downto 33) := mdata_pred_in.w_init_method;
		ser_mdata_pred_in_v(34 downto 34) := mdata_pred_in.w_init_table_flag;
		ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto 35):= mdata_pred_in.w_init_res;
		
		return ser_mdata_pred_in_v;
	end function serial_mdata_pred_primary;

	-- Serializes record "mdata_pred_t" to "std_logic_vector" (from Table 5-5)
	pure function serial_mdata_pred(mdata_pred_in : in mdata_pred_t; w_init_en_in : in std_logic; w_exp_off_en_in : in std_logic; fidel_ctrl_in : in std_logic_vector; sample_order_in : in std_logic_vector; per_err_upd_in : in std_logic; theta_in : in integer; damp_en_in : in std_logic; off_en_in : in std_logic) return std_logic_vector is
		variable ser_mdata_pred_in_v : std_logic_vector(mdata_pred_in.total_width-1 downto 0);
	begin
		ser_mdata_pred_in_v(mdata_pred_in.primary.total_width-1 downto 0)																												:= serial_mdata_pred_primary(mdata_pred_in.primary);
		
		if (w_init_en_in = '1' or w_exp_off_en_in = '1') then
			ser_mdata_pred_in_v(mdata_pred_in.primary.total_width+mdata_pred_in.weight_tables.total_width-1 downto mdata_pred_in.primary.total_width)									:= serial_mdata_pred_weight_tables(mdata_pred_in.weight_tables, w_init_en_in, w_exp_off_en_in);
		end if;
		
		if (fidel_ctrl_in /= LOSSLESS_C) then
			ser_mdata_pred_in_v(mdata_pred_in.primary.total_width+mdata_pred_in.weight_tables.total_width+mdata_pred_in.quantization.total_width-1 downto mdata_pred_in.primary.total_width+mdata_pred_in.weight_tables.total_width) := serial_mdata_pred_quant(mdata_pred_in.quantization, fidel_ctrl_in, sample_order_in, per_err_upd_in);
		end if;

		if (theta_in > 0) then
			ser_mdata_pred_in_v(mdata_pred_in.total_width-1 downto mdata_pred_in.primary.total_width+mdata_pred_in.weight_tables.total_width+mdata_pred_in.quantization.total_width)	:= serial_mdata_pred_smpl_repr(mdata_pred_in.smpl_repr, damp_en_in, off_en_in);
		end if;
		
		return ser_mdata_pred_in_v;
	end function serial_mdata_pred;

	-------------------------------------------------------------------------------------------------------
	-- IMAGE METADATA
	-------------------------------------------------------------------------------------------------------
	
	-- Serializes record "mdata_img_supl_info_t" to "std_logic_vector" (from Table 5-4)
	pure function serial_mdata_img_supl_info(mdata_img_in : in mdata_img_supl_info_t) return std_logic_vector is
		variable ser_mdata_img_in_v : std_logic_vector(mdata_img_in.total_width-1 downto 0);
	begin
		ser_mdata_img_in_v(1 downto 0)	 := mdata_img_in.table_type;
		ser_mdata_img_in_v(3 downto 2)	 := mdata_img_in.reserved_1;
		ser_mdata_img_in_v(7 downto 4)	 := mdata_img_in.table_purpose;
		ser_mdata_img_in_v(8 downto 8)	 := mdata_img_in.reserved_2;
		ser_mdata_img_in_v(10 downto 9)	 := mdata_img_in.table_structure;
		ser_mdata_img_in_v(11 downto 11) := mdata_img_in.reserved_3;
		ser_mdata_img_in_v(15 downto 12) := mdata_img_in.supl_user_def_data;
		-- Only meaningful data from "table_data_subblock" is extracted (as it has a fixed longer width)
		ser_mdata_img_in_v(mdata_img_in.total_width-1 downto 16) := mdata_img_in.table_data_subblock(mdata_img_in.table_data_size-1 downto 0);
		
		return ser_mdata_img_in_v;
	end function serial_mdata_img_supl_info;

	-- Serializes array of record "mdata_img_supl_info_t" to "std_logic_vector" (from Table 5-4)
	pure function serial_mdata_img_supl_info_arr(mdata_img_arr_in : in mdata_img_supl_info_arr_t) return std_logic_vector is
		variable ser_mdata_img_in_v : std_logic_vector(mdata_img_arr_in'length*mdata_img_arr_in(0).total_width-1 downto 0);
	begin		
		for i in 0 to (mdata_img_arr_in'length-1) loop
			ser_mdata_img_in_v((i+1)*mdata_img_arr_in(i).total_width-1 downto i*mdata_img_arr_in(i).total_width) := serial_mdata_img_supl_info(mdata_img_arr_in(i));
		end loop;
		
		return ser_mdata_img_in_v;
	end function serial_mdata_img_supl_info_arr;

	-- Serializes record "mdata_img_essential_t" to "std_logic_vector" (from Table 5-3)
	pure function serial_mdata_img_essential(mdata_img_in : in mdata_img_essential_t) return std_logic_vector is
		variable ser_mdata_img_in_v : std_logic_vector(mdata_img_in.total_width-1 downto 0);
	begin
		ser_mdata_img_in_v(7 downto 0)	 := mdata_img_in.udef_data;
		ser_mdata_img_in_v(23 downto 8)	 := mdata_img_in.x_size;
		ser_mdata_img_in_v(39 downto 24) := mdata_img_in.y_size;
		ser_mdata_img_in_v(55 downto 40) := mdata_img_in.z_size;
		ser_mdata_img_in_v(56 downto 56) := mdata_img_in.smpl_type;
		ser_mdata_img_in_v(57 downto 57) := mdata_img_in.reserved_1;
		ser_mdata_img_in_v(58 downto 58) := mdata_img_in.larg_dyn_rng_flag;
		ser_mdata_img_in_v(62 downto 59) := mdata_img_in.dyn_range;
		ser_mdata_img_in_v(63 downto 63) := mdata_img_in.smpl_enc_order;
		ser_mdata_img_in_v(79 downto 64) := mdata_img_in.sub_frm_intlv_depth;
		ser_mdata_img_in_v(81 downto 80) := mdata_img_in.reserved_2;
		ser_mdata_img_in_v(84 downto 82) := mdata_img_in.out_word_size;
		ser_mdata_img_in_v(86 downto 85) := mdata_img_in.entropy_coder_type;
		ser_mdata_img_in_v(87 downto 87) := mdata_img_in.reserved_3;
		ser_mdata_img_in_v(89 downto 88) := mdata_img_in.quant_fidel_ctrl_mth;
		ser_mdata_img_in_v(91 downto 90) := mdata_img_in.reserved_4;
		ser_mdata_img_in_v(mdata_img_in.total_width-1 downto 92):= mdata_img_in.supl_info_table_cnt;

		return ser_mdata_img_in_v;
	end function serial_mdata_img_essential;

	-- Serializes record "mdata_img_t" to "std_logic_vector" (from Table 5-2)
	pure function serial_mdata_img(mdata_img_in : in mdata_img_t) return std_logic_vector is
		variable ser_mdata_img_in_v : std_logic_vector(mdata_img_in.total_width-1 downto 0);
	begin
		ser_mdata_img_in_v(mdata_img_in.essential.total_width-1 downto 0)							 := serial_mdata_img_essential(mdata_img_in.essential);
		if (TAU_C > 0) then		-- Supplementary Info Tables are added only if requested
			ser_mdata_img_in_v(mdata_img_in.total_width-1 downto mdata_img_in.essential.total_width) := serial_mdata_img_supl_info_arr(mdata_img_in.supl_info_arr);
		end if;
		
		return ser_mdata_img_in_v;
	end function serial_mdata_img;
	
	-------------------------------------------------------------------------------------------------------
	-- TOP ENCODER HEADER
	-------------------------------------------------------------------------------------------------------
	
	-- Serializes record "enc_header_t" to "std_logic_vector" (from Table 5-1) --> TEMPORARY DISABLED!
	-- pure function serial_enc_header(enc_header_in : in enc_header_t) return std_logic_vector is
		-- variable ser_enc_header_in_v : std_logic_vector(enc_header_in.total_width-1 downto 0);
	-- begin
		-- ser_enc_header_in_v(enc_header_in.mdata_img.total_width-1 downto 0)																			:= serial_mdata_img(enc_header_in.mdata_img);
		-- ser_enc_header_in_v(enc_header_in.mdata_img.total_width+enc_header_in.mdata_pred.total_width-1 downto enc_header_in.mdata_img.total_width)	:= serial_mdata_pred(enc_header_in.mdata_pred);
		-- ser_enc_header_in_v(enc_header_in.total_width-1 downto enc_header_in.mdata_img.total_width+enc_header_in.mdata_pred.total_width)				:= serial_mdata_enc(enc_header_in.mdata_enc);
		-- return ser_enc_header_in_v;
	-- end function serial_enc_header;
	
	-------------------------------------------------------------------------------------------------------
	-- STRING FUNCTIONS
	-------------------------------------------------------------------------------------------------------
	
	-- Reads an integer (from 0x0 to 0xF) and returns its character
	pure function int_to_char(integer_in : in integer) return character is
		variable character_v : character := ' ';
	begin
		case (integer_in) is
			when 16#0#	=> character_v := '0';
			when 16#1#	=> character_v := '1';
			when 16#2#	=> character_v := '2';
			when 16#3#	=> character_v := '3';
			when 16#4#	=> character_v := '4';
			when 16#5#	=> character_v := '5';
			when 16#6#	=> character_v := '6';
			when 16#7#	=> character_v := '7';
			when 16#8#	=> character_v := '8';
			when 16#9#	=> character_v := '9';
			when 16#A#	=> character_v := 'A';
			when 16#B#	=> character_v := 'B';
			when 16#C#	=> character_v := 'C';
			when 16#D#	=> character_v := 'D';
			when 16#E#	=> character_v := 'E';
			when others	=> character_v := 'F';
		end case;
	
		return character_v;
	end function int_to_char;
	
	-- Reads a character ('0' to 'F') and returns its integer digit
	pure function char_to_int(char_in : in character) return integer is
		variable integer_v : integer := 0;
	begin
		case (char_in) is
			when '0'	=> integer_v := 16#0#;
			when '1'	=> integer_v := 16#1#;
			when '2'	=> integer_v := 16#2#;
			when '3'	=> integer_v := 16#3#;
			when '4'	=> integer_v := 16#4#;
			when '5'	=> integer_v := 16#5#;
			when '6'	=> integer_v := 16#6#;
			when '7'	=> integer_v := 16#7#;
			when '8'	=> integer_v := 16#8#;
			when '9'	=> integer_v := 16#9#;
			when 'A'	=> integer_v := 16#10#;
			when 'B'	=> integer_v := 16#11#;
			when 'C'	=> integer_v := 16#12#;
			when 'D'	=> integer_v := 16#13#;
			when 'E'	=> integer_v := 16#14#;
			when others	=> integer_v := 16#15#;
		end case;
	
		return integer_v;
	end function char_to_int;
	
	-- Reads a string (character by character) and returns the sum of its integer digits
	pure function string_to_int(string_in : in string) return integer is
		alias	 string_a	: string(1 to string_in'length) is string_in;
		variable multipl_v	: integer := 1;
		variable integer_v	: integer := 0;
	begin
		for i in string_a'length downto 1 loop
			integer_v := integer_v + multipl_v * char_to_int(string_a(i));
			multipl_v := multipl_v * 10;		-- Every new digit moves to the next position (unit, tenth, hundredth...)
		end loop;
	
		return integer_v;
	end function string_to_int;
	
	-- Returns the meaningful number of bits from a string (to remove all spaces)
	pure function clean_string(string_in : in string) return integer is
		variable position_v : integer := 0;
	begin
		for i in 1 to string_in'length loop
			if (string_in(i) = ' ') then
				position_v := i;				
				exit;	-- When space found, no need to keep checking (as the rest can only be spaces as well)
			end if;
		end loop;
		
		if (position_v <= 0) then		-- If no space found, returns the complete string length.
			return string_in'length;
		elsif (position_v = 1) then		-- If space found at the beginning, it means string is empty, so length 0 is sent out.
			return 0;
		else							-- If space found later, only the length with meaningful data is returned.
			return position_v-1;
		end if;
	end function clean_string;
	
	-- Gives the proper format to the incoming string (NOTE: String index must start with 1!)
	-- This function is prepared to handle incoming strings ONLY with the following structure:
	--		string value:	"0^{xxx}xx..."
	--		char. position:	"123456789..."
	pure function format_string(string_in : in string; max_length_in : in integer) return string is
		variable string_out_v : string(1 to max_length_in) := (others => ' ');	-- Biggest possible case!
		variable num_0s_v	  : integer := string_to_int(string_in(4 to 6));
		variable str_length_v : integer := string_in'length - 7;
		variable count_char_v : integer range 8 to string_in'length+1 := 8;		-- Position where "non-left-0" data starts
	begin
		-- Logic is applied ONLY if the string has the expected architecture: "0^{" & "}"
		if (string_in(1 to 3) = "0^{" and string_in(7) = '}') then	-- '' when checking one position (character), just as std_logic
			for i in 1 to string_out_v'length loop
				-- The number of 0s at the beginning (if any) are applied at the output
				if (i <= num_0s_v) then
					string_out_v(i) := '0';
				else	-- i > num_0s_v
					if (string_in(count_char_v) /= ' ') then
						string_out_v(i) := string_in(count_char_v);
						count_char_v	:= count_char_v + 1;
					else
						exit;		-- No need to keep looking after passing the 2 conditions above
					end if;
				end if;
			end loop;
			
			return string_out_v;	-- Formatted data is outputted
		else
			return string_out_v;	-- Unexpected case (wrong string architecture) -> Empty string!
		end if;
	end function format_string;
	
	-- Shifts a string to either left (direct_in = '1') or right (direct_in = '0') a configurable number of steps, adding spaces as new characters
	pure function shift_string(string_in : in string; num_steps_in : in integer; direct_in : in std_logic) return string is
		variable string_out_v : string(1 to string_in'length) := string_in;
	begin
		if (direct_in = '1') then	-- Shift to the left
			for i in string_in'length downto 1 loop
				if (i > num_steps_in) then
					string_out_v(i-num_steps_in) := string_out_v(i);
				else
					string_out_v(string_in'length-i+1) := ' ';
				end if;
			end loop;
		else	-- direct_in = '0'	-- Shift to the right
			for i in string_in'length downto 1 loop
				if (i > num_steps_in) then
					string_out_v(i) := string_out_v(i-num_steps_in);
				else
					string_out_v(i) := ' ';
				end if;
			end loop;
		end if;
		
		return string_out_v;
	end function shift_string;
	
	-------------------------------------------------------------------------------------------------------
	-- OTHER FUNCTIONS
	-------------------------------------------------------------------------------------------------------
	
	-- Returns the position from the Low-Entropy codes table, where the incoming threshold belongs to
	pure function find_pos_low_entr_table(threshold_in : in integer) return integer is
		variable position_v : integer := -1;
		variable max_loop_v : integer := LOW_ENTR_CODES_C.threshold'length-2;
	begin -- Threshold could be bigger than the largest value in the array, so checking process should start from array position 1.
		-- NOTE: Anyway, above position 0 means that Hybrid Encoder works with "high-entropy", not "low-entropy" (as this table is used for).
		if (threshold_in > LOW_ENTR_CODES_C.threshold(1)) then
			position_v := 0;
		elsif (threshold_in <= LOW_ENTR_CODES_C.threshold(15)) then
			position_v := 15;
		else
			for i in 1 to max_loop_v loop	-- Edges (0 and 15) are already checked above
				if ((threshold_in <= LOW_ENTR_CODES_C.threshold(i)) and (threshold_in > LOW_ENTR_CODES_C.threshold(i+1))) then
					position_v := i;
					exit;	-- If a position was already found, no need to keep looking
				end if;
			end loop;
		end if;
		
		return position_v;
	end function find_pos_low_entr_table;
	
	-- Pads a std_logic_vector with defined value until reaching defined number of bits
	pure function padding_bits(slv_in : in std_logic_vector; pad_bits_in : in integer; pad_value_in : in integer) return std_logic_vector is
		variable slv_out_v : std_logic_vector(slv_in'length + pad_bits_in - 1 downto 0);
	begin
		slv_out_v(slv_in'length-1 downto 0) := slv_in(slv_in'length-1 downto 0);
		if (pad_bits_in /= 0) then
			slv_out_v(slv_in'length+pad_bits_in-1 downto slv_in'length) := std_logic_vector(to_unsigned(pad_value_in, pad_bits_in));
		end if;
		
		return slv_out_v;
	end function padding_bits;
	
	-- Find the position (if any) where the current prefix matches the "input codeword" in "hybrid code tables"
	pure function match_code_prefix(string_arr_in : in array_string_t; table_size_in : in integer; curr_prefix_in : in string) return integer is
		variable format_string_v : string(1 to string_arr_in'length) := (others => ' ');
        variable code_pos_v		 : integer := -1;	-- "-1" means there has been NO string match
	begin
		for i in 0 to string_arr_in'length-1 loop
			if (i < table_size_in) then
				format_string_v := format_string(string_arr_in(i), format_string_v'length);		-- Incoming string is given a format
				-- Received prefix is compared with the prefixes from 'code tables' (including spaces, to make it synthetisable)
				if (format_string_v = curr_prefix_in) then
					code_pos_v := i;
					exit;	-- When match, no need to keep looking
				end if;
			else
				exit;		-- No need to keep looking after passing the condition above
			end if;
		end loop;
		
		return code_pos_v;
	end function match_code_prefix;

	-- Find the position (and there MUST be one!) where the current prefix (without lsb) matches the "active prefix" in "hybrid flush tables"
	pure function match_flush_prefix(string_arr_in : in array_string_t; table_size_in : in integer; curr_prefix_in : in string) return integer is
		constant length_prefix_v  : integer := clean_string(curr_prefix_in);
		-- "+1" because the "code table" array is 1 bit longer than "flush table" array
		variable curr_prefix_in_v : string(1 to string_arr_in'length+1) := curr_prefix_in;
		variable format_string_v  : string(1 to string_arr_in'length+1) := (others => ' ');
		variable flush_pos_v	  : integer := -1;		-- "-1" means there has been NO string match
	begin
		-- When the incoming prefix has >1 character, all of them (except the lsb) are used to find the right position inside hybrid flush table
		if (length_prefix_v > 1) then
			-- As here we are looking for the prefix of the "current prefix" signal, its 'Least Significant Bit' must be deleted prior to the comparison
			curr_prefix_in_v(length_prefix_v) := ' ';
			
			for i in 0 to string_arr_in'length-1 loop		-- Here I must ignore the least significant bit
				if (i < table_size_in) then
					format_string_v := format_string(string_arr_in(i), format_string_v'length);		-- Incoming string is given a format
					-- Received prefix is compared with the prefixes from 'code tables' (including spaces, to make it synthetisable)
					if (format_string_v = curr_prefix_in_v) then
						flush_pos_v := i;
						exit;	-- When match, no need to keep looking
					end if;
				else
					exit;		-- No need to keep looking after passing the condition above
				end if;
			end loop;
		else	-- If incoming prefix has just 1 character, then answer is the "null" character (pos.0 in hybrid flush tables)
			flush_pos_v := 0;
		end if;
		
		return flush_pos_v;
	end function match_flush_prefix;
	
end package body utils_encoder;