# ***********************************************************************************
# ************************* VUNIT AND LIBRARY INITIALIZATION ************************
# ***********************************************************************************

# Load required libraries from VUnit
from os.path import join, dirname
# The public interface of VUnit
from vunit import VUnit
# Computes the cartesian product of input iterables
from itertools import product
# Load required functions for testcase files load
from os import listdir

# Returns the directory name where the present file (run.py) is located
root = dirname(__file__)

# Create VUnit instance by parsing command line arguments
ui = VUnit.from_argv()

# Add random numbers generation package
ui.add_random()
# Add verification component library
ui.add_verification_components()
# Add communication package
ui.add_com()

# Xilinx XPM library, already compiled for Windows (for simulation with ModelSim)
ui.add_external_library("xpm", join(root, "../../_external_libs/xpm"))

# Create library 'vunit_lib'
vunit_lib = ui.library("vunit_lib")

# Add all package files
vunit_lib.add_source_files(join(root, "../../Image/_packages/*.vhd"))
vunit_lib.add_source_files(join(root, "../../Predictor/_packages/*.vhd"))
vunit_lib.add_source_files(join(root, "../../Encoder/_packages/*.vhd"))

# Add all sources files from Image IP
vunit_lib.add_source_files(join(root, "../../Image/image_coord_control.vhd"))

# Add necessary source files from Predictor IP
vunit_lib.add_source_files(join(root, "../../Predictor/prediction/shift_register.vhd"))

# Add all sources files from Top Encoder IP
vunit_lib.add_source_files(join(root, "../header/*.vhd"))
vunit_lib.add_source_files(join(root, "../body/hybrid_coder/*.vhd"))
vunit_lib.add_source_files(join(root, "../body/sample_adaptive_coder/*.vhd"))
vunit_lib.add_source_files(join(root, "../body/top_enc_body.vhd"))
vunit_lib.add_source_files(join(root, "../packer/*.vhd"))
vunit_lib.add_source_files(join(root, "../top_encoder.vhd"))

# Add testbench file from Top Encoder IP
vunit_lib.add_source_files(join(root, "tb_top_encoder.vhd"))

# Necessary file (and explicit association for ModelSim) to let the XPM compiled library work
# NOTE: These lines must be added at the end...
vunit_lib.add_source_files(join(root, "../../_external_libs/glbl.v"))
ui.set_sim_option("modelsim.vsim_flags", ["vunit_lib.glbl"])

# ***********************************************************************************
# ************************************ FUNCTIONS ************************************
# ***********************************************************************************

# To encode the parameters, the script must contain the encode function
def encode(tb_cfg):
    return ",".join(["%s:%s" % (key, str(tb_cfg[key])) for key in tb_cfg])

# A list of parameters are defined, then saved and finally encoded to be in the testbench
def gen_top_encod_tests(obj, smpl_type, smpl_order, header_en, predict_mode, lsum_type, w_init_type, per_err_lim_upd, fidel_type, abs_err_band, rel_err_band, encoder_type, supl_tbl_type, supl_tbl_struct, endianness):
    for smpl_type, smpl_order, header_en, predict_mode, lsum_type, w_init_type, per_err_lim_upd, fidel_type, abs_err_band, rel_err_band, encoder_type, supl_tbl_type, supl_tbl_struct, endianness in product(smpl_type, smpl_order, header_en, predict_mode, lsum_type, w_init_type, per_err_lim_upd, fidel_type, abs_err_band, rel_err_band, encoder_type, supl_tbl_type, supl_tbl_struct, endianness):
        tb_cfg = dict(
            SMPL_TYPE_PY=smpl_type,
            SMPL_ORDER_PY=smpl_order,
            HEADER_EN_PY=header_en,
            PREDICT_MODE_PY=predict_mode,
            LSUM_TYPE_PY=lsum_type,
            W_INIT_TYPE_PY=w_init_type,
            PER_ERR_LIM_UPD_PY=per_err_lim_upd,
            FIDEL_CTRL_TYPE_PY=fidel_type,
            ABS_ERR_BAND_TYPE_PY=abs_err_band,
            REL_ERR_BAND_TYPE_PY=rel_err_band,
            ENCODER_TYPE_PY=encoder_type,
            SUPL_TABLE_TYPE_PY=supl_tbl_type,
            SUPL_TABLE_STRUCT_PY=supl_tbl_struct,
            ENDIANNESS_PY=endianness
        )
        config_name = encode(tb_cfg)
        obj.add_config(name=config_name, generics=dict(encoded_tb_cfg=encode(tb_cfg)))

# ***********************************************************************************
# ****************************** GENERATE TESTBENCHES *******************************
# ***********************************************************************************

# Everytime a new testbench is here requested (it can be the same with different parameters),
# all test cases in the VHDL testbench file will be executed again.
tb_top_encoder = vunit_lib.test_bench("tb_top_encoder")
for test in tb_top_encoder.get_tests():
    # gen_top_encod_tests(test, ['0', '1'], [0, 1, 2], ['0', '1'], ['0', '1'], [0, 1, 2, 3], ['0', '1'], ['0', '1'], [0, 1, 2, 3], ['0', '1'], ['0', '1'], [0, 1, 2], [0, 1, 2], [0, 1, 2, 3], ['0', '1'])
    gen_top_encod_tests(test, ['0'], [1, 2], ['0', '1'], ['0', '1'], [0], ['1'], ['0', '1'], [3], ['1'], ['1'], [0, 1], [1], [1], ['0'])

# ***********************************************************************************
# ********************************** MAIN FUNCTION **********************************
# ***********************************************************************************

ui.main()