--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		18/04/2021
--------------------------------------------------------------------------------
-- IP name:		hybrid_low_entr_code
--
-- Description: The low-entropy codeword (up to 16 different ones) to encode
--				with the Hybrid Entropy Coder.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.hybrid_code_tables.all;
use work.utils_encoder.all;

entity hybrid_low_entr_code is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- Maximum number of bits (size) of every encoded data sample
		MAX_NUM_BITS_G		: integer;
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic
	);
	port (
		clock_i			  : in  std_logic;
		reset_i			  : in  std_logic;
		
		enable_i		  : in  std_logic;
		enable_o		  : out std_logic;
		img_coord_i		  : in  img_coord_t;
		img_coord_o		  : out img_coord_t;
		
		err_lim_upd_en_i  : in  std_logic;
		err_lim_val_i	  : in  std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i	  : in  std_logic_vector(D_C-1 downto 0);
		
		high_res_accu_i	  : in  std_logic_vector(D_C*2-1 downto 0);	-- "Σ~z(t)" (high resolution accumulator)
		counter_i		  : in  std_logic_vector(D_C-1 downto 0);	-- "Γ(t)"   (counter)
		reconst_bit_i	  : in  std_logic;							-- Lets decoder reconstruct "Σ~z(t)" values
		
		curr_prefix_arr_o : out array_string_t(0 to NUM_HYBRID_TABLES_G-1)(1 to LONGEST_STRING_SIZE_G);	-- Current prefix for every Hybrid table
		enc_num_bits_o	  : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
		rvd_codeword_o	  : out std_logic_vector(Umax_C+D_C-1 downto 0)	-- "R'k(j)" (reversed codeword)
	);
end entity hybrid_low_entr_code;

architecture behavioural of hybrid_low_entr_code is
	-- Constants declaration
	constant MAX_NUM_BITS_C	  : std_logic_vector(MAX_NUM_BITS_G-1 downto 0)	:= std_logic_vector(to_unsigned(Umax_C+D_C, MAX_NUM_BITS_G));
	
	-- Signals declaration
	signal enable1_s		  : std_logic	:= '0';
	signal enable2_s		  : std_logic	:= '0';
	signal img_coord1_s		  : img_coord_t	:= reset_img_coord(SMPL_ORDER_G);
	signal img_coord2_s		  : img_coord_t	:= reset_img_coord(SMPL_ORDER_G);
	
	signal index_table_s	  : integer range 0 to NUM_HYBRID_TABLES_G-1	:= 0;
	signal input_sym_s		  : character	:= ' ';		-- "character" is a "string" of 1 position
	signal code_prefix_s	  : integer		:= 0;
	
	signal enc_bits0_s		  : integer range 0 to Umax_C+D_C				:= 0;
	signal rvd_code0_s		  : std_logic_vector(Umax_C+D_C-1 downto 0)		:= (others => '0');
	
	signal err_lim_cnt_s	  : integer range 0 to 2*NZ_C					:= 0;
	signal err_lim_upd_en_s   : std_logic	:= '0';
	signal err_lim_val_s	  : std_logic_vector(D_C-1 downto 0)			:= (others => '0');
	signal data_mp_quan_s	  : std_logic_vector(D_C-1 downto 0)			:= (others => '0');
	signal high_res_accu_s	  : std_logic_vector(D_C*2-1 downto 0)			:= (others => '0');
	signal reconst_bit_s	  : std_logic	:= '0';
	
	signal curr_prefix_arr1_s : array_string_t(0 to NUM_HYBRID_TABLES_G-1)(1 to LONGEST_STRING_SIZE_G) := (others => (others => ' '));
	signal curr_prefix_arr2_s : array_string_t(0 to NUM_HYBRID_TABLES_G-1)(1 to LONGEST_STRING_SIZE_G) := (others => (others => ' '));
	signal enc_num_bits_s	  : std_logic_vector(MAX_NUM_BITS_G-1 downto 0)	:= (others => '0');
	signal rvd_codeword_s	  : std_logic_vector(Umax_C+D_C-1 downto 0)		:= (others => '0');

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_hyb_low_entr_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable1_s		   <= '0';
				enable2_s		   <= '0';
				img_coord1_s	   <= reset_img_coord(SMPL_ORDER_G);
				img_coord2_s	   <= reset_img_coord(SMPL_ORDER_G);
				reconst_bit_s	   <= '0';
				err_lim_upd_en_s   <= '0';
				err_lim_val_s	   <= (others => '0');
				data_mp_quan_s	   <= (others => '0');
				high_res_accu_s	   <= (others => '0');
				curr_prefix_arr2_s <= (others => (others => ' '));
			else
				enable1_s		   <= enable_i;
				enable2_s		   <= enable1_s;
				img_coord1_s	   <= img_coord_i;
				img_coord2_s	   <= img_coord1_s;
				reconst_bit_s	   <= reconst_bit_i;
				err_lim_upd_en_s   <= err_lim_upd_en_i;
				err_lim_val_s	   <= err_lim_val_i;
				data_mp_quan_s	   <= data_mp_quan_i;
				high_res_accu_s	   <= high_res_accu_i;
				curr_prefix_arr2_s <= curr_prefix_arr1_s;
			end if;
		end if;
	end process p_hyb_low_entr_delay;
	
	-- Calculation of "code table prefix"
	p_calc_code_tbl_prefix : process(clock_i) is
		variable threshold_v		: integer	:= 0;
		variable index_table_v		: integer range 0 to NUM_HYBRID_TABLES_G-1 := 0;
		variable input_sym_v		: character := ' ';		-- "character" is a "string" of 1 position
		-- Array as big as existing code tables, to ensure consecutive samples from different tables can be managed
		-- NOTE: "string" signals work as "std_logic_vector" too, but always starting from position '1', and better with 'to' on width definition
		variable curr_prefix_arr_v	: array_string_t(0 to NUM_HYBRID_TABLES_G-1)(1 to LONGEST_STRING_SIZE_G) := (others => (others => ' '));
		variable curr_pos_arr_v		: array_integer_t(0 to NUM_HYBRID_TABLES_G-1) := (others => 1);
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				threshold_v		   := 0;
				index_table_v	   := 0;
				index_table_s	   <= 0;
				input_sym_v		   := ' ';
				input_sym_s		   <= ' ';
				curr_prefix_arr_v  := (others => (others => ' '));
				curr_prefix_arr1_s <= (others => (others => ' '));
				curr_pos_arr_v	   := (others => 1);
				code_prefix_s	   <= 0;
			else
				if (enable_i = '1' and err_lim_upd_en_i = '0') then
					if (img_coord_i.t > 0) then
						-- If there is a prefix match, current prefix and its counter (from its specific 'i'th table) are restarted
						-- NOTE: These lines are added at the beginning, so that they can been seen on simulation)
						if (code_prefix_s /= -1) then
							curr_prefix_arr_v(index_table_v) := (others => ' ');
							curr_pos_arr_v(index_table_v)	 := 1;
						end if;
						
						-- New threshold calculated, to know which table (out of 16) is to be used (no need to have a number longer than 30 bits, even if computed so...)
						threshold_v	  := to_integer(resize((resize(unsigned(high_res_accu_i), D_C*3) * to_unsigned(2**14, D_C*3)) / resize(unsigned(counter_i), D_C*3), 30));
						index_table_v := find_pos_low_entr_table(threshold_v);
						index_table_s <= index_table_v;
						
						-- A new input symbol (in hexadecimal format) generated for every incoming sample
						if (unsigned(data_mp_quan_i) <= to_unsigned(LOW_ENTR_CODES_C.in_sym_limit(index_table_v), D_C)) then
							input_sym_v := int_to_char(to_integer(unsigned(data_mp_quan_i)));	-- With the computed symbol, there MIGHT be a match for "code_prefix_s"
						else
							input_sym_v := 'X';													-- With the escape symbol ('X'), there will ALWAYS be a match for "code_prefix_s"
						end if;
						input_sym_s <= input_sym_v;
						
						-- (5.4.3.3.5.3.4) Current prefix for the 'i'th low-entropy code (so one per table) is updated by appending the input symbol to it -> Oldest value on leftmost side
						curr_prefix_arr_v(index_table_v)(curr_pos_arr_v(index_table_v)) := input_sym_v;
						curr_pos_arr_v(index_table_v) := curr_pos_arr_v(index_table_v) + 1;
						-- Try to find out if the current prefix matches an "input codeword" in the corresponding "hybrid code table"
						code_prefix_s <= match_code_prefix(HYBRID_CODE_TABLES_ARR_C(index_table_v).input_code, HYBRID_CODE_TABLES_ARR_C(index_table_v).table_size, curr_prefix_arr_v(index_table_v));
						
						-- All current prefix from all tables are sent out for the image tail generation
						curr_prefix_arr1_s <= curr_prefix_arr_v;
					end if;
				end if;
			end if;
		end if;
	end process p_calc_code_tbl_prefix;
	
	-- (5.4.3.3.5.3.3) Calculation of "R'0(dz(t)-Li-1)" (Reversed Length-Limited GPO2 Codeword of Mapped Quantizer Index)
	p_calc_low_code0 : process(reset_i, input_sym_s) is
	begin
		if (reset_i = '1') then
			enc_bits0_s <= 0;
			rvd_code0_s <= (others => '0');
		else
			if (enable_i = '1' and err_lim_upd_en_i = '0') then
				-- Residual value to send out is calculated
				if (input_sym_s = 'X') then
					-- As kz(t) is fixed to 0 here, the R'0 will always be generated by the D-bit representation of j, followed by Umax '0's.
					enc_bits0_s <= Umax_C + D_C;
					rvd_code0_s <= std_logic_vector(resize(resize(unsigned(data_mp_quan_i) - to_unsigned(LOW_ENTR_CODES_C.in_sym_limit(index_table_s), D_C) - to_unsigned(1, D_C), D_C) & (Umax_C-1 downto 0 => '0'), Umax_C+D_C));
				end if;
			end if;
		end if;
	end process p_calc_low_code0;

	-- Calculation of "R'k(j)" (Reversed Length-Limited GPO2 Codeword of Mapped Quantizer Index)
	p_calc_low_codeword : process(clock_i) is
		variable output_code_v		: integer	:= 0;
		variable output_size_v		: integer	:= 0;
		variable err_lim_type_v		: std_logic	:= '0';
		variable output_code_slv_v	: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
		variable enc_num_bits_v		: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
		variable rvd_codeword_v		: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				output_code_v	  := 0;
				output_size_v	  := 0;
				err_lim_type_v	  := '0';
				output_code_slv_v := (others => '0');
				enc_num_bits_v	  := (others => '0');
				rvd_codeword_v	  := (others => '0');
				err_lim_cnt_s	  <= 0;
				enc_num_bits_s 	  <= (others => '0');
				rvd_codeword_s	  <= (others => '0');
			else
				-- To be sure the packer does not produce problems after this IP, output data is restarted when fully processed...
				if (img_coord2_s = end_img_coord) then
					enc_num_bits_v := (others => '0');
					rvd_codeword_v := (others => '0');
				end if;
				
				if (enable1_s = '1') then
					enc_num_bits_v := (others => '0');
					rvd_codeword_v := (others => '0');
					
					if (err_lim_upd_en_s = '0') then		-- Mapped quantizer indexes ("?z(t)") are encoded here!
						err_lim_cnt_s						<= 0;
						if (img_coord1_s.t = 0) then		-- Sample for t=0 must not be encoded!
							enc_num_bits_v					:= std_logic_vector(to_unsigned(D_C, MAX_NUM_BITS_G));
							rvd_codeword_v					:= data_mp_quan_s & (Umax_C-1 downto 0 => '0');
						else	-- img_coord1_s.t > 0
							-- If the active prefix matches the input codeword, the output codeword is outputted too. Otherwise nothing is done
							if (code_prefix_s /= -1) then	-- Hybrid Code Tables can NOT be read if no "code_prefix_s" found
								-- Extracting values from hybrid code table into dedicated signals, simply for the sake of readability
								output_code_v				:= HYBRID_CODE_TABLES_ARR_C(index_table_s).output_code(code_prefix_s);
								output_size_v				:= HYBRID_CODE_TABLES_ARR_C(index_table_s).output_size(code_prefix_s);
								
								-- Intermediate signal to make the code synthesizable
								output_code_slv_v			:= std_logic_vector(to_unsigned(output_code_v, Umax_C+D_C));
								
								-- With the escape symbol ('X'), there will ALWAYS be a match for the "code_prefix_s"
								if (input_sym_s = 'X') then
									-- 5.4.3.3.5.1.2: When code selection statistics rescaled, LSb from "high_res_accu(t-1) -> here high_res_accu_s" is encoded in bitstream before anything else
									if (reconst_bit_s = '1') then
										enc_num_bits_v		:= std_logic_vector(to_unsigned(enc_bits0_s + output_size_v + 1, MAX_NUM_BITS_G));
										rvd_codeword_v		:= std_logic_vector(resize(unsigned(rvd_code0_s(Umax_C+D_C-1 downto Umax_C)) & unsigned(output_code_slv_v(output_size_v-1 downto 0)) & high_res_accu_s(0), Umax_C+D_C));
									else	-- reconst_bit_s = '0'
										enc_num_bits_v		:= std_logic_vector(to_unsigned(enc_bits0_s + output_size_v, MAX_NUM_BITS_G));
										rvd_codeword_v		:= std_logic_vector(resize(unsigned(rvd_code0_s(Umax_C+D_C-1 downto Umax_C)) & unsigned(output_code_slv_v(output_size_v-1 downto 0)), Umax_C+D_C));
									end if;
								else
									-- 5.4.3.3.5.1.2: When code selection statistics rescaled, LSb from "high_res_accu(t-1) -> here high_res_accu_s" is encoded in bitstream before anything else
									if (reconst_bit_s = '1') then
										enc_num_bits_v		:= std_logic_vector(to_unsigned(output_size_v + 1, MAX_NUM_BITS_G));
										rvd_codeword_v 		:= std_logic_vector(resize(unsigned(output_code_slv_v(output_size_v-1 downto 0)) & high_res_accu_s(0), Umax_C+D_C));
									else	-- reconst_bit_s = '0'
										enc_num_bits_v		:= std_logic_vector(to_unsigned(output_size_v, MAX_NUM_BITS_G));
										rvd_codeword_v 		:= std_logic_vector(resize(unsigned(output_code_slv_v(output_size_v-1 downto 0)), Umax_C+D_C));
									end if;
								end if;
							end if;
						end if;
					
					else	-- err_lim_upd_en_s = '1'		-- Error limit values are encoded here!
						if (FIDEL_CTRL_TYPE_G = ABS_ERR_C) then
							enc_num_bits_v					:= std_logic_vector(to_unsigned(DA_C, MAX_NUM_BITS_G));
							rvd_codeword_v(DA_C-1 downto 0)	:= err_lim_val_s(DA_C-1 downto 0);
						elsif (FIDEL_CTRL_TYPE_G = REL_ERR_C) then
							enc_num_bits_v					:= std_logic_vector(to_unsigned(DR_C, MAX_NUM_BITS_G));
							rvd_codeword_v(DR_C-1 downto 0)	:= err_lim_val_s(DR_C-1 downto 0);
						elsif (FIDEL_CTRL_TYPE_G = ABS_REL_C) then
							if ((ABS_ERR_BAND_TYPE_G = '1' and err_lim_cnt_s = 0) or (ABS_ERR_BAND_TYPE_G = '0' and err_lim_cnt_s < NZ_C)) then
								err_lim_type_v				:= '1';			-- Absolute error limit values (always the first one!)
							else
								err_lim_type_v				:= '0';			-- Relative error limit values (always the second one!)
							end if;
							err_lim_cnt_s					<= err_lim_cnt_s + 1;
							
							if (err_lim_type_v = '1') then	-- Time to output "absolute error limit values"
								enc_num_bits_v				:= std_logic_vector(to_unsigned(DA_C, MAX_NUM_BITS_G));
								rvd_codeword_v(DA_C-1 downto 0)	:= err_lim_val_s(DA_C-1 downto 0);
							else	-- err_lim_type_v = '0'	-- Time to output "relative error limit values"
								enc_num_bits_v				:= std_logic_vector(to_unsigned(DR_C, MAX_NUM_BITS_G));
								rvd_codeword_v(DR_C-1 downto 0)	:= err_lim_val_s(DR_C-1 downto 0);
							end if;
						else	-- FIDEL_CTRL_TYPE_G = LOSSLESS_C		-- This case can never happen (as "err_lim_upd_en_s" would be always '0')
							enc_num_bits_v					:= (others => '0');
							rvd_codeword_v					:= (others => '0');
						end if;
					end if;
					
					-- Limitation to ensure that the calculate number of bits does not pass the maximum possible one
					enc_num_bits_s <= iif(unsigned(enc_num_bits_v) > unsigned(MAX_NUM_BITS_C), MAX_NUM_BITS_C, enc_num_bits_v);
					rvd_codeword_s <= rvd_codeword_v;
				end if;
			end if;
		end if;
	end process p_calc_low_codeword;

	-- Outputs
	enable_o		  <= enable2_s;
	img_coord_o		  <= img_coord2_s;
	curr_prefix_arr_o <= curr_prefix_arr2_s;
	enc_num_bits_o	  <= enc_num_bits_s;
	rvd_codeword_o	  <= rvd_codeword_s;
end architecture behavioural;