--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		18/04/2021
--------------------------------------------------------------------------------
-- IP name:		hybrid_statistic
--
-- Description: The adaptive code selection statistics (for Hybrid Entropy) used
--				to select the codeword type for each mapped quantizer index.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.utils_predictor.all;
use work.comp_predictor.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;

entity hybrid_statistic is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	 : std_logic_vector(1 downto 0)
	);
	port (
		clock_i			 : in  std_logic;
		reset_i			 : in  std_logic;
		
		enable_i		 : in  std_logic;
		enable_o		 : out std_logic;
		img_coord_i		 : in  img_coord_t;
		img_coord_o		 : out img_coord_t;
		
		err_lim_upd_en_i : in  std_logic;
		err_lim_upd_en_o : out std_logic;
		err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
		err_lim_val_o	 : out std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
		data_mp_quan_o	 : out std_logic_vector(D_C-1 downto 0);
		
		high_res_accu_o	 : out std_logic_vector(D_C*2-1 downto 0);	-- "Σ~z(t)" (high resolution accumulator)
		counter_o		 : out std_logic_vector(D_C-1 downto 0);	-- "Γ(t)"   (counter)
		reconst_bit_o	 : out std_logic;							-- Lets decoder reconstruct "Σ~z(t)" values
		
		entropy_select_o : out std_logic							-- Selected Entropy method (1=high-entropy, 0=low-entropy)
	);
end entity hybrid_statistic;

architecture behavioural of hybrid_statistic is	
	-- High-resolution accumulator and counter initial values
	constant COUNTER_INIT_C		: std_logic_vector(D_C-1 downto 0)		:= std_logic_vector(to_unsigned(2**Yo_C, D_C));
	constant HR_ACCU_INIT_C		: std_logic_vector(D_C+Yo_C-1 downto 0) := (2 => '1', others => '0');	-- 0x04 (Random value)
	-- Shortcut to Low-Entropy code table, subfield Threshold, position 0
	constant T0_C				: integer	  := LOW_ENTR_CODES_C.threshold(0);
	-- Number of delay samples to get t-1 (or W position), for BSQ/BIP/BIL orders (delays table defined inside "samples_store" IP)
	constant NEXT_W_C			: integer	  := locate_position(SMPL_ORDER_G, 1, NZ_C, 1);
	constant NEXT_SAMPLE_C		: integer	  := 1;
	
	-- Signals declaration
	signal enable_s				: std_logic   := '0';
	signal img_coord_s			: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal err_lim_upd_en_s		: std_logic	  := '0';
	signal err_lim_val_s		: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_mp_quan_s		: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	
	signal high_res_accu_s		: std_logic_vector(D_C*2-1 downto 0) := (others => '0');
	signal high_res_accu_old_s	: std_logic_vector(D_C*2-1 downto 0) := (others => '0');
	signal counter_s			: std_logic_vector(D_C-1 downto 0)	 := COUNTER_INIT_C;
	signal counter_old_s		: std_logic_vector(D_C-1 downto 0)	 := COUNTER_INIT_C;
	
	signal reconst_bit_s		: std_logic	  := '0';
	signal entropy_select_s		: std_logic	  := '1';		-- High-Entropy processing selected by default

begin
	-- Shift register to compute previous "err_lim_upd_en_i" signal, independently from the samples order
	i_err_lim_upd_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => 1,
		REG_SIZE_G	 => NEXT_SAMPLE_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => enable_s,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => img_coord_s,

		data_i(0)	 => err_lim_upd_en_i,
		data_o(0)	 => err_lim_upd_en_s
	);
	
	-- Shift register to compute previous "err_lim_val_i" signal, independently from the samples order
	i_err_lim_val_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => NEXT_SAMPLE_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => err_lim_val_i,
		data_o		 => err_lim_val_s
	);
	
	-- Shift register to compute previous "data_mp_quan_i" signal, independently from the samples order
	i_mp_quan_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => NEXT_SAMPLE_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => data_mp_quan_i,
		data_o		 => data_mp_quan_s
	);

	-- Shift register to compute previous "counter_s" signal, independently from the samples order
	i_cnt_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => NEXT_SAMPLE_C,
		INIT_VAL_G	 => COUNTER_INIT_C
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => counter_s,
		data_o		 => counter_old_s
	);

	-- Shift register to compute "t-1" from "high_res_accu_s" signal, depending on the samples order
	i_hres_accu_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C*2,
		REG_SIZE_G	 => iif(NEXT_W_C <= 2, NEXT_W_C+1, NEXT_W_C-1),
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => high_res_accu_s,
		data_o		 => high_res_accu_old_s
	);
	
	-- Calculation of the high resolution accumulator Σ~z(t) and counter Γ(t)
	p_calc_hres_accu_count : process(clock_i) is
		variable comp1_v : std_logic_vector(D_C*2-1 downto 0) := (others => '0');
		variable comp2_v : std_logic_vector(D_C-1 downto 0)	  := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				comp1_v			:= (others => '0');
				comp2_v			:= (others => '0');
				high_res_accu_s <= (others => '0');
				counter_s		<= COUNTER_INIT_C;
				reconst_bit_s	<= '0';
			else	-- Apart from "enable_i", this IP does nothing while "error limit values" are being encoded
				if (enable_i = '1' and err_lim_upd_en_i = '0') then
					if (img_coord_i.t = 0) then			-- Initial values (Sample for t=0 must not be encoded)
						-- If estimate of "?z(t)" available (z>0), then a reasonable init value is Σ~z(0)=4*Γ(0)*?z(0)
						if (img_coord_i.z > 0) then
							high_res_accu_s <= std_logic_vector(resize(to_unsigned(4, D_C*2) * resize(unsigned(COUNTER_INIT_C), D_C*2) * resize(unsigned(data_mp_quan_i), D_C*2), D_C*2));
						else	-- img_coord_i.z = 0
							high_res_accu_s <= std_logic_vector(resize(unsigned(HR_ACCU_INIT_C), D_C*2));
						end if;
						counter_s			<= COUNTER_INIT_C;
						reconst_bit_s		<= '0';
					else	-- img_coord_i.t > 0
						-- "high_res_accu_s" and "counter_s" are updated BEFORE coding each input sample (so "?z(t)" is used)
						comp1_v				:= std_logic_vector(resize(unsigned(high_res_accu_old_s) + to_unsigned(4, D_C*2) * resize(unsigned(data_mp_quan_i), D_C*2), D_C*2));
						comp2_v				:= std_logic_vector(resize(unsigned(counter_old_s) + to_unsigned(1, D_C), D_C));
						if (unsigned(counter_old_s) < to_unsigned(2**Y_C-1, D_C)) then
							high_res_accu_s <= comp1_v;
							-- Counter is updated only at the end of the spectral band...
							if (img_coord_i.z >= NZ_C-1) then
								counter_s	<= comp2_v;
							end if;
							reconst_bit_s	<= '0';
						else	-- This can only mean "counter_old_s = 2**Y_C-1", the rescaling (5.4.3.3.4.5)
							high_res_accu_s <= std_logic_vector(shift_usg(resize(unsigned(comp1_v) + to_unsigned(1, D_C*2), D_C*2), -1));
							-- Counter is updated only at the end of the spectral band...
							if (img_coord_i.z >= NZ_C-1) then
								counter_s	<= std_logic_vector(shift_usg(unsigned(comp2_v), -1));
								-- (5.4.3.3.5.1.2) Under this situation, "Σ~z(t-1)" LSb is encoded in bitstream, BEFORE any bits output
								reconst_bit_s <= '1';
							else
								reconst_bit_s <= '0';
							end if;
						end if;
					end if;
				end if;
			end if;
		end if;
	end process p_calc_hres_accu_count;

	-- 'High/Low-Entropy' processing selection (executed immediately after updating "high_res_accu_s" & "counter_s" signals
	p_entropy_select : process(reset_i, high_res_accu_s, counter_s) is
		variable threshold_v : integer := 0;
	begin
		if (reset_i = '1') then
			threshold_v		 := 0;
			entropy_select_s <= '1';					-- 'High-Entropy' processing selected by default
		else	-- Apart from "enable_i", this IP does nothing while "error limit values" are being encoded
			if (enable_i = '1' and err_lim_upd_en_i = '0') then
				if (img_coord_i.t = 0) then
					-- 'High-Entropy' selected when t=0, but it does not matter, as 'mapped quantized index' is not encoded later when t=0
					entropy_select_s <= '1';
				else	-- img_coord_i.t > 0
					-- If D=2, condition for using the 'High-Entropy' will never meet, so 'Low-Entropy' must be selected
					if (D_C = 2) then
						entropy_select_s <= '0';		-- 'Low-Entropy' processing selected
					else
						-- NOTE: No need to have a number longer than 30 bits, even if computed so...
						threshold_v := to_integer(resize((resize(unsigned(high_res_accu_s), D_C*3) * to_unsigned(2**14, D_C*3)) / resize(unsigned(counter_s), D_C*3), 30));

						if (threshold_v >= T0_C) then
							entropy_select_s <= '1';	-- 'High-Entropy' processing selected
						else
							entropy_select_s <= '0';	-- 'Low-Entropy' processing selected
						end if;						
					end if;
				end if;
			end if;
		end if;
	end process p_entropy_select;

	-- Outputs
	enable_o		 <= enable_s;
	img_coord_o		 <= img_coord_s;
	err_lim_upd_en_o <= err_lim_upd_en_s;
	err_lim_val_o	 <= err_lim_val_s;
	data_mp_quan_o	 <= data_mp_quan_s;
	high_res_accu_o  <= high_res_accu_s;
	counter_o		 <= counter_s;
	reconst_bit_o	 <= reconst_bit_s;
	entropy_select_o <= entropy_select_s;
end architecture behavioural;