--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		05/04/2021
--------------------------------------------------------------------------------
-- IP name:		smpl_adap_gpo2_code
--
-- Description: Variable-length binary GPO2 codeword to encode with the
--				Sample Adaptive Entropy Coder.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.utils_predictor.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;

entity smpl_adap_gpo2_code is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- Maximum number of bits (size) of every encoded data sample
		MAX_NUM_BITS_G		: integer;
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic
	);
	port (
		clock_i			 : in  std_logic;
		reset_i			 : in  std_logic;
		
		enable_i		 : in  std_logic;
		enable_o		 : out std_logic;
		img_coord_i		 : in  img_coord_t;
		img_coord_o		 : out img_coord_t;
		
		err_lim_upd_en_i : in  std_logic;
		err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
		
		accu_i			 : in  std_logic_vector(D_C*2-1 downto 0);		-- "Σz(t)" (accumulator)
		counter_i		 : in  std_logic_vector(D_C-1 downto 0);		-- "Γ(t)"  (counter)
		
		enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
		codeword_o		 : out std_logic_vector(Umax_C+D_C-1 downto 0)	-- "Rk(j)" (codeword)
	);
end entity smpl_adap_gpo2_code;

architecture behavioural of smpl_adap_gpo2_code is
	constant CONST_49_C		: unsigned(D_C-1 downto 0) := to_unsigned(49, D_C);
	signal numerator_s		: std_logic_vector(D_C*2-1 downto 0)		  := (others => '0');
	
	signal enable1_s		: std_logic   := '0';
	signal enable2_s		: std_logic   := '0';
	signal img_coord1_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord2_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal err_lim_cnt_s	: integer range 0 to 2*NZ_C					  := 0;
	signal err_lim_upd_en_s	: std_logic	  := '0';
	signal err_lim_val_s	: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	signal data_mp_quan_s	: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	signal counter_s		: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	
	signal kz_s				: integer range 0 to D_C					  := 0;
	signal enc_num_bits_s	: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal codeword_s		: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_smpl_adap_gpo2_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable1_s		 <= '0';
				enable2_s		 <= '0';
				img_coord1_s	 <= reset_img_coord(SMPL_ORDER_G);
				img_coord2_s	 <= reset_img_coord(SMPL_ORDER_G);
				err_lim_upd_en_s <= '0';
				err_lim_val_s	 <= (others => '0');
				data_mp_quan_s	 <= (others => '0');
				counter_s		 <= (others => '0');
			else
				enable1_s		 <= enable_i;
				enable2_s		 <= enable1_s;
				img_coord1_s	 <= img_coord_i;
				img_coord2_s	 <= img_coord1_s;
				err_lim_upd_en_s <= err_lim_upd_en_i;
				err_lim_val_s	 <= err_lim_val_i;
				data_mp_quan_s	 <= data_mp_quan_i;
				counter_s		 <= counter_i;
			end if;
		end if;
	end process p_smpl_adap_gpo2_delay;
	
	-- Calculation of "kz(t)" (Variable length code parameter)
	p_calc_kz : process(clock_i) is
		variable comp1_v : std_logic_vector(D_C*2-1 downto 0) := (others => '0');
		variable total_v : std_logic_vector(D_C*2-1 downto 0) := (others => '0');
		variable log2_v	 : integer range 0 to D_C			  := 0;
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				comp1_v		:= (others => '0');
				total_v		:= (others => '0');
				log2_v		:= 0;
				numerator_s <= (others => '0');
				kz_s		<= 0;
			else
				if (enable_i = '1' and err_lim_upd_en_i = '0') then
					-- NOTE: The "t>0" check must NOT be added here, as it is already integrated when computing the operands of these operations...
					comp1_v		:= std_logic_vector(resize(shift_usg(CONST_49_C * unsigned(counter_i), -7), D_C*2));
					-- As an additional delay is required, this is not a variable anymore, but a signal instead
					numerator_s <= std_logic_vector(resize(unsigned(accu_i) + unsigned(comp1_v), D_C*2));
					
					total_v		:= std_logic_vector(resize(unsigned(numerator_s) / unsigned(counter_s), D_C*2));
					-- To ensure no negative values, this function rounds up automatically
					log2_v		:= log2(to_integer(unsigned(total_v)) + 1);			-- '+1' to deal with the case of exactly matching 2-multiple value...
					
					-- The previous value must be constrained within [0, D-2]
					kz_s		<= iif(log2_v <= 0, 0, iif(log2_v > D_C-2, D_C-2, log2_v-1));	-- '-1' to remove the round up from previous 'log2' function
				end if;
			end if;
		end if;
	end process p_calc_kz;

	-- Calculation of "Rk(j)" (Variable-length GPO2 Codeword of Mapped Quantizer Index)
	p_calc_codeword : process(clock_i) is
		variable var_length_v	: integer range 0 to 2**D_C := 0;		-- No need to be a "std_logic_vector"
		variable err_lim_type_v	: std_logic := '0';
		variable data_mp_quan_v	: std_logic_vector(D_C-1 downto 0)		  := (others => '0');
		variable codeword_v		: std_logic_vector(Umax_C+D_C-1 downto 0) := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				var_length_v	:= 0;
				err_lim_type_v	:= '0';
				codeword_v		:= (others => '0');
				err_lim_cnt_s	<= 0;
				enc_num_bits_s	<= (others => '0');
				codeword_s		<= (others => '0');
				data_mp_quan_v  := (others => '0');
			else
				-- To be sure the packer does not produce problems after this IP, output data is restarted when fully processed...
				if (img_coord2_s = end_img_coord) then
					enc_num_bits_s <= (others => '0');
					codeword_v	   := (others => '0');
				end if;
				
				if (enable1_s = '1') then
					codeword_v	   := (others => '0');
					
					if (err_lim_upd_en_s = '0') then		-- Mapped quantizer indexes ("?z(t)") are encoded here!
						err_lim_cnt_s						<= 0;
						if (img_coord1_s.t = 0) then		-- Sample for t=0 must not be encoded!
							enc_num_bits_s					<= std_logic_vector(to_unsigned(D_C, MAX_NUM_BITS_G));
							codeword_v						:= data_mp_quan_s & (Umax_C-1 downto 0 => '0');
						else	-- img_coord1_s.t > 0
							var_length_v					:= to_integer(shift_usg(unsigned(data_mp_quan_s), -kz_s));
							if (var_length_v < Umax_C) then
								enc_num_bits_s				<= std_logic_vector(to_unsigned(var_length_v + 1 + kz_s, MAX_NUM_BITS_G));
								data_mp_quan_v				:= data_mp_quan_s sll (D_C - kz_s);
								codeword_v(D_C downto 0)	:= '1' & data_mp_quan_v;
								codeword_v					:= codeword_v sll (Umax_C-var_length_v-1);
							else	-- var_length_v >= Umax_C
								enc_num_bits_s				<= std_logic_vector(to_unsigned(Umax_C + D_C, MAX_NUM_BITS_G));
								codeword_v					:= (Umax_C-1 downto 0 => '0') & data_mp_quan_s;
							end if;
						end if;
						
					else	-- err_lim_upd_en_s = '1'		-- Error limit values are encoded here!
						if (FIDEL_CTRL_TYPE_G = ABS_ERR_C) then
							enc_num_bits_s					<= std_logic_vector(to_unsigned(DA_C, MAX_NUM_BITS_G));
							codeword_v(DA_C-1 downto 0)		:= err_lim_val_s(DA_C-1 downto 0);
						elsif (FIDEL_CTRL_TYPE_G = REL_ERR_C) then
							enc_num_bits_s					<= std_logic_vector(to_unsigned(DR_C, MAX_NUM_BITS_G));
							codeword_v(DR_C-1 downto 0)		:= err_lim_val_s(DR_C-1 downto 0);
						elsif (FIDEL_CTRL_TYPE_G = ABS_REL_C) then
							if ((ABS_ERR_BAND_TYPE_G = '1' and err_lim_cnt_s = 0) or (ABS_ERR_BAND_TYPE_G = '0' and err_lim_cnt_s < NZ_C)) then
								err_lim_type_v				:= '1';			-- Absolute error limit values (always the first one!)
							else
								err_lim_type_v				:= '0';			-- Relative error limit values (always the second one!)
							end if;
							err_lim_cnt_s					<= err_lim_cnt_s + 1;
							
							if (err_lim_type_v = '1') then	-- Time to output "absolute error limit values"
								enc_num_bits_s				<= std_logic_vector(to_unsigned(DA_C, MAX_NUM_BITS_G));
								codeword_v(DA_C-1 downto 0)	:= err_lim_val_s(DA_C-1 downto 0);
							else	-- err_lim_type_v = '0'	-- Time to output "relative error limit values"
								enc_num_bits_s				<= std_logic_vector(to_unsigned(DR_C, MAX_NUM_BITS_G));
								codeword_v(DR_C-1 downto 0)	:= err_lim_val_s(DR_C-1 downto 0);
							end if;
						else	-- FIDEL_CTRL_TYPE_G = LOSSLESS_C		-- This case can never happen (as "err_lim_upd_en_s" would be always '0')
							enc_num_bits_s					<= (others => '0');
							codeword_v						:= (others => '0');
						end if;
					end if;
					
					codeword_s <= codeword_v;
				end if;
			end if;
		end if;
	end process p_calc_codeword;

	-- Outputs
	enable_o		<= enable2_s;
	img_coord_o		<= img_coord2_s;
	enc_num_bits_o	<= enc_num_bits_s;
	codeword_o		<= codeword_s;
end architecture behavioural;