--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		16/10/2021
--------------------------------------------------------------------------------
-- IP name:		tb_parallel_sync_fifos
--
-- Description: Testbench for the "Parallel XPM Synchronous FIFOs" module
--
--
--------------------------------------------------------------------------------

-- Library instantiated, then declaration of files/functions to use
library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;

-- A reference to VUnit (and related) set of libraries and packages
library vunit_lib;
context vunit_lib.vunit_context;
context vunit_lib.com_context;
context vunit_lib.data_types_context;

-- Components and packages related to the Image IP
use vunit_lib.param_image.all;
use vunit_lib.types_image.all;
use vunit_lib.utils_image.all;
use vunit_lib.comp_image.all;

-- Components related to the Encoder IP
use vunit_lib.comp_encoder.all;

entity tb_parallel_sync_fifos is
    generic (
        encoded_tb_cfg	: string;
        runner_cfg		: string
    );
end entity tb_parallel_sync_fifos;

architecture behavioural of tb_parallel_sync_fifos is
	-- Record type to pack all signals (from Python script) together
	type tb_cfg_t is record
		SMPL_ORDER_G	: std_logic_vector(1 downto 0);
		DATA_WIDTH_G	: integer;
		DATA_DEPTH_G	: integer;
		NUM_PAR_FIFOS_G	: integer;
	end record tb_cfg_t;

	-- Function to decode the Python signals and connect them into the VHDL testbench
	impure function decode(encoded_tb_cfg : string) return tb_cfg_t is
	begin
	return (
		SMPL_ORDER_G	=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "SMPL_ORDER_PY")), 2)),
		DATA_WIDTH_G	=> integer'value(get(encoded_tb_cfg, "DATA_WIDTH_PY")),
		DATA_DEPTH_G	=> integer'value(get(encoded_tb_cfg, "DATA_DEPTH_PY")),
		NUM_PAR_FIFOS_G	=> integer'value(get(encoded_tb_cfg, "NUM_PAR_FIFOS_PY"))
	);
	end function decode;
	
	-- Constant of type "tb_cfg_t" and initialized by "decode()" function
	constant tb_cfg			: tb_cfg_t	  := decode(encoded_tb_cfg);
	
	signal clock_s			: std_logic   := '0';
	signal reset_s			: std_logic   := '1';

	signal enable_in_s		: std_logic   := '0';
	signal enable_out_s		: std_logic   := '0';
	signal img_coord_in_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal img_coord_out_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	
	signal read_fifo_s		: std_logic	  := '0';
	signal init_fifo_s		: std_logic	  := '0';
	
	signal data_in_arr_s	: array_slv_t(0 to tb_cfg.NUM_PAR_FIFOS_G-1)(tb_cfg.DATA_WIDTH_G-1 downto 0) := (others => (others => '0'));
	signal data_out_arr_s	: array_slv_t(0 to tb_cfg.NUM_PAR_FIFOS_G-1)(tb_cfg.DATA_WIDTH_G-1 downto 0) := (others => (others => '0'));

	signal flag_start, flag_stop : std_logic := '0';

begin
	reset_s <= '0' after 20 ns;
	clock_s <= not clock_s after 6.4 ns;	 -- Main clock frequency: 78125000 Hz
	
	-- Simulation MAIN process (if this one finishes, the whole simulation too)
	test_runner : process is
	begin
		-- Process starts by setting up VUnit using test_runner_setup procedure
		test_runner_setup(runner, runner_cfg);

		while test_suite loop	-- Pay very attention to the testcase name, arrows or accents prevent the system to work...
			if run("Parallel Sync FIFOs") then
				info("Running test case = " & to_string(running_test_case));

				flag_start <= '1';
			end if;
			
			wait until (flag_stop = '1');
		end loop;

		-- The process ends with the test_runner_cleanup procedure which will force the simulation to stop
		test_runner_cleanup(runner);
	end process;

	-- Set Watchdog with limited time (this number must be higher as any "wait for" statement used in the testbench)
	test_runner_watchdog(runner, 1 ms);

	-- Input signals stimulus calculation
	p_data_in_cond : process(clock_s) is
	begin
		if (rising_edge(clock_s)) then
			if (reset_s = '1') then
				read_fifo_s	  <= '0';
				data_in_arr_s <= (others => (others => '0'));
			else
				if (flag_start = '1' and init_fifo_s = '1') then
					if ((img_coord_out_s.x < NX_C-1) or (img_coord_out_s.y < NY_C-1) or (img_coord_out_s.z < NZ_C-1)) then
						if ((img_coord_in_s.x = NX_C/2) and (img_coord_in_s.z = NZ_C/2)) then
							read_fifo_s <= '0';
						else
							read_fifo_s <= '1';
						end if;
						
						for i in 0 to data_in_arr_s'length-1 loop
							if (data_in_arr_s(i) = (data_in_arr_s(i)'length-1 downto 0 => '1')) then
								data_in_arr_s(i) <= (others => '0');
							else
								data_in_arr_s(i) <= std_logic_vector(resize(unsigned(data_in_arr_s(i)) + to_unsigned(i, data_in_arr_s(i)'length), data_in_arr_s(i)'length));
							end if;
						end loop;
					end if;
				end if;
			end if;	
		end if;
	end process p_data_in_cond;

	-- Process to stop the simulation when finished (to use 'wait' statements, the process cannot be clocked)
	p_stop_sim : process is
	begin
		-- After one complete image, wait statement and request to finish the simulation
		wait until ((img_coord_out_s.x >= NX_C-1) and (img_coord_out_s.y >= NY_C-1) and (img_coord_out_s.z >= NZ_C-1));
		wait for 100 ns;
		flag_stop <= '1';
		
		-- Two possibilities to stop the simulation without VUnit framework (only the "assert" line really works, anyway)
		-- assert FALSE Report "Simulation Finished" severity FAILURE;
		-- std.env.finish;
	end process p_stop_sim;
	
	-- Entity to control the image coordinates
	i_image_coord_control : image_coord_control
	generic map(
		SMPL_ORDER_G	=> tb_cfg.SMPL_ORDER_G
	)
	port map(
		clock_i			=> clock_s,
		reset_i			=> reset_s,

		enable_i		=> flag_start and init_fifo_s,
		enable_o		=> enable_in_s,
		
		image_end_i		=> '0',
		img_coord_o		=> img_coord_in_s
	);
	
	-- Entity with FIFOs to store different data in parallel
	i_parallel_sync_fifos : parallel_sync_fifos
	generic map(
		SMPL_ORDER_G	=> tb_cfg.SMPL_ORDER_G,
		DATA_WIDTH_G	=> tb_cfg.DATA_WIDTH_G,
		DATA_DEPTH_G	=> tb_cfg.DATA_DEPTH_G,
		NUM_PAR_FIFOS_G	=> tb_cfg.NUM_PAR_FIFOS_G
	)
	port map(
		clock_i			=> clock_s,
		reset_i			=> reset_s,
		
		enable_i		=> enable_in_s,
		enable_o		=> enable_out_s,
		img_coord_i		=> img_coord_in_s,
		img_coord_o		=> img_coord_out_s,
		
		read_fifo_i		=> read_fifo_s,
		init_fifo_o		=> init_fifo_s,
		
		data_arr_i		=> data_in_arr_s,
		data_arr_o		=> data_out_arr_s
	);

end architecture behavioural;