--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		16/10/2021
--------------------------------------------------------------------------------
-- IP name:		tb_top_packer
--
-- Description: Testbench for the "Parallel XPM Synchronous FIFOs" module
--
--
--------------------------------------------------------------------------------

-- Library instantiated, then declaration of files/functions to use
library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;

-- A reference to VUnit (and related) set of libraries and packages
library vunit_lib;
context vunit_lib.vunit_context;
context vunit_lib.com_context;
context vunit_lib.data_types_context;

-- Components and packages related to the Image IP
use vunit_lib.param_image.all;
use vunit_lib.types_image.all;
use vunit_lib.utils_image.all;
use vunit_lib.comp_image.all;

-- Components and packages related to the Encoder IP
use vunit_lib.param_encoder.all;
use vunit_lib.comp_encoder.all;

-- Library to create random numbers
library osvvm;
use osvvm.RandomPkg.all;

entity tb_top_packer is
    generic (
        encoded_tb_cfg	: string;
        runner_cfg		: string
    );
end entity tb_top_packer;

architecture behavioural of tb_top_packer is
	-- Record type to pack all signals (from Python script) together
	type tb_cfg_t is record
		SMPL_ORDER_G : std_logic_vector(1 downto 0);
		HEADER_EN_G	 : std_logic;
	end record tb_cfg_t;

	-- Function to decode the Python signals and connect them into the VHDL testbench
	impure function decode(encoded_tb_cfg : string) return tb_cfg_t is
	begin
	return (
		SMPL_ORDER_G => std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "SMPL_ORDER_PY")), 2)),
		HEADER_EN_G	 => std_logic'value(get(encoded_tb_cfg, "HEADER_EN_PY"))
	);
	end function decode;
	
	-- Constant of type "tb_cfg_t" and initialized by "decode()" function
	constant tb_cfg				: tb_cfg_t	  := decode(encoded_tb_cfg);
	-- Maximum number of bits (size) of every encoded data sample
	constant MAX_NUM_BITS_C		: integer	  := log2(Umax_C + D_C) + 1;
	
	signal clock_s				: std_logic   := '0';
	signal reset_s				: std_logic   := '1';

	signal enable_header_s		: std_logic   := '0';
	signal enable_body_s		: std_logic   := '0';
	signal enable_body1_s		: std_logic   := '0';
	signal enable_body2_s		: std_logic   := '0';
	signal enable_body3_s		: std_logic   := '0';
	signal enable_out_s			: std_logic   := '0';
	signal img_coord_header_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal img_coord_body_s		: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal img_coord_body1_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal img_coord_body2_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal img_coord_body3_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal img_coord_out_s		: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	
	signal enc_header_width_s	: integer	  := 256;
	signal enc_header_data_s	: std_logic_vector(4095 downto 0)			  := (others => '0');
	signal enc_body_num_bits_s	: std_logic_vector(MAX_NUM_BITS_C-1 downto 0) := (others => '0');
	signal enc_body_data_s		: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
	signal init_fifo_s			: std_logic	  := '0';
	
	signal image_start_s		: std_logic	  := '0';		-- Flag to inform a new image is coming out
	signal image_end_s			: std_logic	  := '0';		-- Flag to inform current image has been finished
	signal valid_data_s			: std_logic	  := '0';		-- Flag to inform valid data is in the output line
	signal packed_data_s		: std_logic_vector(B_C*8-1 downto 0)		  := (others => '0');
	
	signal flag_start, flag_stop : std_logic  := '0';
	shared variable rnd_sv		 : RandomPType;

begin
	reset_s <= '0' after 20 ns;
	clock_s <= not clock_s after 6.4 ns;	 -- Main clock frequency: 78125000 Hz
	
	-- Simulation MAIN process (if this one finishes, the whole simulation too)
	test_runner : process is
	begin
		-- Process starts by setting up VUnit using test_runner_setup procedure
		test_runner_setup(runner, runner_cfg);
		
		-- Define the seed (initial configuration) of the (pseudo) random number generation
		rnd_sv.InitSeed(rnd_sv'instance_name);

		while test_suite loop	-- Pay very attention to the testcase name, arrows or accents prevent the system to work...
			if run("Top Packer") then
				info("Running test case = " & to_string(running_test_case));

				flag_start <= '1';
			end if;
			
			wait until (flag_stop = '1');
		end loop;

		-- The process ends with the test_runner_cleanup procedure which will force the simulation to stop
		test_runner_cleanup(runner);
	end process;

	-- Set Watchdog with limited time (this number must be higher as any "wait for" statement used in the testbench)
	test_runner_watchdog(runner, 1 ms);

	-- Input signals stimulus calculation
	p_data_in_cond : process(clock_s) is
	begin
		if (rising_edge(clock_s)) then
			if (reset_s = '1') then
				enc_body_num_bits_s	<= (others => '0');
				enc_body_data_s		<= (others => '0');
			else
				if (flag_start = '1' and init_fifo_s = '1') then
					if ((img_coord_out_s.x < NX_C-1) or (img_coord_out_s.y < NY_C-1) or (img_coord_out_s.z < NZ_C-1)) then						
						-- One bit less to reset this value, as its length has one additional bit to include ONLY up to Umax_C+D_C, and no more
						enc_body_num_bits_s <= rnd_sv.Randslv(4, Umax_C+D_C, enc_body_num_bits_s'length);
						
						if (enc_body_data_s >= (enc_body_data_s'length-1 downto 0 => '1')) then
							enc_body_data_s <= (others => '0');
						else
							enc_body_data_s <= std_logic_vector(resize(unsigned(enc_body_data_s) + to_unsigned(5, enc_body_data_s'length), enc_body_data_s'length));
						end if;
					end if;
				end if;
			end if;	
		end if;
	end process p_data_in_cond;
	-- Predefined data whose width is 256
	enc_header_data_s(enc_header_width_s-1 downto 0) <= x"0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF";

	-- Process to stop the simulation when finished (to use 'wait' statements, the process cannot be clocked)
	p_stop_sim : process is
	begin
		-- After one complete image, wait statement and request to finish the simulation
		wait until (image_end_s = '1');
		wait for 200 ns;
		flag_stop <= '1';
		
		-- Two possibilities to stop the simulation without VUnit framework (only the "assert" line really works, anyway)
		-- assert FALSE Report "Simulation Finished" severity FAILURE;
		-- std.env.finish;
	end process p_stop_sim;
	
	-- Entity to control the image coordinates
	i_image_coord_control : image_coord_control
	generic map(
		SMPL_ORDER_G		=> tb_cfg.SMPL_ORDER_G
	)
	port map(
		clock_i				=> clock_s,
		reset_i				=> reset_s,

		enable_i			=> flag_start and init_fifo_s,
		enable_o			=> enable_header_s,
		
		image_end_i			=> image_end_s,
		img_coord_o			=> img_coord_header_s
	);
	
	-- Input values delayed to synchronize them with the next modules in chain
	p_tb_packer_delay : process(clock_s) is
	begin
		if rising_edge(clock_s) then
			if (reset_s = '1') then
				enable_body1_s	  <= '0';
				enable_body2_s	  <= '0';
				enable_body3_s	  <= '0';
				img_coord_body1_s <= reset_img_coord(tb_cfg.SMPL_ORDER_G);
				img_coord_body2_s <= reset_img_coord(tb_cfg.SMPL_ORDER_G);
				img_coord_body3_s <= reset_img_coord(tb_cfg.SMPL_ORDER_G);
			else
				enable_body1_s	  <= enable_header_s;
				enable_body2_s	  <= enable_body1_s;
				enable_body3_s	  <= enable_body2_s;
				img_coord_body1_s <= img_coord_header_s;
				img_coord_body2_s <= img_coord_body1_s;
				img_coord_body3_s <= img_coord_body2_s;
			end if;
		end if;
	end process p_tb_packer_delay;
	
	-- Stimulus for the "Encoder Body" part depends on whether the "Encoder Header" was enabled or not
	enable_body_s	 <= enable_body3_s	  when (tb_cfg.HEADER_EN_G = '1') else enable_header_s;
	img_coord_body_s <= img_coord_body3_s when (tb_cfg.HEADER_EN_G = '1') else img_coord_header_s;
	
	-- Entity to pack data
	i_top_packer : top_packer
	generic map(
		SMPL_ORDER_G		=> tb_cfg.SMPL_ORDER_G,
		HEADER_EN_G			=> tb_cfg.HEADER_EN_G,
		MAX_NUM_BITS_G		=> MAX_NUM_BITS_C
	)
	port map(
		clock_i				=> clock_s,
		reset_i				=> reset_s,
		
		enable_header_i		=> enable_header_s,
		enable_body_i		=> enable_body_s,
		enable_o			=> enable_out_s,
		img_coord_header_i	=> img_coord_header_s,
		img_coord_body_i	=> img_coord_body_s,
		img_coord_o			=> img_coord_out_s,
		
		enc_header_width_i	=> enc_header_width_s,
		enc_header_data_i	=> enc_header_data_s,
		enc_body_num_bits_i	=> enc_body_num_bits_s,
		enc_body_data_i		=> enc_body_data_s,
		init_fifo_o			=> init_fifo_s,
		
		image_start_o		=> image_start_s,
		image_end_o			=> image_end_s,
		valid_data_o		=> valid_data_s,
		packed_data_o		=> packed_data_s
	);

end architecture behavioural;