--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		17/10/2021
--------------------------------------------------------------------------------
-- IP name:		top_packer
--
-- Description: Takes "Encoder Header" and (variable-length) "Encoder Body" IPs
--				data and packets it to the defined width.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_encoder.all;
use work.comp_encoder.all;

entity top_packer is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- 0: Encoder Header disabled, 1: Encoder Header enabled
		HEADER_EN_G			: std_logic;
		-- Maximum number of bits (size) of every encoded data sample
		MAX_NUM_BITS_G		: integer
	);
	port (
		clock_i				: in  std_logic;
		reset_i				: in  std_logic;
		
		enable_header_i		: in  std_logic;
		enable_body_i		: in  std_logic;
		enable_o			: out std_logic;
		img_coord_header_i	: in  img_coord_t;
		img_coord_body_i	: in  img_coord_t;
		img_coord_o			: out img_coord_t;
		
		enc_header_width_i	: in  integer;
		enc_header_data_i	: in  std_logic_vector(4095 downto 0);
		enc_body_num_bits_i	: in  std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
		enc_body_data_i		: in  std_logic_vector(Umax_C+D_C-1 downto 0);
		init_fifo_o			: out std_logic;
		
		image_start_o		: out std_logic;			-- Flag to inform a new image is coming out
		image_end_o			: out std_logic;			-- Flag to inform current image has been finished
		valid_data_o		: out std_logic;			-- Flag to inform valid data is in the output line
		packed_data_o		: out std_logic_vector(B_C*8-1 downto 0)
	);
end entity top_packer;

architecture behavioural of top_packer is
	-- Constants declaration
	constant OUTPUT_SIZE_C			: integer	  := 8 * B_C;
	-- Number of parallel sync. FIFOs to use (apart from the one for the 'image coordinates') -> Here 3, for "enc_body_num_bits_i", "enc_body_data_i" and "enable_body_i"
	constant NUM_PAR_FIFOS_C		: integer	  := 3;
	-- FIFO depth should be "enc_header_width_i/OUTPUT_SIZE_C", but as the signal needs to be calculated first, a big enough number is simply used in here
	constant FIFO_DATA_DEPTH_C		: integer	  := 256;

	-- Signals declaration
	signal enable_body_s			: std_logic   := '0';
	signal enable_packer_in_s		: std_logic   := '0';
	signal enable_packer_out1_s		: std_logic   := '0';
	signal enable_packer_out2_s		: std_logic   := '0';
	signal img_coord_body_s			: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord_packer_in_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord_packer_out1_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord_packer_out2_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal enc_header_en_s			: std_logic	  := '0';
	signal enc_body_en_s			: std_logic	  := '0';
	signal cnt_header_s				: integer	  := 0;
	signal cnt_body_s				: integer	  := 0;
	signal cnt_total_s				: integer	  := 0;
	signal cnt_valid_s				: integer	  := 0;
	signal cnt_valid_old_s			: integer	  := 0;
	signal pntr_body_s				: integer range 0 to max(OUTPUT_SIZE_C, NX_C*NY_C*NZ_C*(Umax_C+D_C)) := 0;
	
	signal init_fifo_s				: std_logic	  := '0';
	signal stim_received_s			: std_logic	  := '0';
	signal image_start_s			: std_logic	  := '0';
	signal image_end_s				: std_logic	  := '0';
	signal valid_data_s				: std_logic	  := '0';
	
	signal data_arr_in_s			: array_slv_t(0 to NUM_PAR_FIFOS_C-1)(Umax_C+D_C-1 downto 0) := (others => (others => '0'));
	signal data_arr_out_s			: array_slv_t(0 to NUM_PAR_FIFOS_C-1)(Umax_C+D_C-1 downto 0) := (others => (others => '0'));
	signal enc_body_num_bits_s		: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal enc_body_data_s			: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
	signal packed_data_s			: std_logic_vector(OUTPUT_SIZE_C-1 downto 0)  := (others => '0');

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_packer_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_packer_out1_s	<= '0';
				enable_packer_out2_s	<= '0';
				img_coord_packer_out1_s <= reset_img_coord(SMPL_ORDER_G);
				img_coord_packer_out2_s <= reset_img_coord(SMPL_ORDER_G);
				cnt_valid_old_s			<= 0;
			else
				enable_packer_out1_s	<= enable_packer_in_s;
				enable_packer_out2_s	<= enable_body_s;
				img_coord_packer_out1_s <= img_coord_packer_in_s;
				img_coord_packer_out2_s <= img_coord_body_s;
				cnt_valid_old_s			<= cnt_valid_s;
			end if;
		end if;
	end process p_packer_delay;
	
	-- FIFO to store the "Encoder Body" data, but ONLY needed if the "Encoder Header" is instantiated
	g_gen_fifos : if (HEADER_EN_G = '1') generate
		i_parallel_sync_fifos : parallel_sync_fifos
		generic map(
			SMPL_ORDER_G	=> SMPL_ORDER_G,
			DATA_WIDTH_G	=> Umax_C + D_C,
			DATA_DEPTH_G	=> FIFO_DATA_DEPTH_C,
			NUM_PAR_FIFOS_G	=> NUM_PAR_FIFOS_C
		)
		port map(
			clock_i			=> clock_i,
			reset_i			=> reset_i or image_end_s,
			
			enable_i		=> enable_body_i,
			enable_o		=> open,
			img_coord_i		=> img_coord_body_i,
			img_coord_o		=> img_coord_body_s,
			
			read_fifo_i		=> enc_body_en_s,
			init_fifo_o		=> init_fifo_s,
			
			data_arr_i		=> data_arr_in_s,
			data_arr_o		=> data_arr_out_s
		);
		-- To avoid strange behaviour from the FIFOs and for the sake of simulation, nothing is introduced on signals until "enable_body_i = '1'"
		data_arr_in_s(0)(MAX_NUM_BITS_G-1 downto 0) <= enc_body_num_bits_i when enable_body_i = '1' else (MAX_NUM_BITS_G-1 downto 0 => '0');
		data_arr_in_s(1)	<= enc_body_data_i when enable_body_i = '1' else (Umax_C+D_C-1 downto 0 => '0');
		data_arr_in_s(2)(0)	<= enable_body_i;
		enc_body_num_bits_s	<= data_arr_out_s(0)(MAX_NUM_BITS_G-1 downto 0);
		enc_body_data_s		<= data_arr_out_s(1);
		enable_body_s		<= data_arr_out_s(2)(0);
	else generate
		enc_body_num_bits_s	<= enc_body_num_bits_i;
		enc_body_data_s		<= enc_body_data_i;
		enable_body_s		<= enable_body_i;
		img_coord_body_s	<= img_coord_body_i;
		init_fifo_s			<= '1';					-- If no FIFOs enabled, then this flag is set to '1' not to disturb the system
	end generate g_gen_fifos;
	
	-- Output "enable" and "image coordinates" signals are assigned depending on whether the Encoder Header is enabled or not
	enable_packer_in_s		<= enable_header_i	  when (HEADER_EN_G = '1') else enable_body_i;
	img_coord_packer_in_s	<= img_coord_header_i when (HEADER_EN_G = '1') else img_coord_body_i;
	
	-- Process to pack the incoming data into packets with configurable width
	-- NOTE: Incoming data signal stores its value in the most significant bits...
	p_packing_data : process(clock_i) is
		variable pntr_header_v		 : integer range 0 to OUTPUT_SIZE_C							:= 0;
		variable pntr_body_v		 : integer range 0 to max(OUTPUT_SIZE_C, NX_C*NY_C*NZ_C*(Umax_C+D_C)) := 0;
		variable enc_num_bits_v		 : integer range 0 to Umax_C+D_C							:= 0;
		variable cnt_overflow_v		 : integer range 0 to max(Umax_C+D_C, OUTPUT_SIZE_C)		:= 0;
		variable data_overflow_v	 : std_logic_vector(Umax_C+D_C-1 downto 0)					:= (others => '0');
		variable packed_data_small_v : std_logic_vector(OUTPUT_SIZE_C-1 downto 0)				:= (others => '0');
		variable new_chunk_small_v	 : std_logic_vector(OUTPUT_SIZE_C-1 downto 0)				:= (others => '0');
		variable packed_data_big_v	 : std_logic_vector(NX_C*NY_C*NZ_C*(Umax_C+D_C)-1 downto 0) := (others => '0');
		variable new_chunk_big_v	 : std_logic_vector(NX_C*NY_C*NZ_C*(Umax_C+D_C)-1 downto 0) := (others => '0');
	begin
		if rising_edge(clock_i) then
			-- All signals are restarted with either "reset_i = '1'" or reaching the end of the image ("image_end_s = '1'")
			if (reset_i = '1' or image_end_s = '1') then
				pntr_header_v		:= 0;
				pntr_body_v			:= 0;
				enc_num_bits_v		:= 0;
				cnt_overflow_v		:= 0;
				data_overflow_v		:= (others => '0');
				packed_data_small_v	:= (others => '0');
				new_chunk_small_v	:= (others => '0');
				packed_data_big_v	:= (others => '0');
				new_chunk_big_v		:= (others => '0');
				cnt_header_s		<= 0;
				cnt_body_s			<= 0;
				cnt_total_s			<= 0;
				cnt_valid_s			<= 0;
				pntr_body_s			<= 0;
				valid_data_s		<= '0';
				packed_data_s		<= (others => '0');
			else
				-- This condition statement requires the very first and last enable signals to assert as well as if encoder header is running, so that the whole image is covered
				if (enable_packer_in_s = '1' or enable_packer_out2_s = '1' or enc_header_en_s = '1' or stim_received_s = '1') then
					-- Valid data is by default '0', asserting it only when required
					valid_data_s <= '0';
					
					if (HEADER_EN_G = '1' and (cnt_total_s < enc_header_width_i)) then			-- Encoder header data
						if ((cnt_header_s+1)*OUTPUT_SIZE_C <= enc_header_width_i) then
							packed_data_small_v := enc_header_data_i((cnt_header_s+1)*OUTPUT_SIZE_C-1 downto cnt_header_s*OUTPUT_SIZE_C);
						else
							pntr_header_v := enc_header_width_i - cnt_header_s*OUTPUT_SIZE_C;
							packed_data_small_v := (others => '0');
							packed_data_small_v(pntr_header_v-1 downto 0) := enc_header_data_i(enc_header_width_i-1 downto cnt_header_s*OUTPUT_SIZE_C);
						end if;
						packed_data_s <= packed_data_small_v;
						cnt_header_s  <= cnt_header_s + 1;
						-- Notification (and counter) of new data going out
						valid_data_s  <= '1';
						cnt_valid_s	  <= cnt_valid_s + 1;
						
					else	-- HEADER_EN_G = '0' or (cnt_total_s >= enc_header_width_i)			-- Encoder body data
						enc_num_bits_v := to_integer(unsigned(enc_body_num_bits_s));
						if (Umax_C+D_C <= OUTPUT_SIZE_C) then		-- If "Encoder Body" (maximum) data size is shorter or equal than the output width
							cnt_overflow_v	  := 0;
							data_overflow_v	  := (others => '0');
							new_chunk_small_v := (others => '0');
							
							-- If next data chunk COMPLETELY fits within the output width, and still some bits are left
							if (pntr_body_v+enc_num_bits_v < OUTPUT_SIZE_C) then
								-- Except at the beginning, the new data is fetched, shifted to the right to finally put in the proper position at the output (with XOR gate)
								if (pntr_body_v = 0) then
									packed_data_small_v(enc_num_bits_v-1 downto 0)	:= enc_body_data_s(Umax_C+D_C-1 downto Umax_C+D_C-enc_num_bits_v);
									packed_data_small_v := packed_data_small_v	sll (OUTPUT_SIZE_C-enc_num_bits_v);
								else	-- pntr_body_v > 0
									new_chunk_small_v(enc_num_bits_v-1 downto 0)	:= enc_body_data_s(Umax_C+D_C-1 downto Umax_C+D_C-enc_num_bits_v);
									new_chunk_small_v	:= new_chunk_small_v	sll (OUTPUT_SIZE_C-enc_num_bits_v);
									new_chunk_small_v	:= new_chunk_small_v	srl pntr_body_v;
									packed_data_small_v	:= packed_data_small_v	xor new_chunk_small_v;
								end if;									
								-- Counter increased according to the new number of bits
								pntr_body_v			:= pntr_body_v + enc_num_bits_v;
							
							else	-- If next data chunk either COMPLETELY fits (with NO bits left) or PARTIALLY fits within the output width
								cnt_overflow_v 		:= pntr_body_v + enc_num_bits_v - OUTPUT_SIZE_C;
								-- Only the part that still fits is included in the output vector (after shifting data to the right)
								new_chunk_small_v(enc_num_bits_v-cnt_overflow_v-1 downto 0) := enc_body_data_s(Umax_C+D_C-1 downto Umax_C+D_C-enc_num_bits_v+cnt_overflow_v);
								new_chunk_small_v	:= new_chunk_small_v		sll (OUTPUT_SIZE_C-enc_num_bits_v+cnt_overflow_v);
								new_chunk_small_v	:= new_chunk_small_v		srl pntr_body_v;
								packed_data_small_v	:= packed_data_small_v		xor new_chunk_small_v;
								packed_data_s		<= packed_data_small_v;
								-- The previous vector is cleaned up, and then the remaining data is store there for later (and its counter is updated too!)
								packed_data_small_v	:= (others => '0');
								-- If there is an overflow, such data is extracted and stored for next time
								if (cnt_overflow_v > 0) then
									data_overflow_v(cnt_overflow_v-1 downto 0)		:= enc_body_data_s(Umax_C+D_C-enc_num_bits_v+cnt_overflow_v-1 downto Umax_C+D_C-enc_num_bits_v);
									packed_data_small_v(cnt_overflow_v-1 downto 0) 	:= data_overflow_v(cnt_overflow_v-1 downto 0);
									packed_data_small_v := packed_data_small_v	sll (OUTPUT_SIZE_C-cnt_overflow_v);
								end if;
								pntr_body_v		:= cnt_overflow_v;
								-- Notification (and counter) of new data going out
								valid_data_s	<= '1';
								cnt_valid_s		<= cnt_valid_s + 1;
							end if;
							
							-- Special case for the very last sample to output (either on last input coordinates, or last output coordinates)...
							-- ... one clock cycle later, where remaining data is sent out as well (leaving the rest of "OUTPUT_SIZE_C" as '0'-padding)
							if ((img_coord_body_s = end_img_coord and cnt_overflow_v = 0) or img_coord_packer_out2_s = end_img_coord) then
								pntr_body_v		:= 0;
								packed_data_s	<= packed_data_small_v;
								-- Notification (and counter) of new data going out
								valid_data_s	<= '1';
								cnt_valid_s		<= cnt_valid_s + 1;
							end if;
						
						else	-- Umax_C+D_C > OUTPUT_SIZE_C		-- If "Encoder Body" (maximum) data size is longer than the output width
							-- Store the incoming stimulus ONLY while it is provided
							if (stim_received_s = '0') then
								new_chunk_big_v		  := (others => '0');
								-- Except at the beginning, the new data is fetched, shifted to the left to finally put in the right position (XOR gate) at the output
								if (pntr_body_v = 0) then
									packed_data_big_v(enc_num_bits_v-1 downto 0) := enc_body_data_s(Umax_C+D_C-1 downto Umax_C+D_C-enc_num_bits_v);
								else	-- pntr_body_v /= 0
									new_chunk_big_v(enc_num_bits_v-1 downto 0)	 := enc_body_data_s(Umax_C+D_C-1 downto Umax_C+D_C-enc_num_bits_v);
									new_chunk_big_v	  := new_chunk_big_v	sll pntr_body_v;
									packed_data_big_v := packed_data_big_v	xor new_chunk_big_v;
								end if;
								pntr_body_v			  := pntr_body_v + enc_num_bits_v;
							end if;
							
							-- Only when there is enough data, it is outputted, counters subtracted, and arrays shifted to have the next oldest data at rightmost
							if (pntr_body_v >= OUTPUT_SIZE_C) then
								packed_data_s	  <= packed_data_big_v(OUTPUT_SIZE_C-1 downto 0);
								packed_data_big_v := packed_data_big_v srl OUTPUT_SIZE_C;
								pntr_body_v		  := pntr_body_v - OUTPUT_SIZE_C;
								-- Notification (and counter) of new data going out
								valid_data_s	  <= '1';
								cnt_valid_s		  <= cnt_valid_s + 1;
							-- Special case for the very last sample to output, where remaining data is sent out as well (leaving the rest of "OUTPUT_SIZE_C" as '0'-padding)
							elsif (stim_received_s = '1' and (pntr_body_v < OUTPUT_SIZE_C)) then
								packed_data_s	  <= packed_data_big_v(OUTPUT_SIZE_C-1 downto 0);
								pntr_body_v		  := 0;
								-- Notification (and counter) of new data going out
								valid_data_s	  <= '1';
								cnt_valid_s		  <= cnt_valid_s + 1;
							end if;
						end if;
						cnt_body_s	<= cnt_body_s + 1;
						pntr_body_s <= pntr_body_v;
					end if;
					cnt_total_s	<= cnt_total_s + OUTPUT_SIZE_C;
				else
					-- No valid data when 'enable' signals are deasserted
					valid_data_s <= '0';
				end if;
			end if;
		end if;
	end process p_packing_data;

	-- Handles the flag when the stimulus has been fully received
	p_stimulus_flag : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				stim_received_s <= '0';
			else
				-- Flag to signal when the incoming "Encoder Body Data" has been fully received (which does NOT necessarily mean that all processing is done!)
				if (img_coord_packer_out2_s = second_end_img_coord(SMPL_ORDER_G)) then
					stim_received_s <= '1';
				elsif (image_end_s = '1') then
					stim_received_s <= '0';
				end if;
			end if;
		end if;
	end process p_stimulus_flag;
	
	-- Handles all not-synchronized flags (this process covers all possibilities, even when header is longer than input samples)
	p_encoder_flags : process(reset_i, stim_received_s, cnt_valid_old_s, cnt_valid_s, cnt_total_s, pntr_body_s, enable_packer_out1_s) is
	begin
		if (reset_i = '1') then
			enc_header_en_s <= '0';
			enc_body_en_s	<= '0';
			image_start_s	<= '0';
			image_end_s		<= '0';
		else
			-- Flag to signal when the "Encoder Header Data" is being sent out (right at the beginning)
			if (HEADER_EN_G = '1') then
				if (enable_packer_out1_s = '1' and cnt_total_s <= enc_header_width_i) then
					enc_header_en_s <= '1';
				elsif (cnt_total_s > enc_header_width_i) then
					enc_header_en_s <= '0';
				end if;
			else	-- HEADER_EN_G = '0'
				enc_header_en_s <= '0';
			end if;
			
			-- Flag to signal when the "Encoder Body Data" is being sent out (right after the "Encoder Header")			
			if (HEADER_EN_G = '1') then
				if (enable_packer_out1_s = '1' and cnt_total_s > enc_header_width_i) then
					enc_body_en_s <= '1';
				elsif (cnt_total_s = 0) then
					enc_body_en_s <= '0';
				end if;
			else	-- HEADER_EN_G = '0'
				if (enable_packer_out1_s = '1') then
					enc_body_en_s <= '1';
				elsif (cnt_total_s = 0) then
					enc_body_en_s <= '0';
				end if;
			end if;
			
			-- A one clock cycle flag to say when a new (un)compressed image starts to being sent out
			if (cnt_valid_old_s = 0 and cnt_valid_s = 1) then
				image_start_s <= '1';
			else
				image_start_s <= '0';
			end if;
		
			-- A one clock cycle flag to say when the current (un)compressed image has been fully sent out
			if (stim_received_s = '1' and cnt_total_s /= 0 and pntr_body_s <= 0) then
				image_end_s <= '1';
			else
				image_end_s <= '0';
			end if;
		end if;
	end process p_encoder_flags;
	
	-- Outputs
	enable_o		<= enable_packer_out1_s;
	img_coord_o		<= img_coord_packer_out2_s when enc_body_en_s = '1' else reset_img_coord(SMPL_ORDER_G);
	init_fifo_o		<= init_fifo_s;
	image_start_o	<= image_start_s;
	image_end_o		<= image_end_s;
	valid_data_o	<= valid_data_s;
	packed_data_o	<= packed_data_s;
end architecture behavioural;