library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;

-- Package Declaration Section
package utils_image is
	
	pure function max(max1 : in integer; max2 : in integer) return integer;
	pure function max(max1 : in signed; max2 : in signed) return signed;
	pure function max(max1 : in unsigned; max2 : in unsigned) return unsigned;
	pure function min(min1 : in integer; min2 : in integer) return integer;
	pure function min(min1 : in signed; min2 : in signed) return signed;
	pure function min(min1 : in unsigned; min2 : in unsigned) return unsigned;
	
	pure function log2(val_in : in integer) return natural;
	pure function power2(val_in : in integer) return std_logic_vector;
	
	pure function division_down(dw1 : in integer; dw2 : in integer) return integer;
	pure function division_down(dw1 : in signed; dw2 : in signed) return signed;
	pure function division_down(dw1 : in unsigned; dw2 : in unsigned) return unsigned;
	pure function division_up(up1 : in integer; up2 : in integer) return integer;
	pure function division_up(up1 : in signed; up2 : in signed) return signed;
	pure function division_up(up1 : in unsigned; up2 : in unsigned) return unsigned;

	pure function iif(cond_in : in boolean; true_in : in integer; false_in : in integer) return integer;
	pure function iif(cond_in : in boolean; true_in : in signed; false_in : in signed) return signed;
	pure function iif(cond_in : in boolean; true_in : in unsigned; false_in : in unsigned) return unsigned;
	pure function iif(cond_in : in boolean; true_in : in std_logic; false_in : in std_logic) return std_logic;
	pure function iif(cond_in : in boolean; true_in : in std_logic_vector; false_in : in std_logic_vector) return std_logic_vector;
	pure function iif(cond_in : in boolean; true_in : in array_integer_t; false_in : in array_integer_t) return array_integer_t;
	
	pure function sgl_slv_2_arr_slv(slv_in : in std_logic_vector; width_in : in integer) return array_slv_t;
	pure function sgl_slv_2_arr_sgn(slv_in : in std_logic_vector; width_in : in integer) return array_signed_t;
	pure function sgl_slv_2_arr_usg(slv_in : in std_logic_vector; width_in : in integer) return array_unsigned_t;
	pure function arr_slv_2_sgl_slv(arr_slv_in : in array_slv_t) return std_logic_vector;
	pure function arr_sgn_2_sgl_slv(arr_sgn_in : in array_signed_t) return std_logic_vector;
	pure function arr_usg_2_sgl_slv(arr_usg_in : in array_unsigned_t) return std_logic_vector;
	pure function arr_slv_2_arr_sgn(arr_slv_in : in array_slv_t) return array_signed_t;
	pure function arr_slv_2_arr_usg(arr_slv_in : in array_slv_t) return array_unsigned_t;

	pure function fill_float(seed_in : in integer) return float_t;
	pure function fill_array(length_in : in integer; units_in : integer) return array_integer_t;

	pure function check_array_all_pos_same(array_in : in array_integer_t; orig_val_in : in integer) return std_logic;
	pure function check_matrix_all_pos_same(matrix_in : in matrix_integer_t; orig_val_in : in integer) return std_logic;

	pure function set_smpl_limits(smpl_type_in : std_logic) return smpl_lim_t;
	pure function locate_position(smpl_order : in std_logic_vector; pos1 : in integer; pos2 : in integer; pos3 : in integer) return integer;
	pure function reset_img_coord(smpl_order : in std_logic_vector) return img_coord_t;
	pure function second_end_img_coord(smpl_order : in std_logic_vector) return img_coord_t;
	pure function end_img_coord return img_coord_t;

end package utils_image;

-- Package Body Section
package body utils_image is

	-- Returns the bigger value from the two arguments (integer format)
	pure function max(max1 : in integer; max2 : in integer) return integer is
	begin
		if (max1 > max2) then
			return max1;
		else
			return max2;
		end if;
	end function max;
	
	-- Returns the bigger value from the two arguments (signed format)
	pure function max(max1 : in signed; max2 : in signed) return signed is
		variable max_width_v : integer := max(max1'length, max2'length);
	begin
		if (resize(max1, max_width_v) > resize(max2, max_width_v)) then
			return resize(max1, max_width_v);
		else
			return resize(max2, max_width_v);
		end if;
	end function max;

	-- Returns the bigger value from the two arguments (unsigned format)
	pure function max(max1 : in unsigned; max2 : in unsigned) return unsigned is
		variable max_width_v : integer := max(max1'length, max2'length);
	begin
		if (resize(max1, max_width_v) > resize(max2, max_width_v)) then
			return resize(max1, max_width_v);
		else
			return resize(max2, max_width_v);
		end if;
	end function max;
	
	-- Returns the smaller value from the two arguments (integer format)
	pure function min(min1 : in integer; min2 : in integer) return integer is
	begin
		if (min1 < min2) then
			return min1;
		else
			return min2;
		end if;
	end function min;

	-- Returns the smaller value from the two arguments (signed format)
	pure function min(min1 : in signed; min2 : in signed) return signed is
		variable max_width_v : integer := max(min1'length, min2'length);
	begin
		if (resize(min1, max_width_v) < resize(min2, max_width_v)) then
			return resize(min1, max_width_v);
		else
			return resize(min2, max_width_v);
		end if;
	end function min;

	-- Returns the smaller value from the two arguments (unsigned format)
	pure function min(min1 : in unsigned; min2 : in unsigned) return unsigned is
		variable max_width_v : integer := max(min1'length, min2'length);
	begin
		if (resize(min1, max_width_v) < resize(min2, max_width_v)) then
			return resize(min1, max_width_v);
		else
			return resize(min2, max_width_v);
		end if;
	end function min;
	
	-- Synthetisable 2-based logarithm (Log2) function
	pure function log2(val_in : in integer) return natural is
		constant VAL_IN_C : unsigned(31 downto 0) := to_unsigned(val_in, 32);
		variable temp_v	  : unsigned(31 downto 0) := (0 => '1', others => '0');
		variable result_v : natural := 0;
	begin
		-- Integer values do not accept more than 32 bits
		for i in 0 to VAL_IN_C'length-1 loop
			if (VAL_IN_C <= temp_v) then
				result_v := i;
				exit;
			end if;
			-- Moving bits 1 position to left (equivalent to 2**i)
			temp_v := temp_v sll 1;
		end loop;
		
		return result_v;
	end function log2;
	
	-- Synthetisable Power-of-2 (2**x) function
	pure function power2(val_in : in integer) return std_logic_vector is
		variable result_v : std_logic_vector(val_in downto 0) := (others => '0');
	begin
		-- Power of 2 needs one extra bit to represent the value, as it does not start from 0, but 1
		result_v(val_in) := '1';
		
		return result_v;
	end function power2;

	-- Makes a division and rounds the result down (integer format)
	pure function division_down(dw1 : in integer; dw2 : in integer) return integer is
		variable result_v : integer := 0;
	begin
		if (dw2 /= 0) then
			result_v := dw1/dw2;
		else
			result_v := 0;
		end if;
		
		return result_v;
	end function division_down;

	-- Makes a division and rounds the result down (signed format)
	pure function division_down(dw1 : in signed; dw2 : in signed) return signed is
		variable max_width_v : integer := max(dw1'length, dw2'length);
		variable result_v	 : signed(max_width_v-1 downto 0) := (others => '0');
	begin
		result_v := resize(resize(dw1, max_width_v) / resize(dw2, max_width_v), max_width_v);
		
		return result_v;
	end function division_down;

	-- Makes a division and rounds the result down (unsigned format)
	pure function division_down(dw1 : in unsigned; dw2 : in unsigned) return unsigned is
		variable max_width_v : integer := max(dw1'length, dw2'length);
		variable result_v	 : unsigned(max_width_v-1 downto 0) := (others => '0');
	begin
		result_v := resize(resize(dw1, max_width_v) / resize(dw2, max_width_v), max_width_v);
		
		return result_v;
	end function division_down;

	-- Makes a division and rounds the result up (integer format)
	pure function division_up(up1 : in integer; up2 : in integer) return integer is
		variable result_v : integer := 0;
	begin
		if (up1 mod up2 /= 0) then
			result_v := up1/up2 + 1;
		else
			result_v := up1/up2;
		end if;
		
		return result_v;
	end function division_up;

	-- Makes a division and rounds the result up (signed format)
	pure function division_up(up1 : in signed; up2 : in signed) return signed is
		variable max_width_v : integer := max(up1'length, up2'length);
		variable result_v	 : signed(max_width_v-1 downto 0) := (others => '0');
	begin
		if (resize(resize(up1, max_width_v) mod resize(up2, max_width_v), max_width_v) /= to_signed(0, max_width_v)) then
			result_v := resize(resize(up1, max_width_v) / resize(up2, max_width_v) + to_signed(1, max_width_v), max_width_v);
		else
			result_v := resize(resize(up1, max_width_v) / resize(up2, max_width_v), max_width_v);
		end if;
		
		return result_v;
	end function division_up;

	-- Makes a division and rounds the result up (unsigned format)
	pure function division_up(up1 : in unsigned; up2 : in unsigned) return unsigned is
		variable max_width_v : integer := max(up1'length, up2'length);
		variable result_v	 : unsigned(max_width_v-1 downto 0) := (others => '0');
	begin
		if (resize(resize(up1, max_width_v) mod resize(up2, max_width_v), max_width_v) /= to_unsigned(0, max_width_v)) then
			result_v := resize(resize(up1, max_width_v) / resize(up2, max_width_v) + to_unsigned(1, max_width_v), max_width_v);
		else
			result_v := resize(resize(up1, max_width_v) / resize(up2, max_width_v), max_width_v);
		end if;
		
		return result_v;
	end function division_up;

	-- Immediate IF (iif) function for "integer" values
	pure function iif(cond_in : in boolean; true_in : in integer; false_in : in integer) return integer is
	begin
		if (cond_in = true) then
			return true_in;
		else
			return false_in;
		end if;
	end function iif;

	-- Immediate IF (iif) function for "signed" values
	pure function iif(cond_in : in boolean; true_in : in signed; false_in : in signed) return signed is
	begin
		if (cond_in = true) then
			return true_in;
		else
			return false_in;
		end if;
	end function iif;
	
	-- Immediate IF (iif) function for "unsigned" values
	pure function iif(cond_in : in boolean; true_in : in unsigned; false_in : in unsigned) return unsigned is
	begin
		if (cond_in = true) then
			return true_in;
		else
			return false_in;
		end if;
	end function iif;
	
	-- Immediate IF (iif) function for "std_logic" values
	pure function iif(cond_in : in boolean; true_in : in std_logic; false_in : in std_logic) return std_logic is
	begin
		if (cond_in = true) then
			return true_in;
		else
			return false_in;
		end if;
	end function iif;
	
	-- Immediate IF (iif) function for "std_logic_vector" values
	pure function iif(cond_in : in boolean; true_in : in std_logic_vector; false_in : in std_logic_vector) return std_logic_vector is
	begin
		if (cond_in = true) then
			return true_in;
		else
			return false_in;
		end if;
	end function iif;
	
	-- Immediate IF (iif) function for "array_integer_t" values
	pure function iif(cond_in : in boolean; true_in : in array_integer_t; false_in : in array_integer_t) return array_integer_t is
	begin
		if (cond_in = true) then
			return true_in;
		else
			return false_in;
		end if;
	end function iif;
	
	-- Transforms from "std_logic_vector" into "array_slv_t"
	pure function sgl_slv_2_arr_slv(slv_in : in std_logic_vector; width_in : in integer) return array_slv_t is
		constant SIZE_C		: integer := slv_in'length / width_in;
		variable arr_slv_v	: array_slv_t(0 to SIZE_C-1)(width_in-1 downto 0) := (others => (others => '0'));
	begin
		for i in 0 to SIZE_C-1 loop
			arr_slv_v(i) := std_logic_vector(slv_in((i+1)*width_in-1 downto i*width_in));
		end loop;
		
		return arr_slv_v;
	end function sgl_slv_2_arr_slv;
	
	-- Transforms from "std_logic_vector" into "array_signed_t"
	pure function sgl_slv_2_arr_sgn(slv_in : in std_logic_vector; width_in : in integer) return array_signed_t is
		constant SIZE_C		: integer := slv_in'length / width_in;
		variable arr_sgn_v	: array_signed_t(0 to SIZE_C-1)(width_in-1 downto 0) := (others => (others => '0'));
	begin
		for i in 0 to SIZE_C-1 loop
			arr_sgn_v(i) := signed(slv_in((i+1)*width_in-1 downto i*width_in));
		end loop;
		
		return arr_sgn_v;
	end function sgl_slv_2_arr_sgn;
	
	-- Transforms from "std_logic_vector" into "array_unsigned_t"
	pure function sgl_slv_2_arr_usg(slv_in : in std_logic_vector; width_in : in integer) return array_unsigned_t is
		constant SIZE_C		: integer := slv_in'length / width_in;
		variable arr_usg_v	: array_unsigned_t(0 to SIZE_C-1)(width_in-1 downto 0) := (others => (others => '0'));
	begin
		for i in 0 to SIZE_C-1 loop
			arr_usg_v(i) := unsigned(slv_in((i+1)*width_in-1 downto i*width_in));
		end loop;
		
		return arr_usg_v;
	end function sgl_slv_2_arr_usg;
	
	-- Transforms from "array_slv_t" into "std_logic_vector"
	pure function arr_slv_2_sgl_slv(arr_slv_in : in array_slv_t) return std_logic_vector is
		variable sgl_slv_v : std_logic_vector(arr_slv_in'length*arr_slv_in(0)'length-1 downto 0) := (others => '0');
	begin
		for i in 0 to arr_slv_in'length-1 loop
			sgl_slv_v((i+1)*arr_slv_in(0)'length-1 downto i*arr_slv_in(0)'length) := std_logic_vector(arr_slv_in(i));
		end loop;
		
		return sgl_slv_v;
	end function arr_slv_2_sgl_slv;
	
	-- Transforms from "array_signed_t" into "std_logic_vector"
	pure function arr_sgn_2_sgl_slv(arr_sgn_in : in array_signed_t) return std_logic_vector is
		variable sgl_slv_v : std_logic_vector(arr_sgn_in'length*arr_sgn_in(0)'length-1 downto 0) := (others => '0');
	begin
		for i in 0 to arr_sgn_in'length-1 loop
			sgl_slv_v((i+1)*arr_sgn_in(0)'length-1 downto i*arr_sgn_in(0)'length) := std_logic_vector(arr_sgn_in(i));
		end loop;
		
		return sgl_slv_v;
	end function arr_sgn_2_sgl_slv;
	
	-- Transforms from "array_unsigned_t" into "std_logic_vector"
	pure function arr_usg_2_sgl_slv(arr_usg_in : in array_unsigned_t) return std_logic_vector is
		variable sgl_slv_v : std_logic_vector(arr_usg_in'length*arr_usg_in(0)'length-1 downto 0) := (others => '0');
	begin
		for i in 0 to arr_usg_in'length-1 loop
			sgl_slv_v((i+1)*arr_usg_in(0)'length-1 downto i*arr_usg_in(0)'length) := std_logic_vector(arr_usg_in(i));
		end loop;
		
		return sgl_slv_v;
	end function arr_usg_2_sgl_slv;
	
	-- Transforms from "array_slv_t" into "array_signed_t"
	pure function arr_slv_2_arr_sgn(arr_slv_in : in array_slv_t) return array_signed_t is
		variable arr_sgn_v : array_signed_t(0 to arr_slv_in'length-1)(arr_slv_in(0)'length-1 downto 0) := (others => (others => '0'));
	begin
		for i in 0 to arr_slv_in'length-1 loop
			arr_sgn_v(i) := signed(arr_slv_in(i));
		end loop;
		
		return arr_sgn_v;
	end function arr_slv_2_arr_sgn;
	
	-- Transforms from "array_slv_t" into "array_unsigned_t"
	pure function arr_slv_2_arr_usg(arr_slv_in : in array_slv_t) return array_unsigned_t is
		variable arr_usg_v : array_unsigned_t(0 to arr_slv_in'length-1)(arr_slv_in(0)'length-1 downto 0) := (others => (others => '0'));
	begin
		for i in 0 to arr_slv_in'length-1 loop
			arr_usg_v(i) := unsigned(arr_slv_in(i));
		end loop;
		
		return arr_usg_v;
	end function arr_slv_2_arr_usg;

	-- Fills a float type signal, using an input number as seed
	pure function fill_float(seed_in : in integer) return float_t is
		variable float_v : float_t;
	begin
		float_v.b	  := 1;
		float_v.alpha := seed_in;
		float_v.j	  := seed_in * 2;
		
		return float_v;
	end function fill_float;
	
	-- Fills an array increasing a config number on each iteration
	pure function fill_array(length_in : in integer; units_in : integer) return array_integer_t is
		variable arr_int_v : array_integer_t(0 to length_in-1);
	begin
		arr_int_v(0) := 0;
		for i in 1 to length_in-1 loop
			arr_int_v(i) := arr_int_v(i-1) + units_in;
		end loop;

		return arr_int_v;
	end function fill_array;
	
	-- Returns '0' if all values within an "integer array" are the same. Otherwise '1'
	pure function check_array_all_pos_same(array_in : in array_integer_t; orig_val_in : in integer) return std_logic is
	begin
		for i in 0 to array_in'length-1 loop
			if (array_in(i) /= orig_val_in) then
				return '1';
			end if;
		end loop;
		
		return '0';
	end function check_array_all_pos_same;
	
	-- Returns '0' if all values within an "integer matrix" are the same. Otherwise '1'
	pure function check_matrix_all_pos_same(matrix_in : in matrix_integer_t; orig_val_in : in integer) return std_logic is
	begin
		for i in 0 to matrix_in'length-1 loop
			for j in 0 to matrix_in(0)'length-1 loop
				if (matrix_in(i)(j) /= orig_val_in) then
					return '1';
				end if;
			end loop;
		end loop;
		
		return '0';
	end function check_matrix_all_pos_same;

	-- Defines the sample limit values (0 = samples unsigned type, 1 = samples signed type)
	pure function set_smpl_limits(smpl_type_in : std_logic) return smpl_lim_t is
		variable smpl_lim_v : smpl_lim_t;
	begin
		if (smpl_type_in = '0') then	-- "signed" type
			smpl_lim_v.min := std_logic_vector(S_MIN_SGN_C);
			smpl_lim_v.mid := std_logic_vector(S_MID_SGN_C);
			smpl_lim_v.max := std_logic_vector(S_MAX_SGN_C);
		else							-- "unsigned" type
			smpl_lim_v.min := std_logic_vector(S_MIN_USN_C);
			smpl_lim_v.mid := std_logic_vector(S_MID_USN_C);
			smpl_lim_v.max := std_logic_vector(S_MAX_USN_C);
		end if;
		
		return smpl_lim_v;
	end function set_smpl_limits;
	
	-- Returns the right position value, depending on the input order
	pure function locate_position(smpl_order : in std_logic_vector; pos1 : in integer; pos2 : in integer; pos3 : in integer) return integer is
		variable pos_v : integer := 0;
	begin
		if (smpl_order = BSQ_C) then
			pos_v := pos1;
		elsif (smpl_order = BIP_C) then
			pos_v := pos2;
		elsif (smpl_order = BIL_C) then
			pos_v := pos3;
		else	-- Unexpected case
			pos_v := 0;
		end if;

		return pos_v;
	end function locate_position;
	
	-- Resets the image coordinates record, depending on the chosen sample order
	pure function reset_img_coord(smpl_order : in std_logic_vector) return img_coord_t is
		variable img_coord_v : img_coord_t;
	begin
		if (smpl_order = BSQ_C) then
			img_coord_v.x := -1;
			img_coord_v.y := 0;
			img_coord_v.z := 0;
		elsif (smpl_order = BIP_C) then
			img_coord_v.x := 0;
			img_coord_v.y := 0;
			img_coord_v.z := -1;
		elsif (smpl_order = BIL_C) then
			img_coord_v.x := -1;
			img_coord_v.y := 0;
			img_coord_v.z := 0;
		else
			img_coord_v.x := -1;
			img_coord_v.y := -1;
			img_coord_v.z := -1;
		end if;
		img_coord_v.t := img_coord_v.y * NX_C + img_coord_v.x;

		return img_coord_v;
	end function reset_img_coord;
	
	-- Returns the second last image coordinate
	pure function second_end_img_coord(smpl_order : in std_logic_vector) return img_coord_t is
		variable img_coord_v : img_coord_t;
	begin
		if (smpl_order = BSQ_C) then
			img_coord_v.x := NX_C-2;
			img_coord_v.y := NY_C-1;
			img_coord_v.z := NZ_C-1;
		elsif (smpl_order = BIP_C) then
			img_coord_v.x := NX_C-1;
			img_coord_v.y := NY_C-1;
			img_coord_v.z := NZ_C-2;
		else	-- smpl_order = BIL_C
			img_coord_v.x := NX_C-2;
			img_coord_v.y := NY_C-1;
			img_coord_v.z := NZ_C-1;
		end if;
		img_coord_v.t := img_coord_v.y * NX_C + img_coord_v.x;

		return img_coord_v;
	end function second_end_img_coord;

	-- Returns the last image coordinate
	pure function end_img_coord return img_coord_t is
		variable img_coord_v : img_coord_t;
	begin
		img_coord_v.x := NX_C-1;
		img_coord_v.y := NY_C-1;
		img_coord_v.z := NZ_C-1;
		img_coord_v.t := img_coord_v.y * NX_C + img_coord_v.x;

		return img_coord_v;
	end function end_img_coord;
	
end package body utils_image;