--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		24/10/2020
--------------------------------------------------------------------------------
-- IP name:		image_coord_control
--
-- Description: Keeps track and outputs the current image position of the cube
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_encoder.all;

entity image_coord_control is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G : std_logic_vector(1 downto 0)
	);
	port (
		clock_i		 : in  std_logic;
		reset_i		 : in  std_logic;

		enable_i	 : in  std_logic;
		enable_o	 : out std_logic;
		
		image_end_i	 : in  std_logic;
		img_coord_o	 : out img_coord_t
	);
end entity image_coord_control;

architecture behavioural of image_coord_control is
	-- Regardless of the sample order, ONLY the first coordinate to change must start with '-1' (either 'x' or 'z'), so that the first change outputs (0, 0, 0)
	-- Image coordinates															--					 							   _____
	signal x_s : integer range -1 to NX_C-1 := iif(SMPL_ORDER_G = BIP_C, 0, -1);	-- X coord: Sample/Width			 +--> +x	 /|	   /|
	signal y_s : integer range -1 to NY_C-1 := 0;									-- Y coord: Frame					/|			/_|___/ |
	signal z_s : integer range -1 to NZ_C-1 := iif(SMPL_ORDER_G = BIP_C, -1, 0);	-- Z coord: (Spectral) Band		   / |			| |___|_|
																					--								  v	 v			| /   | /
	signal enable_s		: std_logic	:= '0';											--								+z	 +y			|/____|/
	signal image_end_s	: std_logic	:= '0';
	signal stall_s		: std_logic	:= '0';
	
	-- Intermediate values (only for BIP/BIL input order) --> Milica says to use ONLY either BIP_C (M_C=NZ_C) or BIL_C (M_C=1)
	constant M_C		: integer	:= iif(SMPL_ORDER_G = BIP_C, M_BIP_C, M_BIL_C);
	constant I_MAX_C	: integer	:= division_up(NZ_C, M_C);
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_image_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				image_end_s	 <= '0';
				enable_s	 <= '0';
			else
				-- "image_end_s" is the only signal that must not be restarted with "stall_s", so such condition only applies for "enable_s"
				image_end_s	 <= image_end_i;
				if (stall_s = '1') then
					enable_s <= '0';
				else
					enable_s <= enable_i;
				end if;
			end if;
		end if;
	end process p_image_delay;
	
	-- Once whole image has been introduced, next one must not start until current one has been completely processed
	p_stall_count : process(reset_i, image_end_s, image_end_i, x_s, y_s, z_s) is
	begin
		if (reset_i = '1') then
			stall_s <= '0';
		else
			if (x_s >= NX_C-1 and y_s >= NY_C-1 and z_s >= NZ_C-1) then
				stall_s <= '1';
			else
				-- Due to synchro. issues, the deassertion of "stall_s" can ONLY be done with a falling edge of "image_end_i"
				if (image_end_s = '1' and image_end_i = '0') then
					stall_s <= '0';
				end if;
			end if;
		end if;
	end process p_stall_count;

	-- Coordinates counting for input order BSQ
	g_img_coord_BSQ : if (SMPL_ORDER_G = BSQ_C) generate
		p_img_coord_BSQ : process(clock_i) is
		begin
			if (rising_edge(clock_i)) then
				if (reset_i = '1' or stall_s = '1') then
					x_s <= -1;		-- With BSQ_C, first coordinate to change is always 'x', so it should always start with '-1'
					y_s <= 0;
					z_s <= 0;
				else
					if (enable_i = '1') then
						if (x_s < NX_C-1) then
							x_s <= x_s + 1;
						else	
						x_s <= 0;
							if (y_s < NY_C-1) then
								y_s <= y_s + 1;
							else
								y_s <= 0;
								if (z_s < NZ_C-1) then
									z_s <= z_s + 1;
								else
									z_s <= 0;
								end if;
							end if;
						end if;
					end if;
				end if;
			end if;
		end process p_img_coord_BSQ;
	end generate g_img_coord_BSQ;

	-- Coordinates counting for input order BI (BIP and BIL are just special cases from BI)
	-- If M_C=NZ_C -> BIP_C input order; and if M_C=1 -> BIL_C input order
	g_img_coord_BI : if (SMPL_ORDER_G = BIP_C or SMPL_ORDER_G = BIL_C) generate
		p_img_coord_BI : process(clock_i) is
			variable i_v : integer range 0 to I_MAX_C := 0;
			variable x_v : integer range -1 to 0	  := iif(SMPL_ORDER_G = BIP_C, 0, -1);
			variable z_v : integer range -1 to 0	  := iif(SMPL_ORDER_G = BIP_C, -1, 0);
		begin
			if (rising_edge(clock_i)) then
				if (reset_i = '1' or stall_s = '1') then
					i_v := 0;
					x_s <= x_v;
					y_s <= 0;
					z_s <= z_v;
				else
					if (enable_i = '1') then
						if (z_s < work.utils_image.min((i_v+1)*M_C-1, NZ_C-1)) then
							z_s <= z_s + 1;
						else	
							z_s <= i_v * M_C;
							if (x_s < NX_C-1) then
								x_s <= x_s + 1;
							else
								x_s <= 0;
								-- "i_v" must be variable (and "z_s" updated immediately), so that there is no delays with the image coordinates
								if (i_v < I_MAX_C-1) then
									i_v := i_v + 1;
									z_s <= i_v * M_C;
								else
									i_v := 0;
									z_s <= i_v * M_C;
									if (y_s < NY_C-1) then
										y_s <= y_s + 1;
									else
										y_s <= 0;
									end if;
								end if;
							end if;
						end if;
					end if;
				end if;
			end if;
		end process p_img_coord_BI;
	end generate g_img_coord_BI;

	-- Outputs
	enable_o	<= enable_s;
	img_coord_o	<= (
		x	=> x_s,
		y	=> y_s,
		z	=> z_s,
		t	=> y_s * NX_C + x_s
	);
	-- NOTE: Using "t", one can calculate "x" (x = t mod NX_C) and "y" (y = division_down(t, NX_C)) coordinates
end architecture behavioural;