library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;

use work.param_predictor.all;
use work.types_predictor.all;

-- Package Declaration Section
package comp_predictor is
	
	-----------------------------------------------------------------------------------------------------------------
	------------------------------------------------- PREDICTOR TOP -------------------------------------------------
	-----------------------------------------------------------------------------------------------------------------

	component top_predictor is
		generic (
			-- 0: "signed" type samples, 1: "unsigned" type samples
			SMPL_TYPE_G			: std_logic;
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
			LSUM_TYPE_G			: std_logic_vector(1 downto 0);
			-- 1: Full prediction mode, 0: Reduced prediction mode
			PREDICT_MODE_G		: std_logic;
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic;
			-- 1: enabled, 0: disabled
			PER_ERR_LIM_UPD_G	: std_logic;
			-- 1: Custom weight init, 0: Default weight init
			W_INIT_TYPE_G		: std_logic
		);
		port (
			clock_i			: in  std_logic;
			reset_i			: in  std_logic;

			enable_i		: in  std_logic;
			enable_o		: out std_logic;
			img_coord_i		: in  img_coord_t;
			img_coord_o		: out img_coord_t;
			
			data_s0_i		: in  std_logic_vector(D_C-1 downto 0);		-- "sz(t)" (original sample)
			data_mp_quan_o	: out std_logic_vector(D_C-1 downto 0)		-- "?z(t)" (mapped quantizer index)
		);
	end component top_predictor;

	-----------------------------------------------------------------------------------------------------------------
	-------------------------------------------------- ADDER BLOCK --------------------------------------------------
	-----------------------------------------------------------------------------------------------------------------
	
	component adder is
		generic (
			-- 0: "signed" type samples, 1: "unsigned" type samples
			SMPL_TYPE_G	  : std_logic;
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G  : std_logic_vector(1 downto 0)
		);
		port (
			clock_i		  : in  std_logic;
			reset_i		  : in  std_logic;
			
			img_en_i	  : in  std_logic;
			enable_o	  : out std_logic;
			img_coord_i	  : in  img_coord_t;
			img_coord_o	  : out img_coord_t;
			
			pred_en_i	  : in  std_logic;
			pred_coord_i  : in  img_coord_t;
			data_s3_i	  : in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
			data_s3_o	  : out std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
			data_s4_lsb_i : in  std_logic;							-- "s~z(t)"  (double-resolution predicted sample) LSb
			data_s4_lsb_o : out std_logic;							-- "s~z(t)"  (double-resolution predicted sample) LSb
			
			data_s0_i	  : in  std_logic_vector(D_C-1 downto 0);	-- "sz(t)"  (original sample)
			en_res_o	  : out std_logic;
			data_res_o	  : out std_logic_vector(D_C+2-1 downto 0)	-- "/\z(t)" (prediction residual)
		);
	end component adder;

	-----------------------------------------------------------------------------------------------------------------
	------------------------------------------------ QUANTIZER BLOCK ------------------------------------------------
	-----------------------------------------------------------------------------------------------------------------
	
	component quantizer is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic;
			-- 1: enabled, 0: disabled
			PER_ERR_LIM_UPD_G	: std_logic
		);
		port (
			clock_i		 : in  std_logic;
			reset_i		 : in  std_logic;
			
			enable_i	 : in  std_logic;
			enable_o	 : out std_logic;
			img_coord_i	 : in  img_coord_t;
			img_coord_o	 : out img_coord_t;
			
			data_s3_i	 : in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
			data_res_i	 : in  std_logic_vector(D_C+2-1 downto 0);	-- "/\z(t)" (prediction residual)
			
			data_merr_o	 : out std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)" (maximum error)
			data_quant_o : out std_logic_vector(D_C+2-1 downto 0)	-- "qz(t)" (quantizer index)
		);
	end component quantizer;

	component fidelity_ctrl is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G		: std_logic_vector(1 downto 0);
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
			-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
			ABS_ERR_BAND_TYPE_G	: std_logic;
			REL_ERR_BAND_TYPE_G	: std_logic;
			-- 1: enabled, 0: disabled
			PER_ERR_LIM_UPD_G	: std_logic
		);
		port (
			clock_i		: in  std_logic;
			reset_i		: in  std_logic;
			
			enable_i	: in  std_logic;
			enable_o	: out std_logic;
			img_coord_i	: in  img_coord_t;
			img_coord_o	: out img_coord_t;
			
			data_s3_i	: in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
			data_merr_o	: out std_logic_vector(D_C+1-1 downto 0)	-- "mz(t)"  (maximum error)
		);
	end component fidelity_ctrl;

	-----------------------------------------------------------------------------------------------------------------
	-------------------------------------------------- MAPPER BLOCK -------------------------------------------------
	-----------------------------------------------------------------------------------------------------------------

	component mapper is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	: std_logic_vector(1 downto 0);
			-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
			SMPL_LIMIT_G	: smpl_lim_t
		);
		port (
			clock_i			: in  std_logic;
			reset_i			: in  std_logic;
			
			enable_i		: in  std_logic;
			enable_o		: out std_logic;
			img_coord_i		: in  img_coord_t;
			img_coord_o		: out img_coord_t;
			
			data_s3_i		: in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
			data_s4_lsb_i	: in  std_logic;							-- "s~z(t)"  (double-resolution predicted sample) LSb
			data_merr_i		: in  std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)"  (maximum error)
			data_quant_i	: in  std_logic_vector(D_C+2-1 downto 0);	-- "qz(t)"  (quantizer index)
			data_mp_quan_o	: out std_logic_vector(D_C-1 downto 0)		-- "δz(t)"  (mapped quantizer index)
		);
	end component mapper;

	component scaled_diff is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	: std_logic_vector(1 downto 0);
			-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
			SMPL_LIMIT_G	: smpl_lim_t
		);
		port (
			clock_i			: in  std_logic;
			reset_i			: in  std_logic;
			
			enable_i		: in  std_logic;
			enable_o		: out std_logic;
			img_coord_i		: in  img_coord_t;
			img_coord_o		: out img_coord_t;
			
			data_s3_i		: in  std_logic_vector(D_C+1-1 downto 0);		-- "s^z(t)" (predicted sample)
			data_merr_i		: in  std_logic_vector(D_C+1-1 downto 0);		-- "mz(t)"  (maximum error)
			scaled_diff_o	: out std_logic_vector(D_C+1-1 downto 0)		-- "θz(t)"  (scaled difference)
		);
	end component scaled_diff;

	-----------------------------------------------------------------------------------------------------------------
	------------------------------------------ SAMPLE REPRESENTATIVE BLOCK ------------------------------------------
	-----------------------------------------------------------------------------------------------------------------

	component sample_representative is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	  : std_logic_vector(1 downto 0);
			-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
			SMPL_LIMIT_G	  : smpl_lim_t;
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G : std_logic_vector(1 downto 0)
		);
		port (
			clock_i		: in  std_logic;
			reset_i		: in  std_logic;
			
			enable_i	: in  std_logic;
			enable_o	: out std_logic;
			img_coord_i	: in  img_coord_t;
			img_coord_o	: out img_coord_t;
			
			data_merr_i	: in  std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)"	 (maximum error)
			data_quant_i: in  std_logic_vector(D_C+2-1 downto 0);	-- "qz(t)"   (quantizer index)
			data_s0_i	: in  std_logic_vector(D_C-1 downto 0);		-- "sz(t)"	 (original sample)
			data_s3_i	: in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)"  (predicted sample)
			data_s6_i	: in  std_logic_vector(Re_C-1 downto 0);	-- "s)z(t)"	 (high-resolution predicted sample)

			data_s1_o	: out std_logic_vector(D_C-1 downto 0);		-- "s'z(t)"  (clipped quantizer bin center)
			data_s2_o	: out std_logic_vector(D_C-1 downto 0)		-- "s''z(t)" (sample representative)
		);
	end component sample_representative;

	component dbl_res_smpl_repr is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	  : std_logic_vector(1 downto 0);
			-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
			FIDEL_CTRL_TYPE_G : std_logic_vector(1 downto 0)
		);
		port (
			clock_i		: in  std_logic;
			reset_i		: in  std_logic;
			
			enable_i	: in  std_logic;
			enable_o	: out std_logic;
			img_coord_i	: in  img_coord_t;
			img_coord_o	: out img_coord_t;

			data_merr_i	: in  std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)"	(maximum error)
			data_quant_i: in  std_logic_vector(D_C+2-1 downto 0);	-- "qz(t)"	(quantizer index)		
			data_s6_i	: in  std_logic_vector(Re_C-1 downto 0);	-- "s)z(t)"	(high-resolution predicted sample)
			data_s1_i	: in  std_logic_vector(D_C-1 downto 0);		-- "s'z(t)"	(clipped quantizer bin center)
			
			data_s5_o	: out std_logic_vector(Re_C-1 downto 0)		-- "s~''z(t)" (double-resolution sample representative)
		);
	end component dbl_res_smpl_repr;

	component clip_quant_bin_center is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G : std_logic_vector(1 downto 0);
			-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
			SMPL_LIMIT_G : smpl_lim_t
		);
		port (
			clock_i		 : in  std_logic;
			reset_i		 : in  std_logic;
			
			enable_i	 : in  std_logic;
			enable_o	 : out std_logic;
			img_coord_i	 : in  img_coord_t;
			img_coord_o	 : out img_coord_t;
			
			data_s3_i	 : in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
			data_merr_i	 : in  std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)"  (maximum error)
			data_quant_i : in  std_logic_vector(D_C+2-1 downto 0);	-- "qz(t)"  (quantizer index)
			data_s1_o	 : out std_logic_vector(D_C-1 downto 0)		-- "s'z(t)" (clipped quantizer bin center)
		);
	end component clip_quant_bin_center;
	
	-----------------------------------------------------------------------------------------------------------------
	------------------------------------------------ PREDICTION BLOCK -----------------------------------------------
	-----------------------------------------------------------------------------------------------------------------

	component prediction is
		generic (
			-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
			SMPL_LIMIT_G	: smpl_lim_t;
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	: std_logic_vector(1 downto 0);
			-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
			LSUM_TYPE_G		: std_logic_vector(1 downto 0);
			-- 1: Full prediction mode, 0: Reduced prediction mode
			PREDICT_MODE_G	: std_logic;
			-- 1: Custom weight init, 0: Default weight init
			W_INIT_TYPE_G	: std_logic
		);
		port (
			clock_i		  : in  std_logic;
			reset_i		  : in  std_logic;
			
			enable_i	  : in  std_logic;
			enable_o	  : out std_logic;
			img_coord_i	  : in  img_coord_t;
			img_coord_o	  : out img_coord_t;
			
			pz_i		  : in  integer;
			data_s0_i	  : in  std_logic_vector(D_C-1 downto 0);		-- "sz(t)"   (original sample)
			data_s1_i	  : in  std_logic_vector(D_C-1 downto 0);		-- "s'z(t)"	 (clipped quantizer bin center)
			data_s2_i	  : in  std_logic_vector(D_C-1 downto 0);		-- "s''z(t)" (sample representative)
			
			data_s3_o	  : out std_logic_vector(D_C+1-1 downto 0);		-- "s^z(t)"	 (predicted sample)
			data_s4_lsb_o : out std_logic;								-- "s~z(t)"  (double-resolution predicted sample) LSb
			data_s6_o	  : out std_logic_vector(Re_C-1 downto 0)		-- "s)z(t)"	 (high-resolution predicted sample)
		);
	end component prediction;

	component predicted_sample is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G : std_logic_vector(1 downto 0)
		);
		port (
			clock_i		: in  std_logic;
			reset_i		: in  std_logic;
			
			enable_i	: in  std_logic;
			enable_o	: out std_logic;
			img_coord_i	: in  img_coord_t;
			img_coord_o	: out img_coord_t;

			data_s4_i	: in  std_logic_vector(D_C+2-1 downto 0);	-- "s~z(t)" (double-resolution predicted sample)
			data_s3_o	: out std_logic_vector(D_C+1-1 downto 0)	-- "s^z(t)"	(predicted sample)
		);
	end component predicted_sample;

	component dbl_res_pred_smpl is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G : std_logic_vector(1 downto 0);
			-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
			SMPL_LIMIT_G : smpl_lim_t
		);
		port (
			clock_i		: in  std_logic;
			reset_i		: in  std_logic;
			
			enable_i	: in  std_logic;
			enable_o	: out std_logic;
			img_coord_i	: in  img_coord_t;
			img_coord_o	: out img_coord_t;
			
			data_s0z1_i	: in  std_logic_vector(D_C-1 downto 0);		-- "sz-1(t)" (original sample, but from previous spectral band)
			data_s6_i	: in  std_logic_vector(Re_C-1 downto 0);	-- "s)z(t)"  (high-resolution predicted sample)
			data_s4_o	: out std_logic_vector(D_C+2-1 downto 0)	-- "s~z(t)"  (double-resolution predicted sample)
		);
	end component dbl_res_pred_smpl;
	
	component high_res_pred_smpl is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G : std_logic_vector(1 downto 0);
			-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
			SMPL_LIMIT_G : smpl_lim_t
		);
		port (
			clock_i		: in  std_logic;
			reset_i		: in  std_logic;
			
			enable_i	: in  std_logic;
			enable_o	: out std_logic;
			img_coord_i	: in  img_coord_t;
			img_coord_o	: out img_coord_t;
																	-- "d^z(t)" (predicted central local difference)
			data_pred_cldiff_i : in std_logic_vector(MAX_CZ_C+D_C+4+OMEGA_C+3-1 downto 0);
			data_lsum_i	: in  std_logic_vector(D_C+4-1 downto 0);	-- "σz(t)"  (local sum)
			data_s6_o	: out std_logic_vector(Re_C-1 downto 0)		-- "s)z(t)" (high-resolution predicted sample)
		);
	end component high_res_pred_smpl;

	component pred_ctrl_local_diff is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	: std_logic_vector(1 downto 0);
			-- 1: Full prediction mode, 0: Reduced prediction mode
			PREDICT_MODE_G	: std_logic
		);
		port (
			clock_i		  : in  std_logic;
			reset_i		  : in  std_logic;
			
			enable_i	  : in  std_logic;
			enable_o	  : out std_logic;
			img_coord_i	  : in  img_coord_t;
			img_coord_o	  : out img_coord_t;
			
			pz_i		  : in  integer;
			ldiff_vect_i  : in  array_slv_t(0 to MAX_CZ_C-1)(D_C+4-1 downto 0);			 -- "Uz(t)"  (local difference vector)
			weight_vect_i : in  array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0);		 -- "Wz(t)"  (weight vector)
			
			data_pred_cldiff_o : out std_logic_vector(MAX_CZ_C+D_C+4+OMEGA_C+3-1 downto 0) -- "d^z(t)" (predicted central local difference)
		);
	end component pred_ctrl_local_diff;

	component weights_vector is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	: std_logic_vector(1 downto 0);
			-- 1: Full prediction mode, 0: Reduced prediction mode
			PREDICT_MODE_G	: std_logic;
			-- 1: Custom weight init, 0: Default weight init
			W_INIT_TYPE_G	: std_logic
		);
		port (
			clock_i			: in  std_logic;
			reset_i			: in  std_logic;
			
			enable_i		: in  std_logic;
			enable_o		: out std_logic;
			img_coord_i		: in  img_coord_t;
			img_coord_o		: out img_coord_t;
			
			pred_err_bit_i	: in  integer range -1 to 1;							 -- "ez(t)"  (double-resolution prediction error) bit
			w_scal_exp_i	: in  integer range -127 to 127;						 -- "p(t)"  (weight update scaling exponent)
			ldiff_vect_i	: in  array_slv_t(0 to MAX_CZ_C-1)(D_C+4-1 downto 0);	 -- "Uz(t)" (local difference vector)
			weight_vect_o	: out array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0) -- "Wz(t)" (weight vector)
		);
	end component weights_vector;

	component weight_upd_scal_exp is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G : std_logic_vector(1 downto 0)
		);
		port (
			clock_i		 : in  std_logic;
			reset_i		 : in  std_logic;
			
			enable_i	 : in  std_logic;
			enable_o	 : out std_logic;
			img_coord_i	 : in  img_coord_t;
			img_coord_o	 : out img_coord_t;
			
			w_scal_exp_o : out integer range -127 to 127		-- "p(t)" (weight update scaling exponent)
		);
	end component weight_upd_scal_exp;

	component dbl_res_pred_error is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	: std_logic_vector(1 downto 0)
		);
		port (
			clock_i			: in  std_logic;
			reset_i			: in  std_logic;
			
			enable_i		: in  std_logic;
			enable_o		: out std_logic;
			img_coord_i		: in  img_coord_t;
			img_coord_o		: out img_coord_t;
			
			data_s1_i		: in  std_logic_vector(D_C-1 downto 0);		-- "s'z(t)" (clipped quantizer bin center)
			data_s4_i		: in  std_logic_vector(D_C+2-1 downto 0);	-- "s~z(t)" (double-resolution predicted sample)
			pred_err_bit_o	: out integer range -1 to 1					-- "ez(t)"  (double-resolution prediction error) bit
		);
	end component dbl_res_pred_error;

	component local_diff_vector is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	: std_logic_vector(1 downto 0);
			-- 1: Full prediction mode, 0: Reduced prediction mode
			PREDICT_MODE_G  : std_logic
		);
		port (
			clock_i		: in  std_logic;
			reset_i		: in  std_logic;
			
			enable_i	: in  std_logic;
			enable_o	: out std_logic;
			img_coord_i	: in  img_coord_t;
			img_coord_o	: out img_coord_t;
			
			ldiff_pos_i	: in  ldiff_pos_t;
			ldiff_vect_o: out array_slv_t(0 to MAX_CZ_C-1)(D_C+4-1 downto 0)	-- "Uz(t)" (local difference vector)
		);
	end component local_diff_vector;

	component local_differences is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G	: std_logic_vector(1 downto 0);
			-- 1: Full prediction mode, 0: Reduced prediction mode
			PREDICT_MODE_G	: std_logic
		);
		port (
			clock_i		  : in  std_logic;
			reset_i		  : in  std_logic;
			
			enable_i	  : in  std_logic;
			enable_o	  : out std_logic;
			img_coord_i	  : in  img_coord_t;
			img_coord_o	  : out img_coord_t;
			
			data_lsum_i	  : in  std_logic_vector(D_C+4-1 downto 0);
			data_s2_pos_i : in  s2_pos_t;
			ldiff_pos_o	  : out ldiff_pos_t
		);
	end component local_differences;

	component local_sum is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G  : std_logic_vector(1 downto 0);
			-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
			SMPL_LIMIT_G  : smpl_lim_t;
			-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
			LSUM_TYPE_G	  : std_logic_vector(1 downto 0)
		);
		port (
			clock_i		  : in  std_logic;
			reset_i		  : in  std_logic;
			
			enable_i	  : in  std_logic;
			enable_o	  : out std_logic;
			img_coord_i	  : in  img_coord_t;
			img_coord_o	  : out img_coord_t;
			
			data_s2_pos_i : in  s2_pos_t;
			data_lsum_o	  : out std_logic_vector(D_C+4-1 downto 0)	-- "σz(t)" (Local sum)
		);
	end component local_sum;

	component sample_store is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G  : std_logic_vector(1 downto 0)
		);
		port (
			clock_i		  : in  std_logic;
			reset_i		  : in  std_logic;

			enable_i	  : in  std_logic;
			enable_o	  : out std_logic;
			img_coord_i	  : in  img_coord_t;
			img_coord_o	  : out img_coord_t;
			
			data_s2_i	  : in  std_logic_vector(D_C-1 downto 0);
			data_s2_pos_o : out s2_pos_t
		);
	end component sample_store;

	component shift_register is
		generic (
			-- 00: BSQ order, 01: BIP order, 10: BIL order
			SMPL_ORDER_G : std_logic_vector(1 downto 0);
			DATA_SIZE_G	 : integer;
			REG_SIZE_G	 : integer;
			INIT_VAL_G	 : std_logic_vector(DATA_SIZE_G-1 downto 0)
		);
		port (
			clock_i		: in  std_logic;
			reset_i		: in  std_logic;

			enable_i	: in  std_logic;
			enable_o	: out std_logic;
			img_coord_i	: in  img_coord_t;
			img_coord_o	: out img_coord_t;

			data_i		: in  std_logic_vector(DATA_SIZE_G-1 downto 0);
			data_o		: out std_logic_vector(DATA_SIZE_G-1 downto 0)
		);
	end component shift_register;

end package comp_predictor;