library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;

-- Package Declaration Section
package utils_predictor is

	pure function mod_R(modR_sgd : in signed; R_int : in integer) return signed;
	pure function mod_R(modR_usg : in unsigned; R_int : in integer) return unsigned;

	pure function sgn(sgn_in : in signed) return integer;
	pure function sgnp(sgn_in : in signed) return integer;

	pure function clip(int_in : in integer; min_int_in : in integer; max_int_in : in integer) return integer;
	pure function clip(sgd_in : in signed; min_sgd_in : in signed; max_sgd_in : in signed) return signed;
	pure function clip(usg_in : in unsigned; min_usg_in : in unsigned; max_usg_in : in unsigned) return unsigned;
	
	pure function shift_sgn(sgd_in : in signed; pos_in : in integer) return signed;
	pure function shift_usg(usg_in : in unsigned; pos_in : in integer) return unsigned;

	pure function vector_product(arr1_sgd : in array_signed_t; arr2_sgd : in array_signed_t; pz_in : in integer) return signed;
	pure function vector_product(arr1_usg : in array_unsigned_t; arr2_usg : in array_unsigned_t; pz_in : in integer) return unsigned;

	pure function reset_s2_pos return s2_pos_t;
	pure function reset_ldiff_pos return ldiff_pos_t;

end package utils_predictor;

-- Package Body Section
package body utils_predictor is
	
	-- Modulus*R function for "signed" signals
	pure function mod_R(modR_sgd : in signed; R_int : in integer) return signed is	
		constant max_size_v	: integer						:= max(modR_sgd'length+1, R_int+3);
		constant power0_v	: signed(max_size_v-1 downto 0)	:= (R_int+0 => '1', others => '0');
		constant power1_v	: signed(max_size_v-1 downto 0)	:= (R_int-1 => '1', others => '0');
		variable parent_v	: signed(max_size_v-1 downto 0) := (others => '0');
		variable modulus_v	: signed(max_size_v-1 downto 0) := (others => '0');
	begin	-- If working with big positive numbers in "signed", array should has by default more bits than necessary, so that the MSB will never be 1 (and so, the number is never read as negative)
		parent_v  := resize(resize(modR_sgd, max_size_v) + power1_v, max_size_v);
		modulus_v := resize(parent_v mod power0_v, max_size_v);

		return resize(modulus_v - power1_v, R_int);
	end function mod_R;

	-- Modulus*R function for "unsigned" signals
	pure function mod_R(modR_usg : in unsigned; R_int : in integer) return unsigned is
		constant max_size_v	: integer						  := max(modR_usg'length+1, R_int+3);
		constant power0_v	: unsigned(max_size_v-1 downto 0) := (R_int+0 => '1', others => '0');
		constant power1_v	: unsigned(max_size_v-1 downto 0) := (R_int-1 => '1', others => '0');
		variable parent_v	: unsigned(max_size_v-1 downto 0) := (others => '0');
		variable modulus_v	: unsigned(max_size_v-1 downto 0) := (others => '0');
	begin
		parent_v  := resize(resize(modR_usg, max_size_v) + power1_v, max_size_v);
		modulus_v := resize(parent_v mod power0_v, max_size_v);

		return resize(modulus_v - power1_v, R_int);
	end function mod_R;
	
	-- Sign function for "signed" signals
	pure function sgn(sgn_in : in signed) return integer is
		variable zeros_v : signed(sgn_in'length-1 downto 0) := (others => '0');
	begin
		if (sgn_in > zeros_v) then
			return 1;
		elsif (sgn_in = zeros_v) then
			return 0;
		else
			return -1;
		end if;
	end function sgn;
	
	-- Sign plus function for "signed" signals
	pure function sgnp(sgn_in : in signed) return integer is
		variable zeros_v : signed(sgn_in'length-1 downto 0) := (others => '0');
	begin
		if (sgn_in >= zeros_v) then
			return 1;
		else
			return -1;
		end if;
	end function sgnp;

	-- Clipping function for "integer" signals
	pure function clip(int_in : in integer; min_int_in : in integer; max_int_in : in integer) return integer is
	begin
		if (int_in < min_int_in) then
			return min_int_in;
		elsif (int_in > max_int_in) then
			return max_int_in;
		else
			return int_in;
		end if;
	end function clip;
	
	-- Clipping function for "signed" signals
	pure function clip(sgd_in : in signed; min_sgd_in : in signed; max_sgd_in : in signed) return signed is
		variable max_width_v : integer := max_sgd_in'length;	-- This input value already has the longest width
	begin
		if (resize(sgd_in, max_width_v) < resize(min_sgd_in, max_width_v)) then
			return resize(min_sgd_in, max_width_v);
		elsif (resize(sgd_in, max_width_v) > resize(max_sgd_in, max_width_v)) then
			return resize(max_sgd_in, max_width_v);
		else
			return resize(sgd_in, max_width_v);
		end if;
	end function clip;
	
	-- Clipping function for "unsigned" signals
	pure function clip(usg_in : in unsigned; min_usg_in : in unsigned; max_usg_in : in unsigned) return unsigned is
		variable max_width_v : integer := max_usg_in'length;	-- This input value already has the longest width
	begin
		if (resize(usg_in, max_width_v) < resize(min_usg_in, max_width_v)) then
			return resize(min_usg_in, max_width_v);
		elsif (resize(usg_in, max_width_v) > resize(max_usg_in, max_width_v)) then
			return resize(max_usg_in, max_width_v);
		else
			return resize(usg_in, max_width_v);
		end if;
	end function clip;
	
	-- Shifting function (to either left or right) for "signed" signals
	pure function shift_sgn(sgd_in : in signed; pos_in : in integer) return signed is
		constant ZERO_C			: signed(sgd_in'length-1 downto 0) := (others => '0');
		variable shifted_sgn_v	: signed(sgd_in'length-1 downto 0) := (others => '0');
		variable abs_pos_v		: integer := abs(pos_in);
	begin
		if (sgd_in /= ZERO_C) then
			if (pos_in > 0) then		-- Positive value, so it shifts to the left
				shifted_sgn_v := resize(sgd_in sll abs_pos_v, sgd_in'length);
			elsif (pos_in < 0) then		-- Negative value, so it shifts to the right
				shifted_sgn_v := resize(sgd_in srl abs_pos_v, sgd_in'length);
				-- If MSb = '1', the sign extension must be done manually with sll/srl functions...
				if (sgd_in(sgd_in'length-1) = '1') then
					shifted_sgn_v(sgd_in'length-1 downto sgd_in'length-abs_pos_v) := (others => '1');
				end if;
			else	-- pos_in = 0		-- No need to do anything here
				shifted_sgn_v := sgd_in;
			end if;
		else	-- sgd_in = ZERO_C
			shifted_sgn_v := sgd_in;
		end if;

		return shifted_sgn_v;
	end function shift_sgn;
	
	-- Shifting function (to either left or right) for "unsigned" signals
	pure function shift_usg(usg_in : in unsigned; pos_in : in integer) return unsigned is
		constant ZERO_C			: unsigned(usg_in'length-1 downto 0) := (others => '0');
		variable shifted_usg_v	: unsigned(usg_in'length-1 downto 0) := (others => '0');
		variable abs_pos_v		: integer := abs(pos_in);
	begin
		if (usg_in /= ZERO_C) then
			if (pos_in > 0) then		-- Positive value, so shifts to the left
				shifted_usg_v := resize(usg_in sll abs_pos_v, usg_in'length);
			elsif (pos_in < 0) then		-- Negative value, so shifts to the right
				shifted_usg_v := resize(usg_in srl abs_pos_v, usg_in'length);
			else	-- pos_in = 0		-- No need to do anything here
				shifted_usg_v := usg_in;
			end if;
		else	-- usg_in = ZERO_C
			shifted_usg_v := usg_in;
		end if;

		return shifted_usg_v;
	end function shift_usg;
	
	-- Vector inner product function for "signed" signals
	pure function vector_product(arr1_sgd : in array_signed_t; arr2_sgd : in array_signed_t; pz_in : in integer) return signed is
		-- The product signal length should be an addition of the 2 input signal lengths...
		variable product_width_v : integer := arr1_sgd(0)'length + arr2_sgd(0)'length;		-- Result: OMEGA_C+3 + D_C
		variable total_width_v	 : integer := product_width_v + arr1_sgd'length;			-- Result: OMEGA_C+3 + D_C + MAX_CZ_C
		variable product_v		 : signed(product_width_v-1 downto 0) := (others => '0');
		variable result_v		 : signed(total_width_v-1 downto 0)	  := (others => '0');
	begin
		-- Direct. local diff. are always taken into account here, but they will ONLY affect under "Full Prediction Mode", because such positions are always '0' under "Reduced Prediction Mode"
		for i in 0 to (MAX_CZ_C-1) loop
			if (i < pz_in+3) then
				product_v := resize(resize(arr1_sgd(i), product_width_v) * resize(arr2_sgd(i), product_width_v), product_width_v);
				result_v  := resize(result_v + resize(product_v, total_width_v), total_width_v);
			else
				exit;
			end if;
		end loop;

		return result_v;
	end function vector_product;

	-- Vector inner product function for "unsigned" signals
	pure function vector_product(arr1_usg : in array_unsigned_t; arr2_usg : in array_unsigned_t; pz_in : in integer) return unsigned is
		-- The product signal length should be an addition of the 2 input signal lengths...
		variable product_width_v : integer := arr1_usg(0)'length + arr2_usg(0)'length;		-- Result: OMEGA_C+3 + D_C
		variable total_width_v	 : integer := product_width_v + arr1_usg'length;			-- Result: OMEGA_C+3 + D_C + MAX_CZ_C
		variable product_v		 : unsigned(product_width_v-1 downto 0) := (others => '0');
		variable result_v		 : unsigned(total_width_v-1 downto 0)	:= (others => '0');
	begin
		-- Direct. local diff. are always taken into account here, but they will ONLY affect under "Full Prediction Mode", because such positions are always '0' under "Reduced Prediction Mode"
		for i in 0 to (MAX_CZ_C-1) loop
			if (i < pz_in+3) then
				product_v := resize(resize(arr1_usg(i), product_width_v) * resize(arr2_usg(i), product_width_v), product_width_v);
				result_v  := resize(result_v + resize(product_v, total_width_v), total_width_v);
			else
				exit;
			end if;
		end loop;

		return result_v;
	end function vector_product;
	
	-- Resets the s2 positions record
	pure function reset_s2_pos return s2_pos_t is
		variable s2_pos_v : s2_pos_t;
	begin
		s2_pos_v.cur := (others => '0');
		s2_pos_v.w	 := (others => '0');
		s2_pos_v.wz  := (others => '0');
		s2_pos_v.n	 := (others => '0');
		s2_pos_v.nw  := (others => '0');
		s2_pos_v.ne  := (others => '0');
		
		return s2_pos_v;
	end function reset_s2_pos;

	-- Resets the local differences positions record
	pure function reset_ldiff_pos return ldiff_pos_t is
		variable ldiff_pos_v : ldiff_pos_t;
	begin
		ldiff_pos_v.c  := (others => '0');
		ldiff_pos_v.n  := (others => '0');
		ldiff_pos_v.w  := (others => '0');
		ldiff_pos_v.nw := (others => '0');
		
		return ldiff_pos_v;
	end function reset_ldiff_pos;	

end package body utils_predictor;