--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		30/10/2020
--------------------------------------------------------------------------------
-- IP name:		adder
--
-- Description: Returns the difference between two elements
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

entity adder is
	generic (
		-- 0: "signed" type samples, 1: "unsigned" type samples
		SMPL_TYPE_G	  : std_logic;
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G  : std_logic_vector(1 downto 0)
	);
	port (
		clock_i		  : in  std_logic;
		reset_i		  : in  std_logic;
		
		img_en_i	  : in  std_logic;
		enable_o	  : out std_logic;
		img_coord_i	  : in  img_coord_t;
		img_coord_o	  : out img_coord_t;
		
		pred_en_i	  : in  std_logic;
		pred_coord_i  : in  img_coord_t;
		data_s3_i	  : in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
		data_s3_o	  : out std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
		data_s4_lsb_i : in  std_logic;							-- "s~z(t)"  (double-resolution predicted sample) LSb
		data_s4_lsb_o : out std_logic;							-- "s~z(t)"  (double-resolution predicted sample) LSb
		
		data_s0_i	  : in  std_logic_vector(D_C-1 downto 0);	-- "sz(t)"  (original sample)
		en_res_o	  : out std_logic;
		data_res_o	  : out std_logic_vector(D_C+2-1 downto 0)	-- "/\z(t)" (prediction residual)
	);
end entity adder;

architecture behavioural of adder is
	-- Constants declaration
	constant IMG_SIZE_C		: integer := NX_C*NY_C*NZ_C;
	
	-- Signals declaration
	signal enable_s			: std_logic											:= '0';
	signal img_coord_s		: img_coord_t										:= reset_img_coord(SMPL_ORDER_G);
	
	signal img_cnt_s		: integer range 0 to IMG_SIZE_C-1					:= 0;
	signal pred_cnt_s		: integer range 0 to IMG_SIZE_C-1					:= 0;
	signal final_cnt_s		: integer range 0 to IMG_SIZE_C-1					:= 0;
	signal img_end_flag_s	: std_logic											:= '0';
	signal pred_end_flag_s	: std_logic											:= '0';
	
	signal data_s3_s		: std_logic_vector(D_C+1-1 downto 0)				:= (others => '0');
	signal data_s4_lsb_s	: std_logic											:= '0';
	signal en_res_s			: std_logic											:= '0';
	signal data_res_s		: std_logic_vector(D_C+2-1 downto 0)				:= (others => '0');
	
	signal data_s4_lsb_arr_s: std_logic_vector(IMG_SIZE_C-1 downto 0)			:= (others => '0');
	signal data_s0_arr_s	: array_slv_t(0 to IMG_SIZE_C-1)(D_C-1 downto 0)	:= (others => (others => '0'));
	signal data_s3_arr_s	: array_slv_t(0 to IMG_SIZE_C-1)(D_C+1-1 downto 0)	:= (others => (others => '0'));
	signal img_coord_arr_s	: array_img_coord_t(0 to IMG_SIZE_C-1)				:= (others => reset_img_coord(SMPL_ORDER_G));

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_en_coord_upd : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s <= reset_img_coord(SMPL_ORDER_G);
			else
				if (img_end_flag_s = '0') then
					enable_s	<= img_en_i;
					img_coord_s	<= img_coord_i;
				elsif (img_end_flag_s = '1' and pred_end_flag_s = '1') then
					enable_s	<= '1';
					img_coord_s	<= img_coord_arr_s(final_cnt_s);
				else
					enable_s	<= '0';
					img_coord_s	<= reset_img_coord(SMPL_ORDER_G);
				end if;
			end if;
		end if;
	end process p_en_coord_upd;
	
	-- Monitors and saves both original input and predicted sample values
	p_cnt_flags : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				img_cnt_s		  <= 0;
				pred_cnt_s		  <= 0;
				final_cnt_s		  <= 0;
				img_end_flag_s	  <= '0';
				pred_end_flag_s	  <= '0';
				data_s4_lsb_arr_s <= (others => '0');
				data_s0_arr_s	  <= (others => (others => '0'));
				data_s3_arr_s	  <= (others => (others => '0'));
				img_coord_arr_s	  <= (others => reset_img_coord(SMPL_ORDER_G));
			else
				-- Monitors and saves incoming original input samples (+ its coordinates)
				if (img_en_i = '1' and img_end_flag_s = '0') then
					data_s0_arr_s(img_cnt_s)	<= data_s0_i;
					img_coord_arr_s(img_cnt_s)	<= img_coord_i;
					if (img_cnt_s < IMG_SIZE_C-1) then
						img_cnt_s <= img_cnt_s + 1;
					else
						img_cnt_s <= 0;
						img_end_flag_s <= '1';
					end if;
				end if;
				
				-- Monitors and saves incoming predicted samples
				if (pred_en_i = '1' and pred_end_flag_s = '0') then
					data_s3_arr_s(pred_cnt_s)	<= data_s3_i;
					if (pred_cnt_s < IMG_SIZE_C-1) then
						pred_cnt_s <= pred_cnt_s + 1;
					else
						pred_cnt_s <= 0;
						pred_end_flag_s <= '1';
					end if;
					data_s4_lsb_arr_s(pred_cnt_s) <= data_s4_lsb_i;
				end if;
				
				-- Monitors and waits until residual values are sent out
				if (img_end_flag_s = '1' and pred_end_flag_s = '1') then
					if (final_cnt_s < IMG_SIZE_C-1) then
						final_cnt_s <= final_cnt_s + 1;
					else
						final_cnt_s <= 0;
						img_end_flag_s  <= '0';
						pred_end_flag_s <= '0';
					end if;
				end if;
			end if;
		end if;
	end process p_cnt_flags;
	
	-- Prediction residual calculation
	p_pred_res_calc : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				data_s3_s	  <= (others => '0');
				data_s4_lsb_s <= '0';
				en_res_s	  <= '0';
				data_res_s	  <= (others => '0');
			else
				if (img_en_i = '1' or pred_en_i = '1' or (img_end_flag_s = '1' and pred_end_flag_s = '1')) then
					if (img_end_flag_s = '0' or pred_end_flag_s = '0') then
						if (img_end_flag_s = '0') then
							if (SMPL_TYPE_G = '0') then			-- Incoming data is recognized as "signed"
								data_res_s <= std_logic_vector(resize(resize(signed(data_s0_i), D_C+3) - resize(signed(data_s3_i), D_C+3), D_C+2));
							else	-- SMPL_TYPE_G = '1'		-- Incoming data is recognized as "unsigned"
								data_res_s <= std_logic_vector(resize(signed(resize(unsigned(data_s0_i), D_C+3)) - resize(signed(data_s3_i), D_C+3), D_C+2));
							end if;
						end if;
						if (pred_end_flag_s = '0') then
							data_s3_s	<= data_s3_i;
						end if;
					elsif (img_end_flag_s = '1' and pred_end_flag_s = '1') then
						data_s3_s		<= data_s3_arr_s(final_cnt_s);
						data_s4_lsb_s	<= data_s4_lsb_arr_s(final_cnt_s);
						en_res_s		<= '1';
						if (SMPL_TYPE_G = '0') then				-- Incoming data is recognized as "signed"
							data_res_s	<= std_logic_vector(resize(resize(signed(data_s0_arr_s(final_cnt_s)), D_C+3) - resize(signed(data_s3_arr_s(final_cnt_s)), D_C+3), D_C+2));
						else	-- SMPL_TYPE_G = '1'			-- Incoming data is recognized as "unsigned"
							data_res_s	<= std_logic_vector(resize(signed(resize(unsigned(data_s0_arr_s(final_cnt_s)), D_C+3)) - resize(signed(data_s3_arr_s(final_cnt_s)), D_C+3), D_C+2));
						end if;
					end if;
				end if;
			end if;
		end if;
	end process p_pred_res_calc;
	
	-- Outputs
	enable_o	  <= enable_s;
	img_coord_o	  <= img_coord_s;
	data_s3_o	  <= data_s3_s;
	data_s4_lsb_o <= data_s4_lsb_s;
	en_res_o	  <= en_res_s;
	data_res_o	  <= data_res_s;
end architecture behavioural;