--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		30/10/2020
--------------------------------------------------------------------------------
-- IP name:		scaled_diff
--
-- Description: Computes the scaled difference between sˆz(t) and the nearest
--				endpoint
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;

entity scaled_diff is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	: std_logic_vector(1 downto 0);
		-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
		SMPL_LIMIT_G	: smpl_lim_t
	);
	port (
		clock_i			: in  std_logic;
		reset_i			: in  std_logic;
		
		enable_i		: in  std_logic;
		enable_o		: out std_logic;
		img_coord_i		: in  img_coord_t;
		img_coord_o		: out img_coord_t;
		
		data_s3_i		: in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
		data_merr_i		: in  std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)"  (maximum error)
		scaled_diff_o	: out std_logic_vector(D_C+1-1 downto 0)	-- "θz(t)"  (scaled difference)
	);
end entity scaled_diff;

architecture behavioural of scaled_diff is
	signal enable_s		 : std_logic := '0';
	signal img_coord_s	 : img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal scaled_diff_s : std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_scaled_diff_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s <= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_scaled_diff_delay;
	
	-- Scaled difference value (θz(t)) calculation	
	p_sc_diff_calc : process(clock_i) is
		variable comp1_v, comp2_v, comp3_v, comp4_v, comp5_v : std_logic_vector(D_C+3-1 downto 0) := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				comp1_v := (others => '0');
				comp2_v := (others => '0');
				comp3_v := (others => '0');
				comp4_v := (others => '0');
				comp5_v := (others => '0');
				scaled_diff_s <= (others => '0');
			else
				if (enable_i = '1') then
					if (img_coord_i.t = 0) then
						comp1_v := std_logic_vector(resize(resize(signed(data_s3_i), D_C+3) - resize(signed(SMPL_LIMIT_G.min), D_C+3), D_C+3));
						comp2_v := std_logic_vector(resize(resize(signed(SMPL_LIMIT_G.max), D_C+3) - resize(signed(data_s3_i), D_C+3), D_C+3));
						scaled_diff_s <= std_logic_vector(resize(work.utils_image.min(signed(comp1_v), signed(comp2_v)), D_C+1));
					else	-- img_coord_i.t > 0
						comp3_v := std_logic_vector(resize(resize(signed(data_s3_i), D_C+3) - resize(signed(SMPL_LIMIT_G.min), D_C+3) + resize(signed(data_merr_i), D_C+3), D_C+3));
						comp4_v := std_logic_vector(resize(resize(signed(SMPL_LIMIT_G.max), D_C+3) - resize(signed(data_s3_i), D_C+3) + resize(signed(data_merr_i), D_C+3), D_C+3));
						comp5_v := std_logic_vector(resize(shift_sgn(resize(signed(data_merr_i), D_C+4), 1) + to_signed(1, D_C+4), D_C+3));
						scaled_diff_s <= std_logic_vector(resize(work.utils_image.min(division_down(signed(comp3_v), signed(comp5_v)), division_down(signed(comp4_v), signed(comp5_v))), D_C+1));
					end if;
				end if;
			end if;
		end if;
	end process p_sc_diff_calc;

	-- Outputs
	enable_o		<= enable_s;
	img_coord_o		<= img_coord_s;
	scaled_diff_o	<= scaled_diff_s;
end architecture behavioural;