--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		30/10/2020
--------------------------------------------------------------------------------
-- IP name:		mapper
--
-- Description: Maps the incoming quantized prediction residual 
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.types_predictor.all;
use work.utils_predictor.all;
use work.comp_predictor.all;

entity mapper is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	: std_logic_vector(1 downto 0);
		-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
		SMPL_LIMIT_G	: smpl_lim_t
	);
	port (
		clock_i			: in  std_logic;
		reset_i			: in  std_logic;
		
		enable_i		: in  std_logic;
		enable_o		: out std_logic;
		img_coord_i		: in  img_coord_t;
		img_coord_o		: out img_coord_t;
		
		data_s4_lsb_i	: in  std_logic;							-- "s~z(t)" (double-resolution predicted sample) LSb
		data_s3_i		: in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
		data_merr_i		: in  std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)"  (maximum error)
		data_quant_i	: in  std_logic_vector(D_C+2-1 downto 0);	-- "qz(t)"  (quantizer index)
		data_mp_quan_o	: out std_logic_vector(D_C-1 downto 0)		-- "δz(t)"  (mapped quantizer index)
	);
end entity mapper;

architecture behavioural of mapper is	
	constant ZEROS_C	  : std_logic_vector(D_C+2-1 downto 0) := (others => '0');
	
	signal enable1_s	  : std_logic	:= '0';
	signal enable2_s	  : std_logic	:= '0';
	signal img_coord1_s	  : img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord2_s	  : img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal data_s4_lsb_s  : std_logic	:= '0';
	signal data_quant_s	  : std_logic_vector(D_C+2-1 downto 0)	:= (others => '0');
	signal scaled_diff_s : std_logic_vector(D_C+1-1 downto 0)	:= (others => '0');
	signal data_mp_quan_s : std_logic_vector(D_C-1 downto 0)	:= (others => '0');
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_mapper_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable2_s	  <= '0';
				img_coord2_s  <= reset_img_coord(SMPL_ORDER_G);
				data_s4_lsb_s <= '0';
				data_quant_s  <= (others => '0');
			else
				enable2_s	  <= enable1_s;
				img_coord2_s  <= img_coord1_s;
				data_s4_lsb_s <= data_s4_lsb_i;
				data_quant_s  <= data_quant_i;
			end if;
		end if;
	end process p_mapper_delay;
	
	-- Scaled difference "θz(t)" calculation
	i_scaled_diff : scaled_diff
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		SMPL_LIMIT_G	=> SMPL_LIMIT_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable_i,
		enable_o		=> enable1_s,
		img_coord_i		=> img_coord_i,
		img_coord_o		=> img_coord1_s,
		
		data_s3_i		=> data_s3_i,
		data_merr_i		=> data_merr_i,
		scaled_diff_o	=> scaled_diff_s
	);

	-- Mapped quantizer index value "δz(t)" calculation	
	p_mp_quan_calc : process(clock_i) is
		variable data_quant_usg_v  : unsigned(D_C+4-1 downto 0) := (others => '0');
		variable scaled_diff_usg_v : unsigned(D_C+4-1 downto 0) := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				data_quant_usg_v  := (others => '0');
				scaled_diff_usg_v := (others => '0');
				data_mp_quan_s	  <= (others => '0');
			else
				if (enable1_s = '1') then
					data_quant_usg_v  := unsigned(abs(resize(signed(data_quant_s), D_C+4)));
					scaled_diff_usg_v := unsigned(resize(signed(scaled_diff_s), D_C+4));

					if (data_quant_usg_v > scaled_diff_usg_v) then
						data_mp_quan_s <= std_logic_vector(resize(data_quant_usg_v + scaled_diff_usg_v, D_C));
					elsif ((data_s4_lsb_s = '0' and (signed(data_quant_s) >= signed(ZEROS_C))) or (data_s4_lsb_s = '1' and (signed(data_quant_s) <= signed(ZEROS_C)))) then
						data_mp_quan_s <= std_logic_vector(resize(shift_usg(data_quant_usg_v, 1), D_C));
					else
						data_mp_quan_s <= std_logic_vector(resize(shift_usg(data_quant_usg_v, 1) - to_unsigned(1, D_C+4), D_C));
					end if;
				end if;
			end if;
		end if;
	end process p_mp_quan_calc;

	-- Outputs
	enable_o		<= enable2_s;
	img_coord_o		<= img_coord2_s;
	data_mp_quan_o	<= data_mp_quan_s;
end architecture behavioural;