--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		24/10/2020
--------------------------------------------------------------------------------
-- IP name:		local_sum
--
-- Description: Computes the local sum
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.types_predictor.all;
use work.utils_predictor.all;

entity local_sum is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G  : std_logic_vector(1 downto 0);
		-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
		SMPL_LIMIT_G  : smpl_lim_t;
		-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
		LSUM_TYPE_G	  : std_logic_vector(1 downto 0)
	);
	port (
		clock_i		  : in  std_logic;
		reset_i		  : in  std_logic;
		
		enable_i	  : in  std_logic;
		enable_o	  : out std_logic;
		img_coord_i	  : in  img_coord_t;
		img_coord_o	  : out img_coord_t;
		
		data_s2_pos_i : in  s2_pos_t;
		data_lsum_o	  : out std_logic_vector(D_C+4-1 downto 0)	-- "σz(t)" (Local sum)
	);
end entity local_sum;

architecture behavioural of local_sum is
	signal enable_s		: std_logic   := '0';
	signal img_coord_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal data_lsum_s	: std_logic_vector(D_C+4-1 downto 0) := (others => '0');

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_lsum_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s <= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_lsum_delay;

	-- Local sum (σz(t)) calculation
	g_lsum_calc : case LSUM_TYPE_G generate
		when WID_NEI_C =>	-- Wide neighbour-oriented case
			p_lsum_wi_ne_calc : process(clock_i) is
			begin
				if rising_edge(clock_i) then
					if (reset_i = '1') then
						data_lsum_s <= (others => '0');
					else
						if (enable_i = '1') then
							if (img_coord_i.y > 0 and img_coord_i.x > 0 and img_coord_i.x < NX_C-1) then
								data_lsum_s <= std_logic_vector(resize(resize(signed(data_s2_pos_i.w), D_C+4) + resize(signed(data_s2_pos_i.nw), D_C+4) + resize(signed(data_s2_pos_i.n), D_C+4) + resize(signed(data_s2_pos_i.ne), D_C+4), D_C+4));
							elsif (img_coord_i.y = 0 and img_coord_i.x > 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.w), D_C+4), 2), D_C+4));
							elsif (img_coord_i.y > 0 and img_coord_i.x = 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn((resize(signed(data_s2_pos_i.n), D_C+4) + resize(signed(data_s2_pos_i.ne), D_C+4)), 1), D_C+4));
							elsif (img_coord_i.y > 0 and img_coord_i.x = NX_C-1) then
								data_lsum_s <= std_logic_vector(resize(resize(signed(data_s2_pos_i.w), D_C+4) + resize(signed(data_s2_pos_i.nw), D_C+4) + shift_sgn(resize(signed(data_s2_pos_i.n), D_C+4), 1), D_C+4));
							else		-- Case X=0 and Y=0 here defined, but anyway not used later on...
								-- data_lsum_s <= (others => '0');
							end if;
						end if;
					end if;
				end if;
			end process p_lsum_wi_ne_calc;

		when NAR_NEI_C =>	-- Narrow neighbour-oriented case
			p_lsum_na_ne_calc : process(clock_i) is
			begin
				if rising_edge(clock_i) then
					if (reset_i = '1') then
						data_lsum_s <= (others => '0');
					else
						if (enable_i = '1') then
							if (img_coord_i.y > 0 and img_coord_i.x > 0 and img_coord_i.x < NX_C-1) then
								data_lsum_s <= std_logic_vector(resize(resize(signed(data_s2_pos_i.nw), D_C+4) + shift_sgn(resize(signed(data_s2_pos_i.n), D_C+4), 1) + resize(signed(data_s2_pos_i.ne), D_C+4), D_C+4));
							elsif (img_coord_i.y = 0 and img_coord_i.x > 0 and img_coord_i.z > 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.wz), D_C+4), 2), D_C+4));
							elsif (img_coord_i.y > 0 and img_coord_i.x = 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(resize(signed(data_s2_pos_i.n), D_C+4) + resize(signed(data_s2_pos_i.ne), D_C+4), D_C+4), 1), D_C+4));
							elsif (img_coord_i.y > 0 and img_coord_i.x = NX_C-1) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(resize(signed(data_s2_pos_i.nw), D_C+4) + resize(signed(data_s2_pos_i.n), D_C+4), D_C+4), 1), D_C+4));
							elsif (img_coord_i.y = 0 and img_coord_i.x > 0 and img_coord_i.z = 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(signed(SMPL_LIMIT_G.mid), D_C+4), 2), D_C+4));
							else		-- Case X=0 and Y=0 here defined, but anyway not used later on...
								-- data_lsum_s <= (others => '0');
							end if;
						end if;
					end if;
				end if;
			end process p_lsum_na_ne_calc;

		when WID_COL_C =>	-- Wide column-oriented case
			p_lsum_wi_co_calc : process(clock_i) is
			begin
				if rising_edge(clock_i) then
					if (reset_i = '1') then
						data_lsum_s <= (others => '0');
					else
						if (enable_i = '1') then
							if (img_coord_i.y > 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.n), D_C+4), 2), D_C+4));
							elsif (img_coord_i.y = 0 and img_coord_i.x > 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.w), D_C+4), 2), D_C+4));
							else		-- Case X=0 and Y=0 here defined, but anyway not used later on...
								-- data_lsum_s <= (others => '0');
							end if;
						end if;
					end if;
				end if;
			end process p_lsum_wi_co_calc;
			
		when others =>	-- Narrow column-oriented case (when "NAR_COL_C")
			p_lsum_na_co_calc : process(clock_i) is
			begin
				if rising_edge(clock_i) then
					if (reset_i = '1') then
						data_lsum_s <= (others => '0');
					else
						if (enable_i = '1') then
							if (img_coord_i.y > 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.n), D_C+4), 2), D_C+4));
							elsif (img_coord_i.y = 0 and img_coord_i.x > 0 and img_coord_i.z > 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.wz), D_C+4), 2), D_C+4));
							elsif (img_coord_i.y = 0 and img_coord_i.x > 0 and img_coord_i.z = 0) then
								data_lsum_s <= std_logic_vector(resize(shift_sgn(resize(signed(SMPL_LIMIT_G.mid), D_C+4), 2), D_C+4));
							else		-- Case X=0 and Y=0 here defined, but anyway not used later on...
								-- data_lsum_s <= (others => '0');
							end if;
						end if;
					end if;
				end if;
			end process p_lsum_na_co_calc;
	end generate g_lsum_calc;

	-- Outputs
	enable_o	<= enable_s;
	img_coord_o	<= img_coord_s;
	data_lsum_o	<= data_lsum_s;
end architecture behavioural;