--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		01/11/2020
--------------------------------------------------------------------------------
-- IP name:		pred_ctrl_local_diff
--
-- Description: Computes the predicted central local difference "d^z(t)"
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;

entity pred_ctrl_local_diff is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	: std_logic_vector(1 downto 0);
		-- 1: Full prediction mode, 0: Reduced prediction mode
		PREDICT_MODE_G	: std_logic
	);
	port (
		clock_i			: in  std_logic;
		reset_i			: in  std_logic;
		
		enable_i		: in  std_logic;
		enable_o		: out std_logic;
		img_coord_i		: in  img_coord_t;
		img_coord_o		: out img_coord_t;
		
		pz_i			: in  integer;
		ldiff_vect_i	: in  array_slv_t(0 to MAX_CZ_C-1)(D_C+4-1 downto 0);		 -- "Uz(t)"  (local difference vector)
		weight_vect_i	: in  array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0);	 -- "Wz(t)"  (weight vector)
		
		data_pred_cldiff_o : out std_logic_vector(MAX_CZ_C+D_C+4+OMEGA_C+3-1 downto 0) -- "d^z(t)" (predicted central local difference)
	);
end entity pred_ctrl_local_diff;

architecture behavioural of pred_ctrl_local_diff is
	signal enable_s			  : std_logic := '0';
	signal img_coord_s		  : img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	-- The maximum number of previous spectral bands for prediction (MAX_PZ_C+3) are calculated, but only some of them (until "pz_i") will be used later
	signal data_pred_cldiff_s : std_logic_vector(MAX_CZ_C+D_C+4+OMEGA_C+3-1 downto 0) := (others => '0');
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_pred_cldiff_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	 <= '0';
				img_coord_s	 <= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	 <= enable_i;
				img_coord_s	 <= img_coord_i;
			end if;
		end if;
	end process p_pred_cldiff_delay;
	
	-- Predicted central local difference (d^z(t)) calculation	
	p_pred_cldiff_calc : process(clock_i) is
		variable ldiff_vect_v  : array_signed_t(0 to MAX_CZ_C-1)(D_C+4-1 downto 0)		:= (others => (others => '0'));
		variable weight_vect_v : array_signed_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0)	:= (others => (others => '0'));
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				ldiff_vect_v	   := (others => (others => '0'));
				weight_vect_v	   := (others => (others => '0'));
				data_pred_cldiff_s <= (others => '0');
			else
				if (enable_i = '1') then
					-- Either when t=0 or (z=0 and "Reduced prediction mode"), the output is set to 0
					if ((img_coord_i.t = 0) or (img_coord_i.z = 0 and PREDICT_MODE_G = '0')) then
						data_pred_cldiff_s <= (others => '0');
					else
						-- Input arrays are first transformed into the necessary format (either "signed" or "unsigned")
						ldiff_vect_v  := arr_slv_2_arr_sgn(ldiff_vect_i);
						weight_vect_v := arr_slv_2_arr_sgn(weight_vect_i);

						-- Direct. local differences (positions 0 to 2) are always taken into account inside this function, but they will ONLY affect
						-- under "Full Prediction Mode", because such positions are previously set to 0 if working under "Reduced Prediction Mode"
						data_pred_cldiff_s <= std_logic_vector(vector_product(ldiff_vect_v, weight_vect_v, pz_i));
					end if;
				end if;
			end if;
		end if;
	end process p_pred_cldiff_calc;

	-- Outputs
	enable_o			<= enable_s;
	img_coord_o			<= img_coord_s;
	data_pred_cldiff_o	<= data_pred_cldiff_s;
end architecture behavioural;