--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		24/10/2020
--------------------------------------------------------------------------------
-- IP name:		sample_store
--
-- Description: Given the current sample, it uses shift registers to provide
--				some specifc older samples.
--
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--			   				   								   _____
-- ... | ... | ... | ... | ...			 	 +--> +x		 /|	   /|
-- ----+-----+-----+-----+-----				/|				/_|___/ |
-- ... | NW  |  N  | NE  | ...			   / |				| |___|_|
-- ----+-----+-----+-----+-----			  v	 v				| /   | /
-- ... |  W  | CUR | ... | ...			+z	 +y				|/____|/
-- ----+-----+-----+-----+-----
-- ... | ... | ... | ... | ...			WZ = W, but in the previous spectral band (z-1)
--
--------------------------------------------------------------------------------
-- Depending on input order (BSQ/BIP/BIL), sample distance (delay/reg. size) between current and neighbour samples are:
-- Order | NW			 | N		 | NE			 | W
-- ------+---------------+-----------+---------------+------
-- BSQ	 | NX_C+1		 | NX_C		 | NX_C-1		 | 1
-- BIP	 | (NX_C+1)*NZ_C | NX_C*NZ_C | (NX_C-1)*NZ_C | NZ_C
-- BIL	 | NX_C*NZ_C+1	 | NX_C*NZ_C | NX_C*NZ_C-1	 | 1
--
--------------------------------------------------------------------------------
-- Depending on input order (BSQ/BIP/BIL), sample distance (delay/reg. size) between current sample and same sample in previous bands are:
-- Order | Z-1		 | Z-2		    | ... | z-P_C
-- ------+-----------+--------------+-----+-------------
-- BSQ	 | NX_C*NY_C | 2*NX_C*NY_C	| ... | P_C*NX_C*NY_C
-- BIP	 | 1		 | 2			| ... | P_C
-- BIL	 | NX_C		 | 2*NX_C	    | ... | P_C*NX_C
--
-- NOTE: To get neighbour samples from previous spectral bands, simply ADD the corresponding values from these 2 tables above.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.types_predictor.all;
use work.comp_predictor.all;

entity sample_store is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G  : std_logic_vector(1 downto 0)
	);
	port (
		clock_i		  : in  std_logic;
		reset_i		  : in  std_logic;

		enable_i	  : in  std_logic;
		enable_o	  : out std_logic;
		img_coord_i	  : in  img_coord_t;
		img_coord_o	  : out img_coord_t;
		
		data_s2_i	  : in  std_logic_vector(D_C-1 downto 0);
		data_s2_pos_o : out s2_pos_t
	);
end entity sample_store;

architecture behavioural of sample_store is
	signal enable_s		: std_logic	  := '0';
	signal img_coord_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal s2_cur_s		: std_logic_vector(D_C-1 downto 0) := (others => '0');
	signal s2_w_s		: std_logic_vector(D_C-1 downto 0) := (others => '0');
	signal s2_wz_s		: std_logic_vector(D_C-1 downto 0) := (others => '0');
	signal s2_n_s		: std_logic_vector(D_C-1 downto 0) := (others => '0');
	signal s2_nw_s		: std_logic_vector(D_C-1 downto 0) := (others => '0');
	signal s2_ne_s		: std_logic_vector(D_C-1 downto 0) := (others => '0');
	
	-- Delay for neighbour samples depending on the input order
	constant POS_NW_C	: integer := locate_position(SMPL_ORDER_G, NX_C+1, (NX_C+1)*NZ_C, NX_C*NZ_C+1);
	constant POS_N_C	: integer := locate_position(SMPL_ORDER_G, NX_C, NX_C*NZ_C, NX_C*NZ_C);
	constant POS_NE_C	: integer := locate_position(SMPL_ORDER_G, NX_C-1, (NX_C-1)*NZ_C, NX_C*NZ_C-1);
	constant POS_W_C	: integer := locate_position(SMPL_ORDER_G, 1, NZ_C, 1);
	constant POS_WZ_C	: integer := locate_position(SMPL_ORDER_G, 1+NX_C*NY_C, NZ_C+1, 1+NX_C);

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_sample_store_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s	<= reset_img_coord(SMPL_ORDER_G);
				s2_cur_s	<= (others => '0');
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
				s2_cur_s	<= data_s2_i;
			end if;
		end if;
	end process p_sample_store_delay;

	-- Position "NW" calculation
	i_shift_reg_nw : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => POS_NW_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		=> clock_i,
		reset_i		=> reset_i,
		
		enable_i	=> enable_i,
		enable_o	=> open,
		img_coord_i	=> img_coord_i,
		img_coord_o	=> open,
		
		data_i		=> s2_cur_s,
		data_o		=> s2_nw_s
	);

	-- Position "N" calculation
	i_shift_reg_n : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => POS_N_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		=> clock_i,
		reset_i		=> reset_i,
		
		enable_i	=> enable_i,
		enable_o	=> open,
		img_coord_i	=> img_coord_i,
		img_coord_o	=> open,
		
		data_i		=> s2_cur_s,
		data_o		=> s2_n_s
	);
	
	-- Position "NE" calculation
	i_shift_reg_ne : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => POS_NE_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		=> clock_i,
		reset_i		=> reset_i,
		
		enable_i	=> enable_i,
		enable_o	=> open,
		img_coord_i	=> img_coord_i,
		img_coord_o	=> open,
		
		data_i		=> s2_cur_s,
		data_o		=> s2_ne_s
	);

	-- Position "W" calculation
	i_shift_reg_w : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => POS_W_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		=> clock_i,
		reset_i		=> reset_i,
		
		enable_i	=> enable_i,
		enable_o	=> open,
		img_coord_i	=> img_coord_i,
		img_coord_o	=> open,
		
		data_i		=> s2_cur_s,
		data_o		=> s2_w_s
	);

	-- Position "W" calculation, but delayed 1 spectral band Z
	i_shift_reg_wz : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => POS_WZ_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		=> clock_i,
		reset_i		=> reset_i,
		
		enable_i	=> enable_i,
		enable_o	=> open,
		img_coord_i	=> img_coord_i,
		img_coord_o	=> open,
		
		data_i		=> s2_cur_s,
		data_o		=> s2_wz_s
	);

	-- Outputs
	enable_o	  <= enable_s;
	img_coord_o	  <= img_coord_s;
	data_s2_pos_o <= (
		cur => s2_cur_s,
		nw	=> s2_nw_s,
		n	=> s2_n_s,
		ne	=> s2_ne_s,
		w	=> s2_w_s,
		wz	=> s2_wz_s
	);
end architecture behavioural;