--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		07/11/2020
--------------------------------------------------------------------------------
-- IP name:		prediction
--
-- Description: Computes the predicted sample "s^z(t)"
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;
use work.comp_predictor.all;

entity prediction is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	: std_logic_vector(1 downto 0);
		-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
		SMPL_LIMIT_G	: smpl_lim_t;
		-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
		LSUM_TYPE_G		: std_logic_vector(1 downto 0);
		-- 1: Full prediction mode, 0: Reduced prediction mode
		PREDICT_MODE_G	: std_logic;
		-- 1: Custom weight init, 0: Default weight init
		W_INIT_TYPE_G	: std_logic
	);
	port (
		clock_i		  : in  std_logic;
		reset_i		  : in  std_logic;
		
		enable_i	  : in  std_logic;
		enable_o	  : out std_logic;
		img_coord_i	  : in  img_coord_t;
		img_coord_o	  : out img_coord_t;
		
		pz_i		  : in  integer;
		data_s0_i	  : in  std_logic_vector(D_C-1 downto 0);		-- "sz(t)"   (original sample)
		data_s1_i	  : in  std_logic_vector(D_C-1 downto 0);		-- "s'z(t)"	 (clipped quantizer bin center)
		data_s2_i	  : in  std_logic_vector(D_C-1 downto 0);		-- "s''z(t)" (sample representative)
		
		data_s3_o	  : out std_logic_vector(D_C+1-1 downto 0);		-- "s^z(t)"	 (predicted sample)
		data_s4_lsb_o : out std_logic;								-- "s~z(t)"  (double-resolution predicted sample) LSb
		data_s6_o	  : out std_logic_vector(Re_C-1 downto 0)		-- "s)z(t)"	 (high-resolution predicted sample)
	);
end entity prediction;

architecture behavioural of prediction is
	-- The number of delay samples to get z-1, for BSQ/BIP/BIL orders (delays table defined inside "samples_store" IP)
	constant NEXT_Z_C		 : integer := locate_position(SMPL_ORDER_G, NX_C*NY_C, 1, NX_C);
	-- Maximum delay to define arrays size (which COULD be different from the top IP processing-time)
	constant MAX_DELAY_C	 : integer := 7;

	-- Enable and coordinate signals among all sub-blocks
	signal enable1_s, enable2_s, enable3_s, enable4_s, enable5_s, enable6_s, enable7_s, enable8_s                         : std_logic   := '0';
	signal img_coord1_s, img_coord2_s, img_coord3_s, img_coord4_s, img_coord5_s, img_coord6_s, img_coord7_s, img_coord8_s : img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal pz_arr_s			 : array_integer_t(0 to MAX_DELAY_C-1)						:= (others => 0);
	signal data_s0_arr_s	 : array_slv_t(0 to MAX_DELAY_C-1)(D_C-1 downto 0)			:= (others => (others => '0'));
	signal data_s1_arr_s	 : array_slv_t(0 to MAX_DELAY_C-1)(D_C-1 downto 0)			:= (others => (others => '0'));
	signal data_s6_arr_s	 : array_slv_t(0 to MAX_DELAY_C-1)(Re_C-1 downto 0)			:= (others => (others => '0'));
	signal data_lsum_arr_s	 : array_slv_t(0 to MAX_DELAY_C-1)(D_C+4-1 downto 0)		:= (others => (others => '0'));
	signal data_s2_pos_arr_s : s2_pos_arr_t(0 to MAX_DELAY_C-1)							:= (others => reset_s2_pos);
	signal ldiff_vect_arr_s	 : matrix_slv_t(0 to MAX_DELAY_C-1)(0 to MAX_CZ_C-1)(D_C+4-1 downto 0) := (others => (others => (others => '0')));

	-- For whatever reason, these arrays cannot be used to save the OUTPUT value from an IP (e.g. data_quant_o => data_quant_arr_s(0) FAIL),
	-- so independent signals are used instead, and later assigned to these arrays (e.g. data_quant_o => data_quant_s PASS).

	signal data_s0z1_s			: std_logic_vector(D_C-1 downto 0)						:= (others => '0');		-- s0_z-1(t)
	signal data_s3_s			: std_logic_vector(D_C+1-1 downto 0)					:= (others => '0');
	signal data_s4_s			: std_logic_vector(D_C+2-1 downto 0)					:= (others => '0');
	signal data_s4_lsb_s		: std_logic												:= '0';
	signal data_s6_s			: std_logic_vector(Re_C-1 downto 0)						:= (others => '0');
	signal data_lsum_s			: std_logic_vector(D_C+4-1 downto 0)					:= (others => '0');
	signal data_s2_pos_s		: s2_pos_t												:= reset_s2_pos;
	signal ldiff_pos_s			: ldiff_pos_t											:= reset_ldiff_pos;
	signal pred_err_bit_s		: integer range -1 to 1									:= 0;
	signal w_scal_exp_s			: integer range -127 to 127								:= 0;
	signal ldiff_vect_s			: array_slv_t(0 to MAX_CZ_C-1)(D_C+4-1 downto 0)		:= (others => (others => '0'));
	signal weight_vect_s		: array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0)	:= (others => (others => '0'));
	signal data_pred_cldiff_s	: std_logic_vector(MAX_CZ_C+D_C+4+OMEGA_C+3-1 downto 0) := (others => '0');

begin
	-- Input values delayed to synchronize them with the next modules in chain
	-- NOTE: Not all arrays need to use all positions, but in this way everything is within one single process
	p_prediction_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				data_s4_lsb_s		<= '0';
				pz_arr_s			<= (others => 0);
				data_s0_arr_s		<= (others => (others => '0'));
				data_s1_arr_s		<= (others => (others => '0'));
				data_s6_arr_s		<= (others => (others => '0'));
				data_lsum_arr_s		<= (others => (others => '0'));
				data_s2_pos_arr_s	<= (others => reset_s2_pos);
				ldiff_vect_arr_s	<= (others => (others => (others => '0')));
			else
				data_s4_lsb_s		<= data_s4_s(0);
				pz_arr_s(0)			<= pz_i;
				data_s0_arr_s(0)	<= data_s0_i;
				data_s1_arr_s(0)	<= data_s1_i;
				data_s6_arr_s(0)	<= data_s6_s;
				data_lsum_arr_s(0)	<= data_lsum_s;
				data_s2_pos_arr_s(0)<= data_s2_pos_s;
				ldiff_vect_arr_s(0)	<= ldiff_vect_s;

				for i in 1 to (MAX_DELAY_C-1) loop
					pz_arr_s(i)			<= pz_arr_s(i-1);
					data_s0_arr_s(i)	<= data_s0_arr_s(i-1);
					data_s1_arr_s(i)	<= data_s1_arr_s(i-1);
					data_s6_arr_s(i)	<= data_s6_arr_s(i-1);
					data_lsum_arr_s(i)	<= data_lsum_arr_s(i-1);
					data_s2_pos_arr_s(i)<= data_s2_pos_arr_s(i-1);
					ldiff_vect_arr_s(i)	<= ldiff_vect_arr_s(i-1);
				end loop;
			end if;
		end if;
	end process p_prediction_delay;

	i_smpl_store : sample_store
	generic map(
		SMPL_ORDER_G  => SMPL_ORDER_G
	)
	port map(
		clock_i		  => clock_i,
		reset_i		  => reset_i,
		
		enable_i	  => enable_i,
		enable_o	  => enable1_s,
		img_coord_i	  => img_coord_i,
		img_coord_o	  => img_coord1_s,
		
		data_s2_i	  => data_s2_i,
		data_s2_pos_o => data_s2_pos_s
	);
	
	i_local_sum : local_sum
	generic map(
		SMPL_ORDER_G  => SMPL_ORDER_G,
		SMPL_LIMIT_G  => SMPL_LIMIT_G,
		LSUM_TYPE_G	  => LSUM_TYPE_G
	)
	port map(
		clock_i		  => clock_i,
		reset_i		  => reset_i,
		
		enable_i	  => enable1_s,
		enable_o	  => enable2_s,
		img_coord_i	  => img_coord1_s,
		img_coord_o	  => img_coord2_s,
		
		data_s2_pos_i => data_s2_pos_s,
		data_lsum_o	  => data_lsum_s
	);
	
	i_local_differences : local_differences
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		PREDICT_MODE_G	=> PREDICT_MODE_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable2_s,
		enable_o		=> enable3_s,
		img_coord_i		=> img_coord2_s,
		img_coord_o		=> img_coord3_s,
		
		data_lsum_i		=> data_lsum_s,
		data_s2_pos_i	=> data_s2_pos_arr_s(0),
		ldiff_pos_o		=> ldiff_pos_s
	);

	i_ldiff_vector : local_diff_vector
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		PREDICT_MODE_G	=> PREDICT_MODE_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable3_s,
		enable_o		=> enable4_s,
		img_coord_i		=> img_coord3_s,
		img_coord_o		=> img_coord4_s,
		
		ldiff_pos_i		=> ldiff_pos_s,
		ldiff_vect_o	=> ldiff_vect_s
	);
	
	-- This IP is not enabled until "data_s4_s" is calculated (and so, "data_s1_s" delayed until then too)
	i_dbl_res_pred_err : dbl_res_pred_error
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable7_s,
		enable_o		=> open,
		img_coord_i		=> img_coord7_s,
		img_coord_o		=> open,
		
		data_s1_i		=> data_s1_arr_s(6),
		data_s4_i		=> data_s4_s,
		pred_err_bit_o	=> pred_err_bit_s
	);
	
	-- This IP is enabled together with "dbl_res_pred_error", as both generate the input for IP "weights_vector"
	i_w_upd_scal_exp : weight_upd_scal_exp
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable7_s,
		enable_o		=> open,
		img_coord_i		=> img_coord7_s,
		img_coord_o		=> open,
		
		w_scal_exp_o	=> w_scal_exp_s
	);
	
	i_weights_vector : weights_vector
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		PREDICT_MODE_G	=> PREDICT_MODE_G,
		W_INIT_TYPE_G	=> W_INIT_TYPE_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable8_s,
		enable_o		=> open,
		img_coord_i		=> img_coord8_s,
		img_coord_o		=> open,
		
		pred_err_bit_i	=> pred_err_bit_s,
		w_scal_exp_i	=> w_scal_exp_s,
		ldiff_vect_i	=> ldiff_vect_arr_s(3),
		weight_vect_o	=> weight_vect_s
	);
	
	i_pr_ctrl_ldiff : pred_ctrl_local_diff
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		PREDICT_MODE_G	=> PREDICT_MODE_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable4_s,
		enable_o		=> enable5_s,
		img_coord_i		=> img_coord4_s,
		img_coord_o		=> img_coord5_s,
		
		pz_i			=> pz_arr_s(3),
		ldiff_vect_i	=> ldiff_vect_s,
		weight_vect_i	=> weight_vect_s,
		
		data_pred_cldiff_o => data_pred_cldiff_s
	);

	i_high_res_pred_smpl : high_res_pred_smpl
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		SMPL_LIMIT_G	=> SMPL_LIMIT_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable5_s,
		enable_o		=> enable6_s,
		img_coord_i		=> img_coord5_s,
		img_coord_o		=> img_coord6_s,
		
		data_pred_cldiff_i => data_pred_cldiff_s,
		data_lsum_i		=> data_lsum_arr_s(2),
		data_s6_o		=> data_s6_s
	);
	
	-- Delay of one complete spectral band to get value (z-1)
	i_sh_reg_s0z1_dbl_res : shift_register
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		DATA_SIZE_G		=> D_C,
		REG_SIZE_G		=> NEXT_Z_C,
		INIT_VAL_G		=> (others => '0')
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable5_s,
		enable_o		=> open,
		img_coord_i		=> img_coord5_s,
		img_coord_o		=> open,
		
		data_i			=> data_s0_arr_s(5),	-- Might seem to be 'data_s0_arr_s(4)', but it needs one extra delay to first produce 'output z', so that later produce 'output z-1'
		data_o			=> data_s0z1_s
	);

	i_dbl_res_pred_smpl : dbl_res_pred_smpl
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		SMPL_LIMIT_G	=> SMPL_LIMIT_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable6_s,
		enable_o		=> enable7_s,
		img_coord_i		=> img_coord6_s,
		img_coord_o		=> img_coord7_s,
		
		data_s0z1_i		=> data_s0z1_s,
		data_s6_i		=> data_s6_s,
		data_s4_o		=> data_s4_s
	);

	i_pred_smpl : predicted_sample
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		enable_i		=> enable7_s,
		enable_o		=> enable8_s,
		img_coord_i		=> img_coord7_s,
		img_coord_o		=> img_coord8_s,

		data_s4_i		=> data_s4_s,
		data_s3_o		=> data_s3_s
	);

	-- Outputs
	enable_o	  <= enable8_s;
	img_coord_o	  <= img_coord8_s;
	data_s3_o	  <= data_s3_s;
	data_s4_lsb_o <= data_s4_lsb_s;
	data_s6_o	  <= data_s6_arr_s(1);
end architecture behavioural;