--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		01/11/2020
--------------------------------------------------------------------------------
-- IP name:		dbl_res_smpl_repr
--
-- Description: Computes the double-resolution sample representative "s~''z(t)"
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.utils_predictor.all;

entity dbl_res_smpl_repr is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	  : std_logic_vector(1 downto 0);
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G : std_logic_vector(1 downto 0)
	);
	port (
		clock_i		: in  std_logic;
		reset_i		: in  std_logic;
		
		enable_i	: in  std_logic;
		enable_o	: out std_logic;
		img_coord_i	: in  img_coord_t;
		img_coord_o	: out img_coord_t;

		data_merr_i	: in  std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)"		(maximum error)
		data_quant_i: in  std_logic_vector(D_C+2-1 downto 0);	-- "qz(t)"		(quantizer index)		
		data_s6_i	: in  std_logic_vector(Re_C-1 downto 0);	-- "s)z(t)"		(high-resolution predicted sample)
		data_s1_i	: in  std_logic_vector(D_C-1 downto 0);		-- "s'z(t)"		(clipped quantizer bin center)
		
		data_s5_o	: out std_logic_vector(Re_C-1 downto 0)		-- "s~''z(t)"	(double-resolution sample representative)
	);
end entity dbl_res_smpl_repr;

architecture behavioural of dbl_res_smpl_repr is
	constant PW_OM1_C	: std_logic_vector(OMEGA_C+3+THETA_C-1 downto 0) := (OMEGA_C+1 => '1', others => '0');
	
	-- THETA_C can be '4' as maximum, so integer signals can be used here safely
	signal fi_s			: integer range 0 to (2**THETA_C-1) := 0;
	signal psi_s		: integer range 0 to (2**THETA_C-1) := 0;
	
	signal enable_s		: std_logic := '0';
	signal img_coord_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal data_s5_s	: std_logic_vector(Re_C-1 downto 0) := (others => '0');
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_dbl_res_smpl_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s <= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_dbl_res_smpl_delay;
	
	-- Current sample representative damping and offset values, updated according to the spectral band Z
	fi_s  <= FI_ARR_C(img_coord_i.z) when enable_i = '1' else 0;
	-- 4.9.1.2: If lossless fidelity control is used, ѱz = 0 for all spectral bands z
	psi_s <= PSI_ARR_C(img_coord_i.z) when (enable_i = '1' and FIDEL_CTRL_TYPE_G /= LOSSLESS_C) else 0;

	-- Double-resolution sample representative (s~''z(t)) calculation
	p_dbl_res_smpl_repr_calc : process(clock_i) is
		variable comp1_v		  : std_logic_vector(7 downto 0)			  := (others => '0');
		variable comp2_v, comp6_v : std_logic_vector(Re_C+THETA_C-1 downto 0) := (others => '0');
		variable comp3_v		  : integer range -1 to 1					  := 0;
		variable comp4_v, comp5_v : std_logic_vector(Re_C*2-1 downto 0)		  := (others => '0');
		variable comp7_v		  : std_logic_vector(Re_C*2+D_C-1 downto 0)	  := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				comp1_v   := (others => '0');
				comp2_v   := (others => '0');
				comp3_v   := 0;
				comp4_v   := (others => '0');
				comp5_v   := (others => '0');
				comp6_v   := (others => '0');
				comp7_v   := (others => '0');
				data_s5_s <= (others => '0');
			else
				if (enable_i = '1') then
					comp1_v   := std_logic_vector(to_signed(4 * (2**THETA_C - fi_s), comp1_v'length));
					comp2_v   := std_logic_vector(resize(shift_sgn(resize(signed(data_s1_i), Re_C+THETA_C), OMEGA_C), Re_C+THETA_C));
					comp3_v   := sgn(signed(data_quant_i));
					comp4_v   := std_logic_vector(resize(to_signed(comp3_v, Re_C*2) * resize(signed(data_merr_i), Re_C*2) * resize(shift_sgn(to_signed(psi_s, Re_C*2), OMEGA_C-THETA_C), Re_C*2), Re_C*2));
					comp5_v   := std_logic_vector(resize(signed(comp2_v) - signed(comp4_v), Re_C*2));
					comp6_v   := std_logic_vector(resize(to_signed(fi_s, Re_C+THETA_C) * (resize(signed(data_s6_i), Re_C+THETA_C) - resize(signed(PW_OM1_C), Re_C+THETA_C)), Re_C+THETA_C));
					comp7_v	  := std_logic_vector(resize(resize(signed(comp1_v), Re_C*2+D_C) * resize(signed(comp5_v), Re_C*2+D_C) + resize(signed(comp6_v), Re_C*2+D_C), Re_C*2+D_C));
					data_s5_s <= std_logic_vector(resize(shift_sgn(signed(comp7_v), -(OMEGA_C+THETA_C+1)), Re_C));
				end if;
			end if;
		end if;
	end process p_dbl_res_smpl_repr_calc;
	
	-- Outputs
	enable_o	<= enable_s;
	img_coord_o	<= img_coord_s;
	data_s5_o	<= data_s5_s;
end architecture behavioural;