--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		01/11/2020
--------------------------------------------------------------------------------
-- IP name:		sample_representative
--
-- Description: Computes the sample_representative "s''z(t)"
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;
use work.comp_predictor.all;

entity sample_representative is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	  : std_logic_vector(1 downto 0);
		-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
		SMPL_LIMIT_G	  : smpl_lim_t;
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G : std_logic_vector(1 downto 0)
	);
	port (
		clock_i		: in  std_logic;
		reset_i		: in  std_logic;
		
		enable_i	: in  std_logic;
		enable_o	: out std_logic;
		img_coord_i	: in  img_coord_t;
		img_coord_o	: out img_coord_t;
		
		data_merr_i	: in  std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)"	 (maximum error)
		data_quant_i: in  std_logic_vector(D_C+2-1 downto 0);	-- "qz(t)"   (quantizer index)
		data_s0_i	: in  std_logic_vector(D_C-1 downto 0);		-- "sz(t)"	 (original sample)
		data_s3_i	: in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)"  (predicted sample)
		data_s6_i	: in  std_logic_vector(Re_C-1 downto 0);	-- "s)z(t)"	 (high-resolution predicted sample)

		data_s1_o	: out std_logic_vector(D_C-1 downto 0);		-- "s'z(t)"  (clipped quantizer bin center)
		data_s2_o	: out std_logic_vector(D_C-1 downto 0)		-- "s''z(t)" (sample representative)
	);
end entity sample_representative;

architecture behavioural of sample_representative is
	signal enable1_s	: std_logic	  := '0';
	signal enable2_s	: std_logic	  := '0';
	signal enable3_s	: std_logic	  := '0';
	signal img_coord1_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord2_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord3_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal data_merr_s	: std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	signal data_quant_s	: std_logic_vector(D_C+2-1 downto 0) := (others => '0');
	signal data_s0_1s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_s0_2s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_s1_1s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_s1_2s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_s1_3s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_s2_s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_s5_s	: std_logic_vector(Re_C-1 downto 0)	 := (others => '0');
	signal data_s6_s	: std_logic_vector(Re_C-1 downto 0)	 := (others => '0');
	
	signal data_s0_3s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal diff_s		: std_logic							 := '0';
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_smpl_repr_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable3_s	 <= '0';
				img_coord3_s <= reset_img_coord(SMPL_ORDER_G);
				data_merr_s	 <= (others => '0');
				data_quant_s <= (others => '0');
				data_s0_1s	 <= (others => '0');
				data_s0_2s	 <= (others => '0');
				data_s1_2s	 <= (others => '0');
				data_s1_3s	 <= (others => '0');
				data_s6_s	 <= (others => '0');
			else
				enable3_s	 <= enable2_s;
				img_coord3_s <= img_coord2_s;
				data_merr_s	 <= data_merr_i;
				data_quant_s <= data_quant_i;
				data_s0_1s	 <= data_s0_i;
				data_s0_2s	 <= data_s0_1s;
				data_s1_2s	 <= data_s1_1s;
				data_s1_3s	 <= data_s1_2s;
				data_s6_s	 <= data_s6_i;
			end if;
		end if;
	end process p_smpl_repr_delay;
	
	-- Clipped quantizer bin center "s'z(t)" calculation
	i_clip_qua_bin_cnt : clip_quant_bin_center
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		SMPL_LIMIT_G => SMPL_LIMIT_G
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => enable1_s,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => img_coord1_s,
		
		data_s3_i	 => data_s3_i,
		data_merr_i	 => data_merr_i,
		data_quant_i => data_quant_i,
		data_s1_o	 => data_s1_1s
	);

	-- Sample representative "s''z(t)" calculation
	i_dbl_res_smpl_repr : dbl_res_smpl_repr
	generic map(
		SMPL_ORDER_G	  => SMPL_ORDER_G,
		FIDEL_CTRL_TYPE_G => FIDEL_CTRL_TYPE_G
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable1_s,
		enable_o	 => enable2_s,
		img_coord_i	 => img_coord1_s,
		img_coord_o	 => img_coord2_s,

		data_merr_i	 => data_merr_s,
		data_quant_i => data_quant_s,
		data_s6_i	 => data_s6_s,
		data_s1_i	 => data_s1_1s,
		data_s5_o	 => data_s5_s
	);

	-- Sample representative (s''z(t)) calculation
	p_smpl_repr_calc : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				data_s2_s <= (others => '0');
			else
				if (enable2_s = '1') then
					if (img_coord2_s.t = 0) then
						data_s2_s <= data_s0_2s;
					else	-- img_coord2_s.t > 0
						data_s2_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s5_s) + to_signed(1, Re_C), Re_C), -1), D_C));
					end if;
					
					data_s0_3s <= data_s0_2s;
					if (data_s0_3s /= data_s2_s) then
						diff_s <= '1';
					else
						diff_s <= '0';
					end if;
				end if;
			end if;
		end if;
	end process p_smpl_repr_calc;

	-- Outputs
	enable_o	<= enable3_s;
	img_coord_o	<= img_coord3_s;
	data_s1_o	<= data_s1_3s;
	data_s2_o	<= data_s2_s;
end architecture behavioural;