--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		13/10/2021
--------------------------------------------------------------------------------
-- IP name:		pred_enc_interconn
--
-- Description: Manages data going from Predictor's output into Encoder's input
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.error_limit_values_table.all;

use work.comp_encoder.all;
	
entity pred_enc_interconn is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic
	);
	port (
		clock_i				: in  std_logic;
		reset_i				: in  std_logic;
		
		enable_i			: in  std_logic;
		enable_o			: out std_logic;
		img_coord_i			: in  img_coord_t;
		img_coord_o			: out img_coord_t;
		
		init_fifo_o			: out std_logic;
		
		err_lim_upd_en_o	: out std_logic;
		err_lim_val_o		: out std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i		: in  std_logic_vector(D_C-1 downto 0);
        data_mp_quan_o		: out std_logic_vector(D_C-1 downto 0)
	);
end entity pred_enc_interconn;

architecture behavioural of pred_enc_interconn is	
	-- Constants declaration
	-- FIFO depth is the error limit values matrix worst case: matrix size * array size twice (for absolute and relative array values).
	-- As such depth must be a power of 2, the power is extracted and rounded-up (log2), and the proper value finally calculated (power2)
	constant FIFO_DATA_DEPTH_C	: integer := to_integer(unsigned(power2(log2(ERR_LIM_VAL_ARR_C'length * (ERR_LIM_VAL_ARR_C(0).abs_arr'length * 2)))));
	-- Number of parallel sync. FIFOs to use (apart from the one for the 'image coordinates') -> Here 2, for "data_mp_quan_i" and "enable_i"
	constant NUM_PAR_FIFOS_C	: integer := 2;
	
	-- Signals declaration
	signal enable_s			: std_logic	  := '0';
	signal img_coord_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal init_fifo_s		: std_logic	  := '0';
	signal y_old_s			: integer	  := -1;
	
	signal err_arr_cnt_s	: integer range 0 to NZ_C := 0;
	signal abs_rel_sel_s	: std_logic	  := '0';
	signal err_lim_latch_s	: std_logic	  := '0';
	signal err_lim_upd_en_s	: std_logic	  := '0';
	
	signal data_arr_in_s	: array_slv_t(0 to NUM_PAR_FIFOS_C-1)(D_C-1 downto 0) := (others => (others => '0'));
	signal data_arr_out_s	: array_slv_t(0 to NUM_PAR_FIFOS_C-1)(D_C-1 downto 0) := (others => (others => '0'));

	signal err_lim_val_s	: std_logic_vector(D_C-1 downto 0) := (others => '0');
	signal data_mp_quan_s	: std_logic_vector(D_C-1 downto 0) := (others => '0');

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_interconn_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				y_old_s <= -1;
			else
				if (enable_i = '1') then
					y_old_s <= img_coord_i.y;
				end if;
			end if;
		end if;
	end process p_interconn_delay;

	-- Outputs the proper error limit values every 2**U_C clock cycles
	p_err_lim_val_calc : process(clock_i) is
		variable pointer_v : integer range 0 to ERR_LIM_ARR_SIZE_C := 0;
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				pointer_v		 := 0;
				err_arr_cnt_s	 <= 0;
				abs_rel_sel_s	 <= '0';
				err_lim_latch_s	 <= '0';
				err_lim_upd_en_s <= '0';
				err_lim_val_s	 <= (others => '0');
			else				
				if (enable_i = '1') then
					if (((img_coord_i.y /= y_old_s) and (img_coord_i.y mod 2**U_C = 0)) or err_lim_latch_s = '1') then
						err_lim_latch_s  <= '1';
						err_lim_upd_en_s <= '1';	-- To inform when new error limit values are being sent out
						if (pointer_v >= ERR_LIM_VAL_ARR_C'length) then
							pointer_v := 0;
						end if;
						
						case FIDEL_CTRL_TYPE_G is
							when ABS_ERR_C =>
								if (ABS_ERR_BAND_TYPE_G = '0') then		-- Band-independent
									err_lim_val_s	<= std_logic_vector(to_unsigned(ERR_LIM_VAL_ARR_C(pointer_v).abs_c, D_C));
									pointer_v		:= pointer_v + 1;
									err_lim_latch_s	<= '0';
								else	-- ABS_ERR_BAND_TYPE_G = '1'	-- Band-dependent
									err_lim_val_s <= std_logic_vector(to_unsigned(ERR_LIM_VAL_ARR_C(pointer_v).abs_arr(err_arr_cnt_s), D_C));
									
									if (err_arr_cnt_s < ERR_LIM_VAL_ARR_C(pointer_v).abs_arr'length-1) then
										err_arr_cnt_s <= err_arr_cnt_s + 1;
									else
										err_arr_cnt_s	<= 0;
										pointer_v		:= pointer_v + 1;
										err_lim_latch_s <= '0';
									end if;
								end if;
								
							when REL_ERR_C =>
								if (REL_ERR_BAND_TYPE_G = '0') then		-- Band-independent
									err_lim_val_s	<= std_logic_vector(to_unsigned(ERR_LIM_VAL_ARR_C(pointer_v).rel_c, D_C));
									pointer_v		:= pointer_v + 1;
									err_lim_latch_s <= '0';
								else	-- REL_ERR_BAND_TYPE_G = '1'	-- Band-dependent
									err_lim_val_s	<= std_logic_vector(to_unsigned(ERR_LIM_VAL_ARR_C(pointer_v).rel_arr(err_arr_cnt_s), D_C));
									
									if (err_arr_cnt_s < ERR_LIM_VAL_ARR_C(pointer_v).rel_arr'length-1) then
										err_arr_cnt_s <= err_arr_cnt_s + 1;
									else
										err_arr_cnt_s	<= 0;
										pointer_v		:= pointer_v + 1;
										err_lim_latch_s <= '0';
									end if;
								end if;

							when ABS_REL_C =>
								if (abs_rel_sel_s = '0') then			-- Absolute error limits selected
									if (ABS_ERR_BAND_TYPE_G = '0') then		-- Band-independent
										err_lim_val_s <= std_logic_vector(to_unsigned(ERR_LIM_VAL_ARR_C(pointer_v).abs_c, D_C));
										abs_rel_sel_s <= '1';
									else	-- ABS_ERR_BAND_TYPE_G = '1'	-- Band-dependent
										err_lim_val_s <= std_logic_vector(to_unsigned(ERR_LIM_VAL_ARR_C(pointer_v).abs_arr(err_arr_cnt_s), D_C));
										
										if (err_arr_cnt_s < ERR_LIM_VAL_ARR_C(pointer_v).abs_arr'length-1) then
											err_arr_cnt_s <= err_arr_cnt_s + 1;
										else
											err_arr_cnt_s <= 0;
											abs_rel_sel_s <= '1';
										end if;
									end if;
								else	-- abs_rel_sel_s = '1'			-- Relative error limits selected
									if (REL_ERR_BAND_TYPE_G = '0') then		-- Band-independent
										err_lim_val_s	<= std_logic_vector(to_unsigned(ERR_LIM_VAL_ARR_C(pointer_v).rel_c, D_C));
										pointer_v		:= pointer_v + 1;
										abs_rel_sel_s	<= '0';
										err_lim_latch_s <= '0';
									else	-- REL_ERR_BAND_TYPE_G = '1'	-- Band-dependent
										err_lim_val_s <= std_logic_vector(to_unsigned(ERR_LIM_VAL_ARR_C(pointer_v).rel_arr(err_arr_cnt_s), D_C));
										
										if (err_arr_cnt_s < ERR_LIM_VAL_ARR_C(pointer_v).rel_arr'length-1) then
											err_arr_cnt_s <= err_arr_cnt_s + 1;
										else
											err_arr_cnt_s	<= 0;
											pointer_v		:= pointer_v + 1;
											abs_rel_sel_s	<= '0';
											err_lim_latch_s <= '0';
										end if;
									end if;
								end if;

							when others =>		-- when LOSSLESS_C
								err_lim_upd_en_s <= '0';
								err_lim_val_s	 <= (others => '0');
						end case;							
					else
						err_lim_upd_en_s <= '0';
						err_lim_val_s	 <= (others => '0');
					end if;
				end if;
			end if;
		end if;
	end process p_err_lim_val_calc;
	
	-- Parallel sync FIFOs to store Predictor output data (and its coordinates) when 'Error Limit Values' must be updated
	i_parallel_sync_fifos : parallel_sync_fifos
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		DATA_WIDTH_G	=> D_C,
		DATA_DEPTH_G	=> iif(FIFO_DATA_DEPTH_C < 16, 16, FIFO_DATA_DEPTH_C),		-- XPM Sync FIFO requires this field to be at least 16...
		NUM_PAR_FIFOS_G	=> NUM_PAR_FIFOS_C
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,

		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => img_coord_s,
		
		read_fifo_i	 => not err_lim_upd_en_s,	-- Data shall stall when 'Error Limit Values' are NOT being updated
		init_fifo_o	 => init_fifo_s,
		
		data_arr_i	 => data_arr_in_s,
		data_arr_o	 => data_arr_out_s
	);
	-- To avoid strange behaviour from the FIFOs and for the sake of simulation, nothing is introduced on signals until "enable_i = '1'"
	data_arr_in_s(0)	<= data_mp_quan_i when enable_i = '1' else (D_C-1 downto 0 => '0');
	data_arr_in_s(1)(0) <= enable_i;
	data_mp_quan_s		<= data_arr_out_s(0);
	enable_s			<= data_arr_out_s(1)(0);
	
	-- Outputs
	enable_o         <= enable_s;
	img_coord_o      <= img_coord_s;
	init_fifo_o      <= init_fifo_s;
    err_lim_upd_en_o <= err_lim_upd_en_s;
	err_lim_val_o	 <= err_lim_val_s;
	data_mp_quan_o   <= data_mp_quan_s;
end architecture behavioural;