####################################################################################################
# Synthesis/Implementation processes & BitStream/HW Description files generation
####################################################################################################

# Xilinx Vivado "Synthesis" process execution
proc run_synthesis {ooc_enable cpu_cores} {

	# Out-Of-Context option disabled for all Block Designs and IPs
	if { ${ooc_enable} == false } {
		# Disable Out Of Context (OOC) build for all Block Designs
		foreach f [get_files -filter {FILE_TYPE == "Block Designs"}] {
			catch {set_property GENERATE_SYNTH_CHECKPOINT 0 $f}
		}
		# Disable Out Of Context (OOC) build for all IPs
		foreach f [get_files -filter {FILE_TYPE == IP}] {
			catch {set_property GENERATE_SYNTH_CHECKPOINT 0 $f}
		}
		puts " -----> INFO: Synthesis Out-Of-Context (OOC) DISABLED!"
		
	} else {
		puts " -----> INFO: Synthesis Out-Of-Context (OOC) ENABLED!"
	}
	
	puts " -----> INFO: SYNTHESIS PROCESS STARTS (using ${cpu_cores} CPU cores)!"
	# Resets the specified run to an unsynthesized state, just in case for a clean run.
	reset_run synth_1

	# Launches synthesis runs when running the Vivado tools in Project Mode.
	# "-jobs arg". The number of parallel jobs to run on the local host.
	launch_runs synth_1 -jobs ${cpu_cores}

	# Blocks execution of TCL commands until the specified run has finished, either successfully or in error (or until the specified amount of time has elapsed).
	wait_on_run synth_1
	puts " -----> INFO: SYNTHESIS PROCESS FINISHED!"
}

# Xilinx Vivado "Implementation" process execution
proc run_implementation {cpu_cores origin_dir viv_proj_name entity_name} {

	puts " -----> INFO: IMPLEMENTATION PROCESS STARTS (using ${cpu_cores} CPU cores)!"
	# Resets the specified run to an unimplemented state, just in case for a clean run.
	reset_run impl_1

	# Launches "implementation runs", and generates bitstream file for Xilinx device config
	launch_runs impl_1 -to_step write_bitstream -jobs ${cpu_cores}

	# Blocks execution of TCL commands until the specified run has finished, either successfully or in error.
	wait_on_run impl_1

	# Opens the specified "implementation run" into an "Implemented Design"
	open_run impl_1
	puts " -----> INFO: IMPLEMENTATION PROCESS FINISHED!"
	
	# A folder for all necessary output files is created
	exec mkdir -p ${origin_dir}/artifacts
	
	# Relative path into the implementation (impl_1) folder
	set impl_1_path "${viv_proj_name}/${viv_proj_name}.runs/impl_1"
	
	# Copies final bitstream file (.bit) into folder "artifacts"
	file copy -force ${origin_dir}/${impl_1_path}/${entity_name}.bit ${origin_dir}/artifacts/${entity_name}.bit
	puts " -----> INFO: Bitstream (.bit) file saved into folder ARTIFACTS!"

	# In case it exists, copies ILA/VIO probes file (.ltx) into folder "artifacts"
	if { [file exists ${origin_dir}/${impl_1_path}/${entity_name}.ltx] } {
		file copy -force ${origin_dir}/${impl_1_path}/${entity_name}.ltx ${origin_dir}/artifacts/${entity_name}.ltx
		puts " -----> INFO: ILA/VIO probes (.ltx) file saved into folder ARTIFACTS!"
	}
	
	# Generate relevant (timing and utilization) reports
	# report_timing_summary -delay_type min_max -report_unconstrained -check_timing_verbose -max_paths 10 -input_pins -routable_nets -name timing_1 -file $origin_dir/artifacts/timing_report.txt
	# report_utilization -file ${origin_dir}/artifacts/resource_usage_top.rpt -hierarchical -hierarchical_depth 1 -hierarchical_percentages
	# report_utilization -file ${origin_dir}/artifacts/resource_usage.rpt -hierarchical -hierarchical_depth 4 -hierarchical_percentages
}

# Xilinx Vivado HW Description file generation
proc gen_hw_files {origin_dir viv_proj_name sdk_proj_name entity_name} {
	
	# write_hwdef -force -file ${origin_dir}/artifacts/${entity_name}.hdf
	file copy -force ${origin_dir}/${viv_proj_name}/${viv_proj_name}.runs/impl_1/${entity_name}.sysdef ${origin_dir}/artifacts/${entity_name}.hdf
	puts " -----> INFO: Hardware Description File (.hdf) saved into folder ARTIFACTS!"

	# The above ".hdf" file is copied into the necessary folder to be used by Xilinx SDK
	file copy -force ${origin_dir}/artifacts/${entity_name}.hdf ${origin_dir}/../sw/${sdk_proj_name}.sdk/${entity_name}.hdf
	puts " -----> INFO: Hardware Description File (.hdf) moved to Xilinx ${sdk_proj_name} SDK folder!"
}